/**CFile***********************************************************************

  FileName    [cuplusProfile.c]

  PackageName [cuplus]

  Synopsis    [Managing Activity profiles.]

  Description [This file includes routines for creating activity profiles,
    collecting statistics within recursive And-Abstract operations,
    pruning BDDs using activity profiles.
    Since CUDD does not provide room for extra pointers within
    the BDD node structure, we use hashing (st package)
    to implement the correspondence between a node and a profile
    info.
    ]

  SeeAlso     []

  Author      [Gianpiero Cabodi and Stefano Quer]

  Copyright   [This file was created at the Politecnico di Torino,
    Torino, Italy.
    The  Politecnico di Torino makes no warranty about the suitability of
    this software for any purpose.
    It is presented on an AS IS basis.]

  Revision    []

******************************************************************************/

#include "cuplusInt.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

# define USE_DEAD 1

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/

struct profile_info_summary {
  st_table *table;
  unsigned char dac99_compatible;
  unsigned char total_summary_done;
  unsigned char node_summary_done;
  int total_nodes;
  int total_active_nodes;
  double total_full_subtrees;
  double total_non_zero_nodes;
  double total_cached_nodes;
  double total_recursions;
  double total_cache_hits;
  double total_size_cost;
  double avg_recursions;
  double avg_cache_hits;
  double avg_size_cost;
  double std_recursions;
  double std_cache_hits;
  double std_size_cost;
  double max_recursions;
  double max_cache_hits;
  double max_norm_size_cost;
  double min_norm_size_cost;
  double max_BDD_size;
  int max_var_index;
  int profile[10];
  int profile_by_var[10][10];
  int profile_by_size[10][10];
  int rec_profile_by_var[10][10];
  int cache_profile_by_var[10][10];
  int *size_by_var;
};

typedef struct profile_info_item {
  int profile_subtree;
  int cache_hits;
  int node_visit_cnt;
  int size_cost;
  int norm_size_cost;
  int tot_act;
  int sharing;
  DdNode *ref_node;
  DdNode *pruned_bdd;
  struct profile_info_item *next;
  int num;
} profile_info_item_t;

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

#define my_square(a) (((double)a)*((double)a))

#if USE_DEAD
# define MANAGER_KEYS(manager,profile) \
((profile->dac99_compatible)?(manager->keys):(manager->keys - manager->dead))
#else
# define MANAGER_KEYS(manager,profile) (manager->keys)
#endif

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static void ProfileTotalSummary(cuplus_profile_info_t *profile);
static enum st_retval ProfileCentralSummary(char * key, char * value, char * arg);
static enum st_retval ProfileNodeSummary(char * key, char * value, char * arg);
static enum st_retval ProfileInfoItemFree(char * key, char * value, char * arg);
static DdNode * cuplusProfileAndAbstractRecur(DdManager *manager, DdNode *f, DdNode *g, DdNode *cube, cuplus_profile_info_t *profile);
static DdNode * cuplusPruneProfiledRecur(DdManager * manager, DdNode * f, cuplus_profile_info_t *profile, Cuplus_PruneHeuristic_e prune_heuristic, int threshold, int upper_activity);
static void cuplusCollectProfileSummaryRecur(DdManager * manager, DdNode * f, cuplus_profile_info_t *profile);
static void BfvProfiles(DdManager * manager, cuplus_profile_info_t *profile);
static enum st_retval ProfileInfoByVars(char * key, char * value, char * arg);
static void ddClearFlag(DdNode * f);

/**AutomaticEnd***************************************************************/

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis [Takes the AND of two BDDs and simultaneously abstracts the
    variables in cube.]

  Description [A patched version of the original CUDD function.]

  SideEffects [None]

  SeeAlso     [Cudd_bddAndAbstract]

******************************************************************************/

DdNode *
Cuplus_ProfileAndAbstract(
  DdManager *manager,
  DdNode    *f,
  DdNode    *g,
  DdNode    *cube,
  cuplus_profile_info_t *profile
)
{
  DdNode *res;

  /* reordering NOT allowed in present implementation ! */
  manager->reordered = 0;

  res = cuplusProfileAndAbstractRecur(manager, f, g, cube, profile);

  assert (manager->reordered == 0);

  return(res);
}

/**Function********************************************************************

  Synopsis    []

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

DdNode *
Cuplus_ProfilePrune(
  DdManager *manager,
  DdNode *f,
  cuplus_profile_info_t *profile,
  Cuplus_PruneHeuristic_e prune_heuristic,
  int threshold
)
{
  DdNode *res;
  st_table *table = profile->table;
  struct profile_info_item *info_item_ptr;
  int nvars;

  nvars = Cudd_ReadSize (manager);

  ProfileTotalSummary(profile);

#if 0
  if ((profile->size_by_var = ALLOC(int,nvars)) == NULL) {
    fprintf (stderr, "Out of memory collecting profile summary\n");
    return (NULL);
  }

  for (i=0; i<nvars; i++) {
    profile->size_by_var[i] = 0;
  }
#endif

  cuplusCollectProfileSummaryRecur (manager, f, profile);
  ddClearFlag (Cudd_Regular (f));

  BfvProfiles (manager, profile);

#if 0
  fprintf (stdout, "Var-by-size:");
  for (i=0; i<nvars; i++)
    fprintf (stdout, "%d ", profile->size_by_var[i]);
  fprintf (stdout, "\n");
#endif

  if (st_lookup(table, (char *)f, (char **)&info_item_ptr)) {
    fprintf (stdout, "  Root size cost/BDD size  : %d/%d\n",
      info_item_ptr->size_cost, Cudd_DagSize(f));
  }

  /* Reordering NOT allowed in present implementation! */
  manager->reordered = 0;

  res = cuplusPruneProfiledRecur(manager, f, profile,
      prune_heuristic, threshold, 0);

  assert (manager->reordered == 0);

  return(res);
}

/**Function********************************************************************

  Synopsis [Create a new profile]

  Description [Allocate a structure for global statistics of a profile.
    Initialize fields. Create a symbol table for the correspondence
    node-profile info. The dac99_compatible flag must be 1
    for experimental results as in the dac99 paper. It must be set
    now.
    ]

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

cuplus_profile_info_t *
Cuplus_ProfileInfoGen(
  unsigned char dac99_compatible
)
{
  cuplus_profile_info_t *profile;
  st_table *table;

  table = st_init_table (st_ptrcmp, st_ptrhash);
  if (table == NULL) {
    fprintf (stderr,
      "Error : Out-of-memory, couldn't alloc profile info table\n");
    return (NULL);
  }
  profile = ALLOC(cuplus_profile_info_t,1);
  if (profile == NULL) {
    st_free_table(table);
    return(NULL);
  }
  profile->table = table;

  profile->dac99_compatible = dac99_compatible;
  profile->total_summary_done = 0;
  profile->node_summary_done = 0;
  profile->total_nodes = 0;
  profile->total_active_nodes = 0;
  profile->total_full_subtrees = 0;
  profile->total_cached_nodes = 0;
  profile->total_non_zero_nodes = 0;
  profile->max_recursions = 0;
  profile->max_cache_hits = 0;
  profile->max_norm_size_cost = 0;
  profile->min_norm_size_cost = 100000;
  profile->total_recursions = 0;
  profile->total_cache_hits = 0;
  profile->total_size_cost = 0;
  profile->avg_recursions = 0;
  profile->avg_cache_hits = 0;
  profile->avg_size_cost = 0;
  profile->std_recursions = 0;
  profile->std_cache_hits = 0;
  profile->std_size_cost = 0;
  profile->max_var_index = 0;
  profile->max_BDD_size = 0;

  return (profile);
}

/**Function********************************************************************

  Synopsis    [Free a profile]

  Description [Free a profile struct and the associated symbol table.]

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

void
Cuplus_ProfileInfoFree(
  cuplus_profile_info_t *profile
)
{
  st_table      *table = profile->table;

  if (table != NULL) {
    st_foreach(table, ProfileInfoItemFree, NULL);
    st_free_table(table);
  }
  FREE(profile);

  return;
}

/**Function********************************************************************

  Synopsis    [Print global profile statistics]

  Description [Print some global statistics about a profile, e.g. average and
    max values of activity indicators.]

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

void
Cuplus_ProfileInfoPrint(
  cuplus_profile_info_t *profile
)
{
#if 0
  int i,j;
#endif

  ProfileTotalSummary(profile);

  fprintf (stdout, "Total active/profiled nodes   : %d/%d\n",
    profile->total_active_nodes, profile->total_nodes);
  fprintf (stdout, "Total full subtrees           : %g\n",
    profile->total_full_subtrees);
  fprintf (stdout, "Total cached nodes            : %g\n",
    profile->total_cached_nodes);
  fprintf (stdout, "Total non 0 nodes             : %g\n",
    profile->total_non_zero_nodes);

  fprintf (stdout, "Recursions Tot/Avg/Std/Max    : %.4g/%.4g/%.4g/%.4g\n", 
    profile->total_recursions, profile->avg_recursions,
    profile->std_recursions,  profile->max_recursions);

  fprintf (stdout, "Cache hits Tot/Avg/Std/Max    : %.4g/%.4g/%.4g/%.4g\n",
    profile->total_cache_hits, profile->avg_cache_hits,
    profile->std_cache_hits, profile->max_cache_hits);

  fprintf (stdout, "Size cost Tot/Avg/Std/Min/Max : %.4g/%.4g/%g/%.4g/%.4g\n",
    profile->total_size_cost, profile->avg_size_cost,
    profile->std_size_cost, profile->min_norm_size_cost,
    profile->max_norm_size_cost);

#if 0
  fprintf (stdout, "Recursion profile by increasing var :\n");
  for (j=0; j<10; j++) {x
    fprintf (stdout, "v: %4d)", j);
    for (i=0; i<10; i++)
      fprintf (stdout, "%4d ", profile->rec_profile_by_var[i][j]);
    fprintf (stdout, "\n");
  }
  fprintf (stdout, "Cache profile by increasing var :\n");
  for (j=0; j<10; j++) {
    fprintf (stdout, "v: %4d)", j);
    for (i=0; i<10; i++)
      fprintf (stdout, "%4d ", profile->cache_profile_by_var[i][j]);
    fprintf (stdout, "\n");
  }
  fprintf (stdout, "Size profile           :\n");
  for (i=0; i<10; i++)
    fprintf (stdout, "%4d ", profile->profile[i]);
  fprintf (stdout, "\n");
  fprintf (stdout, "Size profile by increasing var :\n");
  for (j=0; j<10; j++) {
    fprintf (stdout, "v: %4d)", j);
    for (i=0; i<10; i++)
      fprintf (stdout, "%4d ", profile->profile_by_var[i][j]);
    fprintf (stdout, "\n");
  }
  fprintf (stdout, "Size profile by decreasing size :\n");
  for (j=9; j>=0; j--) {
    fprintf (stdout, "s: %4d)", j);
    for (i=0; i<10; i++)
      fprintf (stdout, "%4d ", profile->profile_by_size[i][j]);
    fprintf (stdout, "\n");
  }
  fprintf (stdout, "\n");

#endif

  return;
}

/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis [Routine computing global profile statistics.]

  Description [Collect final global profile statistics at the end of the
    learning phase.]

  SideEffects [None]

******************************************************************************/

static void
ProfileTotalSummary(
  cuplus_profile_info_t *profile
)
{
  st_table      *table = profile->table;

  /* avoid a second summary */
  if (profile->total_summary_done == 1) {
    return;
  }

  st_foreach(table, ProfileCentralSummary, (char *)(profile));

  assert (profile->total_nodes == st_count(table));
  profile->total_active_nodes = profile->total_non_zero_nodes
                       /* +profile->total_full_subtrees */ ;
  profile->avg_recursions = profile->total_recursions/profile->total_nodes;
  profile->avg_cache_hits = profile->total_cache_hits/profile->total_nodes;
  profile->avg_size_cost  = profile->total_size_cost/profile->total_nodes;

  st_foreach (table, ProfileNodeSummary, (char *)(profile));

  profile->std_recursions = sqrt(profile->std_recursions/profile->total_nodes-
    profile->avg_recursions*profile->avg_recursions);
  profile->std_cache_hits = sqrt(profile->std_cache_hits/profile->total_nodes-
    profile->avg_cache_hits*profile->avg_cache_hits);
  profile->std_size_cost  = sqrt(profile->std_size_cost/profile->total_nodes-
    profile->avg_size_cost*profile->avg_size_cost);

  profile->total_summary_done = 1;

  return;
}

/**Function********************************************************************

  Synopsis [Profile summary using node statistics.]

  Description [Profile summary using node statistics. Computing total and
    max values.]

  SideEffects [None]

******************************************************************************/

static enum st_retval
ProfileCentralSummary(
  char * key,
  char * value,
  char * arg
)
{
  cuplus_profile_info_t *profile;
  struct profile_info_item *info_item_ptr;
  double norm_size_cost, BDD_size;

  profile = (cuplus_profile_info_t *) arg;
  info_item_ptr = (struct profile_info_item *) value;

  if (info_item_ptr->profile_subtree > 0) {
    profile->total_full_subtrees++;
  }

  profile->total_cache_hits += info_item_ptr->cache_hits;
  if (profile->max_cache_hits < info_item_ptr->cache_hits) {
    profile->max_cache_hits = info_item_ptr->cache_hits;
  }
  if (info_item_ptr->cache_hits > 0) {
    profile->total_cached_nodes++;
  }

  if (info_item_ptr->node_visit_cnt > 0) {

    profile->total_recursions += info_item_ptr->node_visit_cnt;

    if (profile->max_recursions < info_item_ptr->node_visit_cnt) {
      profile->max_recursions = info_item_ptr->node_visit_cnt;
    }

    BDD_size = (double) info_item_ptr->size_cost;

    if (profile->max_BDD_size < BDD_size) {
      profile->max_BDD_size = BDD_size;
    }

    norm_size_cost = BDD_size/info_item_ptr->node_visit_cnt;

    profile->total_size_cost += norm_size_cost;

    if (profile->max_norm_size_cost < norm_size_cost) {
      profile->max_norm_size_cost = norm_size_cost;
    }

    if (profile->min_norm_size_cost > norm_size_cost) {
      profile->min_norm_size_cost = norm_size_cost;
    }

    if (profile->max_var_index < Cudd_Regular(info_item_ptr->ref_node)->index) {
      profile->max_var_index = Cudd_Regular(info_item_ptr->ref_node)->index;
    }

    profile->total_non_zero_nodes++;
  }

  return (ST_CONTINUE);
}

/**Function********************************************************************

  Synopsis [Node statistic summary. ]

  Description [Node statistic summary. Preparing average statistics
    and node normalized size costs.]

  SideEffects [None]

******************************************************************************/

static enum st_retval
ProfileNodeSummary(
  char * key,
  char * value,
  char * arg
)
{
  cuplus_profile_info_t *profile;
  struct profile_info_item *info_item_ptr;
  double norm_size_cost, size_cost;

  profile = (cuplus_profile_info_t *) arg;
  info_item_ptr = (struct profile_info_item *) value;

  profile->std_recursions +=
    my_square((info_item_ptr->node_visit_cnt));

  profile->std_cache_hits +=
    my_square((info_item_ptr->cache_hits));

  if (info_item_ptr->node_visit_cnt > 0) {
    size_cost = (double) info_item_ptr->size_cost;
    norm_size_cost = size_cost/info_item_ptr->node_visit_cnt;

    profile->std_size_cost += my_square (norm_size_cost);

    info_item_ptr->size_cost =
      (1000.0 * size_cost)/profile->max_BDD_size;
    info_item_ptr->norm_size_cost =
      (1000.0 * norm_size_cost)/profile->max_norm_size_cost;

    assert (info_item_ptr->size_cost <= 1000);

  }  else {
    info_item_ptr->size_cost = 0;
    info_item_ptr->norm_size_cost = 0;
  }

  return(ST_CONTINUE);
}

/**Function********************************************************************

  Synopsis [Frees the memory used for info items.]

  Description []

  SideEffects [None]

******************************************************************************/

static enum st_retval
ProfileInfoItemFree(
  char * key,
  char * value,
  char * arg
)
{
  struct profile_info_item *info_item_ptr;

  info_item_ptr = (struct profile_info_item *)value;
  FREE(info_item_ptr);

  return(ST_CONTINUE);
}

/**Function********************************************************************

  Synopsis    [Recursive stap of AndAbstract with profile handling.]

  Description [Takes the AND of two BDDs and simultaneously abstracts
    the variables in cube. The variables are existentially abstracted.
    Returns a pointer to the result is successful; NULL otherwise.
    Profile statistics are upgraded. Profile is associated to the g operand.]

  SideEffects [None]

  SeeAlso     [Cudd_bddAndAbstract]

******************************************************************************/

static DdNode *
cuplusProfileAndAbstractRecur(
  DdManager *manager,
  DdNode    *f,
  DdNode    *g,
  DdNode    *cube,
  cuplus_profile_info_t *profile
)
{
  DdNode *F, *fv, *fnv, *G, *gv, *gnv;
  DdNode *one, *zero, *r, *t, *e, *Cube;
  unsigned int topf, topg, topcube, top, index;
  struct profile_info_item *info_item_ptr;
  unsigned char fg_swap;
  unsigned int keys0;

  st_table  *table = profile->table;
  one = DD_ONE(manager);
  zero = Cudd_Not(one);

  /* Terminal cases. */
  if (f == zero || g == zero || f == Cudd_Not(g)) {
    return (zero);
  }

  if (f == one && g == one) {
    return(one);
  }

/*
  handled as AndAbstractRecur
  if (cube == one) {
    return(cuddBddAndRecur(manager, f, g));
  }
*/

  if (g == one) {
    /* g constant: nothing to profile */
    return (cuddBddExistAbstractRecur(manager, f, cube));
  }

  F = Cudd_Regular(f);
  G = Cudd_Regular(g);

  if (!st_lookup(table, (char *)g, (char **)&info_item_ptr)) {
    info_item_ptr = ALLOC(struct profile_info_item,1);
    if (info_item_ptr == NULL) {
      return(NULL);
    }
   
    if (st_add_direct(table,(char *)g,(char *)info_item_ptr)==ST_OUT_OF_MEM) {
      FREE(info_item_ptr);
      return(NULL);
    }

    info_item_ptr->profile_subtree = 0;
    info_item_ptr->cache_hits = 0;
    info_item_ptr->node_visit_cnt = 0;
    info_item_ptr->ref_node = g;
    info_item_ptr->pruned_bdd = NULL;
    info_item_ptr->size_cost = 0;
    info_item_ptr->num = (profile->total_nodes)++;
  }

  keys0 = MANAGER_KEYS(manager,profile);

  if (f == one || f == g) {
    /* g must be all profiled */
    info_item_ptr->profile_subtree++;
    r = cuddBddExistAbstractRecur(manager, g, cube);
    info_item_ptr->size_cost += (MANAGER_KEYS(manager,profile) - keys0);
    return r;
  }

  /* At this point f and g are not constant. */

  fg_swap = 0;

  if (profile->dac99_compatible) {
  /* disabled because profiling is on g */
  } else {
    if (f < g) { /* Try to increase cache efficiency. */
      fg_swap = 1;
    }
  }

  /* Check cache. */
  if (fg_swap)
    r = cuddCacheLookup(manager, DD_BDD_AND_ABSTRACT_TAG, g, f, cube);
  else
    r = cuddCacheLookup(manager, DD_BDD_AND_ABSTRACT_TAG, f, g, cube);
  if (r != NULL) {
    if (r != zero)
      info_item_ptr->cache_hits++;
      return(r);
  }

  /*
   *  Here we can skip the use of cuddI, because the operands are known
   *  to be non-constant.
   */

  topf = manager->perm[F->index];
  topg = manager->perm[G->index];
  top = ddMin(topf, topg);

  if (cube != one) {
    topcube = manager->perm[cube->index];
    if (topcube < top) {
      return(cuplusProfileAndAbstractRecur(manager,
        f, g, cuddT(cube), profile));
    }
  } else
    topcube = one->index;
    /* Now, topcube >= top. */

  if (topf == top) {
    index = F->index;
    fv = cuddT(F);
    fnv = cuddE(F);
    if (Cudd_IsComplement(f)) {
      fv = Cudd_Not(fv);
      fnv = Cudd_Not(fnv);
    }
  } else {
    index = G->index;
    fv = fnv = f;
  }

  if (topg == top) {
    gv = cuddT(G);
    gnv = cuddE(G);
    if (Cudd_IsComplement(g)) {
      gv = Cudd_Not(gv);
      gnv = Cudd_Not(gnv);
    }
  } else {
    gv = gnv = g;
  }

  if ((cube != one)&&(topcube == top)) {
    Cube = cuddT(cube);
  } else {
    Cube = cube;
  }

  if (fv == zero) {
    t = zero;
  } else {
    t = cuplusProfileAndAbstractRecur(manager, fv, gv, Cube, profile);
    if (t == NULL) return(NULL);
  }

#if 0
    /* Special case: 1 OR anything = 1. Hence, no need to compute
    ** the else branch if t is 1.
    */
    if (t == one && topcube == top) {
      info_item_ptr->node_visit_cnt++;
      cuddCacheInsert(manager, DD_BDD_AND_ABSTRACT_TAG, f, g, cube, one);
      info_item_ptr->size_cost += (MANAGER_KEYS(manager,profile) - keys0);
        return(one);
    }
#endif
  
  cuddRef(t);

  if (fnv == zero) {
    e = zero;
  } else {
    e = cuplusProfileAndAbstractRecur(manager, fnv, gnv, Cube, profile);
    if (e == NULL) {
      Cudd_RecursiveDeref(manager, t);
      return(NULL);
    }
  }
  cuddRef(e);

  if (topcube == top) {       /* abstract */
    r = cuddBddAndRecur(manager, Cudd_Not(t), Cudd_Not(e));
    if (r == NULL) {
      Cudd_RecursiveDeref(manager, t);
      Cudd_RecursiveDeref(manager, e);
      return(NULL);
    }
    r = Cudd_Not(r);
    cuddRef(r);
    Cudd_RecursiveDeref(manager, t);
    Cudd_RecursiveDeref(manager, e);
    cuddDeref(r);
  } else
    if (t == e) {
      r = t;
      cuddDeref(t);
      cuddDeref(e);
    } else {
      if (Cudd_IsComplement(t)) {
         r = cuddUniqueInter(manager,(int)index,Cudd_Not(t),Cudd_Not(e));
         if (r == NULL) {
           Cudd_RecursiveDeref(manager, t);
           Cudd_RecursiveDeref(manager, e);
           return(NULL);
         }
         r = Cudd_Not(r);
      } else {
         r = cuddUniqueInter(manager,(int)index,t,e);
         if (r == NULL) {
           Cudd_RecursiveDeref(manager, t);
           Cudd_RecursiveDeref(manager, e);
           return(NULL);
         }
      }
      cuddDeref(e);
      cuddDeref(t);
    }

  if (fg_swap)
    cuddCacheInsert(manager, DD_BDD_AND_ABSTRACT_TAG, g, f, cube, r);
  else
    cuddCacheInsert(manager, DD_BDD_AND_ABSTRACT_TAG, f, g, cube, r);
  info_item_ptr->size_cost += (MANAGER_KEYS(manager,profile) - keys0);

#if 1
  info_item_ptr->node_visit_cnt++;
#else
  if (r!=zero) {
    info_item_ptr->node_visit_cnt++;
  }
#endif
  
  return (r);
}

/**Function********************************************************************

  Synopsis    [Recursive step of BDD pruning based on profiles.]

  Description [Several heuristics are allowed to bias the pruning/clipping
    process:
    <ul>
    <li>1: recursions
    <li>2: recursions on active nodes,
    <li>3: size
    <li>4: normalized sizelight (prune lighter branch, recur on heavier one)
    <li>5: normalized sizeheavy (prune heavier branch, recur on lighter),
    <li>6: sizelight (prune lighter branch, recur on heavier one),
    <li>7: sizeheavy (prune heavier branch, recur on lighter), ]
    <li>8: sizelight no rec. (prune lighter branch, don't recur on
    heavier one)
    <li>9: sizeheavy no rec. (prune heavier branch,
    don't recur on lighter).
    <\ul>
    ]

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

static DdNode *
cuplusPruneProfiledRecur(
  DdManager * manager,
  DdNode * f,
  cuplus_profile_info_t *profile,
  Cuplus_PruneHeuristic_e prune_heuristic,
  int threshold,
  int upper_activity
)
{
  DdNode *F, *T, *E, *res, *res1, *res2, *one, *zero;
  struct profile_info_item *info_item_ptr, *info_item_ptr_t, *info_item_ptr_e;
  int do_size_pruning = 0;

  st_table  *table=profile->table;

  one = DD_ONE (manager);
  zero = Cudd_Not (one);

  if (f == zero || f == one) {
    return (f);
  }

  if (prune_heuristic == Cuplus_None)
    return (f);

  F = Cudd_Regular(f);
  if (!st_lookup(table, (char *)f, (char **)&info_item_ptr)) {
    /* not profiled: return zero if only active recursions required */
    if (prune_heuristic == Cuplus_ActiveRecursions)
      return zero;
    else
      return f;
  }

  /* From now on, the f top node is profiled */

  /*
   * Check if already computed: no cache table used,
   * caching done in profile info.
   */

  if (info_item_ptr->pruned_bdd != NULL)
    return (info_item_ptr->pruned_bdd);

  /* not cached: check for pruning */

  upper_activity += info_item_ptr->profile_subtree;
  upper_activity += info_item_ptr->cache_hits;

      /* full subtree profiled: decrease threshold for recursions */

  switch (prune_heuristic) {

    case Cuplus_RecursionsWithSharing:

#if 0
  if (prune_heuristic == Cuplus_RecursionsWithSharing) {
    if ((info_item_ptr->node_visit_cnt+upper_activity)*100 <=
      threshold /* *info_item_ptr->sharing*/
      /*(profile->size_by_var[F->index])*/ ) {
#if 1
          info_item_ptr->pruned_bdd = zero;
#endif
          return zero;
    }
  }
#endif

    case Cuplus_Recursions:
    case Cuplus_ActiveRecursions:
      if ((info_item_ptr->node_visit_cnt+upper_activity) <= threshold)
        return zero;
      if (info_item_ptr->profile_subtree > 0) {
        /* whole BDD profiled - threshold already checked - NO recursion ! */
        return f;
      }
      break;
    case Cuplus_NormSizePrune:
      if (info_item_ptr->norm_size_cost > threshold)
        return (zero);
      break;
    case Cuplus_NormSizePruneLight:
    case Cuplus_NormSizePruneHeavy:
      if (profile->dac99_compatible) {
        if (info_item_ptr->node_visit_cnt > 0)
          if (info_item_ptr->norm_size_cost > threshold)
            do_size_pruning = 1;
      }
      else
        if (info_item_ptr->norm_size_cost > threshold)
          do_size_pruning = 1;
      break;
    case Cuplus_SizePrune:
      if (info_item_ptr->norm_size_cost > threshold)
        return (zero);
      break;
    case Cuplus_SizePruneLight:
    case Cuplus_SizePruneHeavy:
    case Cuplus_SizePruneLightNoRecur:
    case Cuplus_SizePruneHeavyNoRecur:
      if (profile->dac99_compatible) {
        if (info_item_ptr->node_visit_cnt > 0)
          if (info_item_ptr->size_cost > threshold)
            do_size_pruning = 1;
      }
      else
        if (info_item_ptr->size_cost > threshold)
          do_size_pruning = 1;
      break;
    default:
      Pdtutil_Assert(1,"Wrong switch in Profile Pruning");
  }

  if (profile->dac99_compatible) {
    if (info_item_ptr->profile_subtree > 0) {
      /* whole BDD profiled - threshold already checked - NO more ! */
      return f;
    }
  }

  /* Compute the cofactors of f and recur */

  T = cuddT(F); E = cuddE(F);
  if (f != F) {
    T = Cudd_Not(T); E = Cudd_Not(E);
  }

  if (((profile->dac99_compatible)
    && (do_size_pruning) && (T != one) && (E != one)) ||
     ((!profile->dac99_compatible)
    && (do_size_pruning)&&(!Cudd_IsConstant(T))&&(!Cudd_IsConstant(E)))) {

    if ((st_lookup(table, (char *)T, (char **)&info_item_ptr_t)) &&
        (st_lookup(table, (char *)E, (char **)&info_item_ptr_e))) {

      switch (prune_heuristic) {
        case Cuplus_NormSizePruneLight:
          if (info_item_ptr_t->norm_size_cost >
              info_item_ptr_e->norm_size_cost)
            E = zero;
          else
            T = zero;
          break;
        case Cuplus_NormSizePruneHeavy:
          if (info_item_ptr_t->norm_size_cost >
              info_item_ptr_e->norm_size_cost)
            T = zero;
          else
            E = zero;
          break;
        case Cuplus_SizePruneLight:
          if (info_item_ptr_t->size_cost > info_item_ptr_e->size_cost)
            E = zero;
          else
            T = zero;
          break;
        case Cuplus_SizePruneLightNoRecur:
          if (info_item_ptr_t->size_cost > info_item_ptr_e->size_cost)
            E = zero;
          else
            T = zero;
          prune_heuristic = Cuplus_None;
          break;
        case Cuplus_SizePruneHeavy:
          if (info_item_ptr_t->size_cost > info_item_ptr_e->size_cost)
            T = zero;
          else
            E = zero;
          break;
        case Cuplus_SizePruneHeavyNoRecur:
          if (info_item_ptr_t->size_cost > info_item_ptr_e->size_cost)
            T = zero;
          else
            E = zero;
          prune_heuristic = Cuplus_None;
          break;
        default:
          Pdtutil_Assert(1,"Wrong switch in Profile Pruning");
      }

    }
  }

  /* recur on cofactors */

  res1 = cuplusPruneProfiledRecur(manager, T, profile,
    prune_heuristic, threshold, upper_activity);
  if (res1 == NULL) return(NULL);
  cuddRef(res1);

  res2 = cuplusPruneProfiledRecur(manager, E, profile,
    prune_heuristic, threshold, upper_activity);
  if (res2 == NULL) {
    Cudd_IterDerefBdd(manager, res1);
    return(NULL);
  }
  cuddRef(res2);

  /* ITE takes care of possible complementation of res1 */

  res = cuddBddIteRecur(manager,Cudd_bddIthVar(manager,F->index),res1,res2);
  if (res == NULL) {
    Cudd_IterDerefBdd(manager, res1);
    Cudd_IterDerefBdd(manager, res2);
    return(NULL);
  }
  cuddDeref(res1);
  cuddDeref(res2);
  if (F->ref != 1) {
    info_item_ptr->pruned_bdd = res;
  }

  return (res);
}

/**Function********************************************************************

  Synopsis [DFV visit with statistics upgrading.]

  Description [This is rather a template for BDD visit. Only BDD size by
    variable index is collected so far.]

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

static void
cuplusCollectProfileSummaryRecur(
  DdManager * manager,
  DdNode * f,
  cuplus_profile_info_t *profile
)
{
  struct profile_info_item *info_item_ptr;
  DdNode *F;
  st_table * table=profile->table;

  F = Cudd_Regular(f);

  if (Cudd_IsComplement(F->next)) {
    return;
  }
  F->next = Cudd_Not(F->next);
  if (cuddIsConstant(F)) {
    return;
  }

  if (!st_lookup(table, (char *)F, (char **)&info_item_ptr)) {
    /* not profiled: return */
    return;
  }

#if 0
  profile->size_by_var[F->index]++;
#endif

  cuplusCollectProfileSummaryRecur(manager,cuddT(F),profile);
  cuplusCollectProfileSummaryRecur(manager,Cudd_Regular(cuddE(F)),
    profile);

  return;
}

/**Function********************************************************************

  Synopsis [Undocumented]

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

static void
BfvProfiles(
  DdManager * manager,
  cuplus_profile_info_t *profile
)
{
  struct profile_info_item *info_item_ptr, *info_item_ptr_t, *info_item_ptr_e;
  struct profile_info_item **info_items_by_var;
  DdNode *f, *F, *t, *e;
  int i, j, nvars;
  st_table * table=profile->table;

  nvars = Cudd_ReadSize(manager);

  if ((info_items_by_var = ALLOC(struct profile_info_item *, nvars)) == NULL) {
    fprintf (stderr, "out of memory visiting profiles\n");
    return;
  }

  for (i=0; i<nvars; i++)
    info_items_by_var[i] = NULL;

  st_foreach(table, ProfileInfoByVars, (char *)info_items_by_var);

  for (i=0; i<nvars; i++) {
    j = Cudd_ReadInvPerm(manager,i);
    for (info_item_ptr = info_items_by_var[j];
         info_item_ptr != NULL;
         info_item_ptr = info_item_ptr->next) {

      /* visit items */

      f = info_item_ptr->ref_node;
      F = Cudd_Regular(f);

      t = cuddT(F);
      e = cuddE(F);
      if (Cudd_IsComplement(f)) {
        t = Cudd_Not(t);
        e = Cudd_Not(e);
      }

      info_item_ptr->tot_act += info_item_ptr->profile_subtree;
      info_item_ptr->tot_act += info_item_ptr->cache_hits;

      if (info_item_ptr->sharing == 0)
        info_item_ptr->sharing = 1;

      if (st_lookup(table, (char *)t, (char **)&info_item_ptr_t)) {
        if (info_item_ptr_t->tot_act < info_item_ptr->tot_act)
          info_item_ptr_t->tot_act = info_item_ptr->tot_act;
        info_item_ptr_t->sharing++;
      }
      if (st_lookup(table, (char *)e, (char **)&info_item_ptr_e)) {
        if (info_item_ptr_e->tot_act < info_item_ptr->tot_act)
          info_item_ptr_e->tot_act = info_item_ptr->tot_act;
        info_item_ptr_e->sharing++;
      }

      /* end visit */
    }
  }

  return;

}

/**Function********************************************************************

  Synopsis    [Undocumented]

  Description []

  SideEffects [None]

******************************************************************************/

static enum st_retval
ProfileInfoByVars(
  char * key,
  char * value,
  char * arg
)
{
  struct profile_info_item *info_item_ptr, **info_items_by_var;
  DdNode *f;
  info_items_by_var = (struct profile_info_item **)arg;
  info_item_ptr = (struct profile_info_item *)value;

  f = Cudd_Regular(info_item_ptr->ref_node);
  info_item_ptr->next = info_items_by_var[f->index];
  info_items_by_var[f->index] = info_item_ptr;

  info_item_ptr->tot_act = 0;
  info_item_ptr->sharing = 0;

  return(ST_CONTINUE);
}

/**Function********************************************************************

  Synopsis    [Performs a DFS from f, clearing the LSB of the next
    pointers.]

  Description []

  SideEffects [None]

  SeeAlso     [ddSupportStep ddDagInt]

******************************************************************************/

static void
ddClearFlag(
  DdNode * f)
{
    if (!Cudd_IsComplement(f->next)) {
        return;
    }
    /* Clear visited flag. */
    f->next = Cudd_Regular(f->next);
    if (cuddIsConstant(f)) {
        return;
    }
    ddClearFlag(cuddT(f));
    ddClearFlag(Cudd_Regular(cuddE(f)));
    return;

}





