/**CHeaderFile*****************************************************************

  FileName    [ddi.h]

  PackageName [ddi]

  Synopsis    [Portability interface with the Decision Diagram package.]

  Description [The <b>DDI</b> package provides functions to manipulate the 
    following objects:
    <ul>
    <li>boolean functions (Ddi_Dd_t)
    <li>arrays of boolean functions (Ddi_DdArray_t)
    <li>variables (Ddi_Var_t)
    <li>arrays of variables (Ddi_VarArray_t)
    <li>sets of variables (Ddi_VarSet_t).
    </ul>
    It can be viewed both as a portability interface, concentrating
    all dependancies from the BDD package, and as an upper level,
    manipulating the previously listed BDD based data types.
    A particular feature, distinguishing this interface from other ones,
    is the support for partitioned (conjunctive as well as disjunctive)
    functions.
    This implementation is written on top of the CUDD package.
    ]
                  
  SeeAlso     []  

  Author    [Gianpiero Cabodi and Stefano Quer]

  Copyright [This file was created at the Politecnico di Torino,
    Torino, Italy. 
    The  Politecnico di Torino makes no warranty about the suitability of 
    this software for any purpose.  
    It is presented on an AS IS basis.]
  
  Revision    []

******************************************************************************/

#ifndef _DDI
#define _DDI

/*---------------------------------------------------------------------------*/
/* Nested includes                                                           */
/*---------------------------------------------------------------------------*/

#include "pdtutil.h"
#include "cuplus.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*
 *  Codes for generalized Boolean functions.
 */

#define DDI_MONO_BOOLF                    0
#define DDI_CPART_BOOLF                   1
#define DDI_DPART_BOOLF                   2

/*
 *  Default settings.
 */

#define DDI_NO_LIMIT                      ((1<<30)-2)
#define DDI_DFLT_ITE_ON                   TRUE
#define DDI_DFLT_ITE_RESIZE_AT            75
#define DDI_DFLT_ITE_MAX_SIZE             1000000
#define DDI_DFLT_ITE_CONST_ON             TRUE
#define DDI_DFLT_ITE_CONST_RESIZE_AT      75
#define DDI_DFLT_ITE_CONST_MAX_SIZE       1000000
#define DDI_DFLT_ADHOC_ON                 TRUE
#define DDI_DFLT_ADHOC_RESIZE_AT          0
#define DDI_DFLT_ADHOC_MAX_SIZE           10000000
#define DDI_DFLT_GARB_COLLECT_ON          TRUE
#define DDI_DFLT_DAEMON                   NIL(void)
#define DDI_DFLT_MEMORY_LIMIT             BDD_NO_LIMIT
#define DDI_DFLT_NODE_RATIO               2.0
#define DDI_DFLT_INIT_BLOCKS              10

#define DDI_UNIQUE_SLOTS CUDD_UNIQUE_SLOTS
#define DDI_CACHE_SLOTS CUDD_CACHE_SLOTS
#define DDI_MAX_CACHE_SIZE 0

/*
 *  Dynamic reordering.
 */

#define DDI_REORDER_SAME 0
#define DDI_REORDER_NONE 1
#define DDI_REORDER_RANDOM 2
#define DDI_REORDER_RANDOM_PIVOT 3
#define DDI_REORDER_SIFT 4
#define DDI_REORDER_SIFT_CONVERGE 5
#define DDI_REORDER_SYMM_SIFT 6
#define DDI_REORDER_SYMM_SIFT_CONV 7
#define DDI_REORDER_WINDOW2 8
#define DDI_REORDER_WINDOW3 9
#define DDI_REORDER_WINDOW4 10
#define DDI_REORDER_WINDOW2_CONV 11
#define DDI_REORDER_WINDOW3_CONV 12
#define DDI_REORDER_WINDOW4_CONV 13
#define DDI_REORDER_GROUP_SIFT 14
#define DDI_REORDER_GROUP_SIFT_CONV 15
#define DDI_REORDER_ANNEALING 16
#define DDI_REORDER_GENETIC 17
#define DDI_REORDER_LINEAR 18
#define DDI_REORDER_LINEAR_CONVERGE 19
#define DDI_REORDER_EXACT 20

/*---------------------------------------------------------------------------*/
/* Structure declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/**Enum************************************************************************

  Synopsis    [Type for Dense Subsetting.]

  Description []

******************************************************************************/

typedef enum
  {
  Ddi_DenseNone_c,
  Ddi_DenseSubHeavyBranch_c,
  Ddi_DenseSupHeavyBranch_c,
  Ddi_DenseSubShortPath_c,
  Ddi_DenseSupShortPath,
  Ddi_DenseUnderApprox_c,
  Ddi_DenseOverApprox_c,
  Ddi_DenseRemapUnder_c,
  Ddi_DenseRemapOver_c,
  Ddi_DenseSubCompress_c,
  Ddi_DenseSupCompress_c
  }
Ddi_DenseMethod_e;

typedef struct Ddi_DdMgr Ddi_DdMgr;

typedef struct DdNode Ddi_DdNode;

typedef struct Ddi_Dd_t Ddi_Dd_t;

typedef struct Ddi_DdArray_t Ddi_DdArray_t;

typedef Ddi_DdNode Ddi_Var_t;

typedef Ddi_Dd_t Ddi_VarSet_t;

typedef struct Ddi_VarArray_t Ddi_VarArray_t;

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

#define DDI_MAX(a,b)         ((a>b)? a:b)

/* hide dependancy on the CUDD package */

#define ddiP(F) cuddP((Ddi_DdReadMgr(F))->mgrCU,Ddi_DdToCU(F))
#define Ddi_MgrAutodynDisable(dd) Cudd_AutodynDisable(Ddi_MgrReadMgrCU(dd))
#define Ddi_MgrAutodynEnable(dd,n) Cudd_AutodynEnable(Ddi_MgrReadMgrCU(dd),n)
#define Ddi_ReduceHeap(dd,m,n) Cudd_ReduceHeap(Ddi_MgrReadMgrCU(dd),m,n)
#define Ddi_MakeTreeNode(dd,i,j,mode) Cudd_MakeTreeNode(Ddi_MgrReadMgrCU(dd),i,j,mode)
#define Ddi_DdIncluded(f,g)\
 (Cudd_bddIteConstant(Ddi_MgrReadMgrCU(Ddi_DdReadMgr(f)),\
                      Ddi_DdToCU(f),Ddi_DdToCU(g),\
                      Ddi_DdToCU(Ddi_MgrReadOne(Ddi_DdReadMgr(f)))\
                     )\
                   == Ddi_DdToCU(Ddi_MgrReadOne(Ddi_DdReadMgr(f)))\
 )
#define Ddi_CountMinterm(F,n) \
 (Cudd_CountMinterm(Ddi_MgrReadMgrCU(Ddi_DdReadMgr(F)),Ddi_DdToCU(F),n))
#define Ddi_CheckZeroRef(dd) Cudd_CheckZeroRef(Ddi_MgrReadMgrCU(dd))
#define Ddi_ReadReorderingStatus(dd,pmethod) \
                       Cudd_ReorderingStatus(Ddi_MgrReadMgrCU(dd),\
                        ((Cudd_ReorderingType *)pmethod))
#define Ddi_ReadPerm(dd,i) Cudd_ReadPerm(Ddi_MgrReadMgrCU(dd),i)
#define Ddi_ReadInvPerm(dd,i) Cudd_ReadInvPerm(Ddi_MgrReadMgrCU(dd),i)
#define Ddi_ReadSize(dd) Cudd_ReadSize(Ddi_MgrReadMgrCU(dd))
#define Ddi_ReadNumVars(dd) (Cudd_ReadSize(Ddi_MgrReadMgrCU(dd)))
#define Ddi_ReadSlots(dd) Cudd_ReadSlots(Ddi_MgrReadMgrCU(dd))
#define Ddi_ReadKeys(dd) Cudd_ReadKeys(Ddi_MgrReadMgrCU(dd))
#define Ddi_ReadMinDead(dd) Cudd_ReadMinDead(Ddi_MgrReadMgrCU(dd))
#define Ddi_ReadMemoryInUse(dd) Cudd_ReadMemoryInUse(Ddi_MgrReadMgrCU(dd))
#define Ddi_ReadReorderings(dd) Cudd_ReadReorderings(Ddi_MgrReadMgrCU(dd))
#define Ddi_ReadReorderingTime(dd) Cudd_ReadReorderingTime(Ddi_MgrReadMgrCU(dd))
#define Ddi_ReadGarbageCollections(dd) Cudd_ReadGarbageCollections(Ddi_MgrReadMgrCU(dd))
#define Ddi_ReadGarbageCollectionTime(dd) Cudd_ReadGarbageCollectionTime(Ddi_MgrReadMgrCU(dd))

#define Ddi_DdEval(dd,f,inputs) Cudd_Eval(Ddi_MgrReadMgrCU(dd),Ddi_DdToCU(f),inputs)
#define Ddi_DdNodeIsConstant(F) Cudd_IsConstant(Ddi_DdNodeToCU(F))
#define Ddi_DdIsConstant(F) Cudd_IsConstant(Ddi_DdToCU(F))
#define Ddi_DdIsComplement(F) Cudd_IsComplement(Ddi_DdToCU(F))

#define Ddi_DdSetMaxCacheHard(dd,n) \
 Cudd_SetMaxCacheHard(Ddi_MgrReadMgrCU(dd),n)
#define Ddi_DdSetMaxMemory(dd,n) Cudd_SetMaxMemory(Ddi_MgrReadMgrCU(dd),n)
#define Ddi_DiSetNextReordering(dd,n) \
  Cudd_SetNextReordering(Ddi_MgrReadMgrCU(dd),n)

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Function prototypes                                                       */
/*---------------------------------------------------------------------------*/

EXTERN Ddi_Dd_t * Ddi_DdDenseSet(Ddi_DenseMethod_e method, Ddi_Dd_t *f, int threshold, int safe, int quality, double hardlimit);
EXTERN int Ddi_DdReadCode(Ddi_Dd_t * f);
EXTERN void Ddi_DdSetCode(Ddi_Dd_t * f, int code);
EXTERN Ddi_DdMgr * Ddi_DdReadMgr(Ddi_Dd_t * f);
EXTERN void Ddi_DdSetMgr(Ddi_Dd_t * f, Ddi_DdMgr * dd);
EXTERN Ddi_DdArray_t * Ddi_DdReadPart(Ddi_Dd_t * f);
EXTERN void Ddi_DdSetPart(Ddi_Dd_t * f, Ddi_DdArray_t *p);
EXTERN int Ddi_DdReadPartNum(Ddi_Dd_t * f);
EXTERN Ddi_Dd_t * Ddi_DdReadIthPart(Ddi_Dd_t * f, int i);
EXTERN Ddi_Dd_t * Ddi_DdCreateFromCU(Ddi_DdMgr *mgr, DdNode * f);
EXTERN void Ddi_DdPrint(Ddi_Dd_t *f);
EXTERN void Ddi_DdSupportPrint(Ddi_Dd_t *f);
EXTERN int Ddi_DdSize(Ddi_Dd_t *f);
EXTERN void Ddi_DdPrintStats(Ddi_Dd_t *f);
EXTERN void Ddi_DdFree(Ddi_Dd_t *f);
EXTERN Ddi_Dd_t * Ddi_DdDup(Ddi_Dd_t *f);
EXTERN Ddi_Dd_t * Ddi_DdCopy(Ddi_DdMgr *dd2, Ddi_Dd_t *f);
EXTERN Ddi_Dd_t * Ddi_DdCopyWithAuxMgr(Ddi_DdMgr *dd2, Ddi_Dd_t *f, Ddi_DdMgr *ddAux);
EXTERN Ddi_Dd_t * Ddi_DdSwapVariablesWithAuxMgr(Ddi_Dd_t *f, Ddi_VarArray_t *src, Ddi_VarArray_t *dst, Ddi_DdMgr *ddAux);
EXTERN Ddi_Dd_t * Ddi_DdEvalFree(Ddi_Dd_t *f, Ddi_Dd_t *g);
EXTERN int Ddi_DdTopvarIndex(Ddi_Dd_t *f);
EXTERN Ddi_Dd_t * Ddi_DdLiteral(Ddi_DdMgr *dd, Ddi_Var_t *v);
EXTERN Ddi_Dd_t * Ddi_DdComplLiteral(Ddi_DdMgr *dd, Ddi_Var_t *v);
EXTERN int Ddi_DdEqual(Ddi_Dd_t *f1, Ddi_Dd_t *f2);
EXTERN int Ddi_DdIsZero(Ddi_Dd_t *f);
EXTERN int Ddi_DdIsOne(Ddi_Dd_t *f);
EXTERN Ddi_Dd_t * Ddi_DdNot(Ddi_Dd_t *f);
EXTERN Ddi_Dd_t * Ddi_DdAnd(Ddi_Dd_t *f, Ddi_Dd_t *g);
EXTERN Ddi_Dd_t * Ddi_DdOr(Ddi_Dd_t *f, Ddi_Dd_t *g);
EXTERN Ddi_Dd_t * Ddi_DdXor(Ddi_Dd_t *f, Ddi_Dd_t *g);
EXTERN Ddi_Dd_t * Ddi_DdXnor(Ddi_Dd_t *f, Ddi_Dd_t *g);
EXTERN Ddi_Dd_t * Ddi_DdIte(Ddi_Dd_t *f, Ddi_Dd_t *g, Ddi_Dd_t *h);
EXTERN Ddi_Dd_t * Ddi_DdAndExist(Ddi_Dd_t *f, Ddi_Dd_t *g, Ddi_VarSet_t *supp);
EXTERN Ddi_Dd_t * Ddi_DdExist(Ddi_Dd_t *f, Ddi_VarSet_t *supp);
EXTERN Ddi_Dd_t * Ddi_DdConstrain(Ddi_Dd_t *f, Ddi_Dd_t *c);
EXTERN Ddi_Dd_t * Ddi_DdRestrict(Ddi_Dd_t *f, Ddi_Dd_t *c);
EXTERN Ddi_Dd_t * Ddi_DdSwapVariables(Ddi_Dd_t *f, Ddi_VarArray_t *x, Ddi_VarArray_t *y);
EXTERN Ddi_VarSet_t * Ddi_DdSupport(Ddi_Dd_t *f);
EXTERN int Ddi_DdStore(Ddi_Dd_t *f, char *ddname, char mode, char *filename, FILE *fp);
EXTERN Ddi_Dd_t * Ddi_DdLoad(Ddi_DdMgr *dd, int varmatchmode, char mode, char *filename, FILE *fp);
EXTERN Ddi_Dd_t * Ddi_DdAddCpart(Ddi_Dd_t *f, Ddi_Dd_t *new);
EXTERN Ddi_Dd_t * Ddi_DdAddDpart(Ddi_Dd_t *f, Ddi_Dd_t *new);
EXTERN void Ddi_DdReplaceIthPart(Ddi_Dd_t *f, int i, Ddi_Dd_t *p);
EXTERN Ddi_Dd_t * Ddi_DdRemPart(Ddi_Dd_t *f, int pi);
EXTERN Ddi_Dd_t * Ddi_DdPart2Mono(Ddi_Dd_t *f);
EXTERN Ddi_Dd_t * Ddi_DdPartFromArray(Ddi_DdArray_t *array, int code);
EXTERN void Ddi_ProfileInfoInit(Ddi_Dd_t *f, cuplus_profile_info_t *profileInfo, unsigned char dac99_compatible);
EXTERN void Ddi_ProfileInfoFree(Ddi_Dd_t *f);
EXTERN void Ddi_ProfileInfoPrint(Ddi_Dd_t *f);
EXTERN Ddi_Dd_t * Ddi_PruneProfiled(Ddi_Dd_t *f, Cuplus_PruneHeuristic_e prune_heuristic, int threshold);
EXTERN void Ddi_MulvarInterleave(Ddi_Dd_t *f, Ddi_Dd_t *g);
EXTERN Ddi_DdNode * Ddi_DdToCU(Ddi_Dd_t *f);
EXTERN Ddi_DdArray_t * Ddi_DdArrayAlloc(int length);
EXTERN void Ddi_DdArrayFree(Ddi_DdArray_t *array);
EXTERN int Ddi_DdArrayInsert(Ddi_DdArray_t *array, int idx, Ddi_Dd_t *f);
EXTERN int Ddi_DdArrayInsertLast(Ddi_DdArray_t *array, Ddi_Dd_t *f);
EXTERN int Ddi_DdArrayInclude(Ddi_DdArray_t *array, int idx, Ddi_Dd_t *f);
EXTERN Ddi_Dd_t * Ddi_DdArrayFetch(Ddi_DdArray_t *array, int i);
EXTERN Ddi_Dd_t * Ddi_DdArrayExtract(Ddi_DdArray_t *array, int i);
EXTERN int Ddi_DdArray_n(Ddi_DdArray_t *array);
EXTERN Ddi_Dd_t * Ddi_DdArrayFetchLast(Ddi_DdArray_t *array);
EXTERN Ddi_DdArray_t * Ddi_DdArrayDup(Ddi_DdArray_t *old);
EXTERN Ddi_DdArray_t * Ddi_DdArrayCopy(Ddi_DdMgr *dd2, Ddi_DdArray_t *F);
EXTERN int Ddi_DdArrayAppend(Ddi_DdArray_t *array1, Ddi_DdArray_t *array2);
EXTERN Ddi_DdArray_t * Ddi_DdArrayJoin(Ddi_DdArray_t *array1, Ddi_DdArray_t *array2);
EXTERN Ddi_Dd_t ** Ddi_DdArrayReadData(Ddi_DdArray_t * array);
EXTERN DdNode ** Ddi_DdArrayToCU(Ddi_DdArray_t *array);
EXTERN Ddi_DdArray_t * Ddi_DdArrayFromCU(Ddi_DdMgr *dd, DdNode **array, int n);
EXTERN void Ddi_DdArrayConstrainIthToN(Ddi_DdArray_t *array, Ddi_Dd_t *C, int i0, int n);
EXTERN void Ddi_DdArrayConstrain(Ddi_DdArray_t *array, Ddi_Dd_t *C);
EXTERN void Ddi_DdArrayRestrictIthToN(Ddi_DdArray_t *array, Ddi_Dd_t *C, int i0, int n);
EXTERN void Ddi_DdArrayRestrict(Ddi_DdArray_t *array, Ddi_Dd_t *C);
EXTERN void Ddi_DdArrayPrint(Ddi_DdArray_t *array);
EXTERN int Ddi_DdArraySize(Ddi_DdArray_t *array);
EXTERN void Ddi_DdArrayPrintStats(Ddi_DdArray_t *array);
EXTERN void Ddi_DdArrayAndIthToN(Ddi_DdArray_t *array, Ddi_Dd_t *c, int i0, int n);
EXTERN void Ddi_DdArrayAnd(Ddi_DdArray_t *array, Ddi_Dd_t *C);
EXTERN void Ddi_DdArrayExistIthToN(Ddi_DdArray_t *array, Ddi_VarSet_t *V, int i0, int n);
EXTERN void Ddi_DdArrayExist(Ddi_DdArray_t *array, Ddi_VarSet_t *V);
EXTERN Ddi_DdArray_t * Ddi_DdArrayOrDdArray(Ddi_DdArray_t *array1, Ddi_DdArray_t *array2);
EXTERN int Ddi_DdArrayStore(Ddi_DdArray_t *array, char *ddname, char **vnames, char **rnames, int *auxids, int mode, char *fname, FILE *fp);
EXTERN Ddi_DdArray_t * Ddi_DdArrayLoad(Ddi_DdMgr *dd, char **vnames, int *vauxids, int mode, char *file, FILE *fp);
EXTERN Ddi_VarSet_t * Ddi_DdArraySupportIthToN(Ddi_DdArray_t *array, int i0);
EXTERN Ddi_VarSet_t * Ddi_DdArraySupport(Ddi_DdArray_t *array);
EXTERN Ddi_VarSet_t ** Ddi_DdArraySupportArray(Ddi_DdArray_t *Fa);
EXTERN Ddi_DdNode * Ddi_DdRootNode(Ddi_Dd_t *f);
EXTERN int Ddi_DdNodeReadIndex(Ddi_DdNode *f);
EXTERN Ddi_DdNode * Ddi_DdNodeThen(Ddi_DdNode *f);
EXTERN Ddi_DdNode * Ddi_DdNodeElse(Ddi_DdNode *f);
EXTERN int Ddi_DdNodeIsVisited(DdNode *f);
EXTERN void Ddi_DdNodeSetVisited(DdNode *f);
EXTERN void Ddi_DdNodeClearVisited(DdNode *f);
EXTERN DdNode * Ddi_DdNodeToCU(Ddi_DdNode *n);
EXTERN Ddi_DdNode * Ddi_DdNodeFromCU(DdNode *n);
EXTERN int Ddi_MgrOperation(Ddi_DdMgr **ddMgrP, char *string, Pdtutil_MgrOp_t operationFlag, void **voidPointer, Pdtutil_MgrRet_t *returnFlagP);
EXTERN void Ddi_MgrPrintStats(Ddi_DdMgr *dd);
EXTERN Ddi_DdMgr * Ddi_MgrInit(char *ddiName, DdManager *CUMgr, unsigned int nvar, unsigned int numSlots, unsigned int cacheSize, unsigned long memorySizeMax);
EXTERN void Ddi_DdiMgrQuit(Ddi_DdMgr *dd);
EXTERN Ddi_DdMgr * Ddi_MgrDup(Ddi_DdMgr *dd);
EXTERN void Ddi_MgrShuffle(Ddi_DdMgr *dd, int *sortedIds, int nids);
EXTERN void Ddi_MgrAlign(Ddi_DdMgr *dd, Ddi_DdMgr *ddRef);
EXTERN void Ddi_MgrCreateGroups2(Ddi_DdMgr *dd, Ddi_VarArray_t *vfix, Ddi_VarArray_t *vmov);
EXTERN int Ddi_MgrWriteOrd(Ddi_DdMgr *dd, char *filename, FILE *fp, Pdtutil_VariableOrderFormat_e fileFormat);
EXTERN int Ddi_MgrReadOrdNamesAuxids(Ddi_DdMgr *dd, char *filename, FILE *fp);
EXTERN void Ddi_MgrAutodynSuspend(Ddi_DdMgr *dd);
EXTERN void Ddi_MgrAutodynResume(Ddi_DdMgr *dd);
EXTERN DdManager * Ddi_MgrReadMgrCU(Ddi_DdMgr *dd);
EXTERN Ddi_Dd_t * Ddi_MgrReadOne(Ddi_DdMgr *dd);
EXTERN Ddi_Dd_t * Ddi_MgrReadZero(Ddi_DdMgr *dd);
EXTERN char** Ddi_MgrReadVarnames(Ddi_DdMgr *dd);
EXTERN int * Ddi_MgrReadVarauxids(Ddi_DdMgr *dd);
EXTERN int Ddi_MgrReadDdNum(Ddi_DdMgr *dd);
EXTERN unsigned int Ddi_MgrReadDynordThresh(Ddi_DdMgr *dd);
EXTERN char * Ddi_ReadDdiName(Ddi_DdMgr *dd);
EXTERN void Ddi_SetDdiName(Ddi_DdMgr *dd, char *ddiName);
EXTERN Pdtutil_VerbLevel_e Ddi_MgrReadVerbosity(Ddi_DdMgr *ddiMgr);
EXTERN void Ddi_MgrSetVerbosity(Ddi_DdMgr *ddiMgr, Pdtutil_VerbLevel_e verbosity);
EXTERN void Ddi_MgrSetMgrCU(Ddi_DdMgr *dd, DdManager *m);
EXTERN void Ddi_MgrSetOne(Ddi_DdMgr *dd, Ddi_Dd_t *one);
EXTERN void Ddi_MgrSetZero(Ddi_DdMgr *dd, Ddi_Dd_t *zero);
EXTERN void Ddi_MgrSetVarnames(Ddi_DdMgr *dd, char **vn);
EXTERN void Ddi_MgrSetVarauxids(Ddi_DdMgr *dd, int *va);
EXTERN void Ddi_MgrSetDynordThresh(Ddi_DdMgr *dd, unsigned int th);
EXTERN unsigned int Ddi_ReadCacheSlots(Ddi_DdMgr *dd);
EXTERN double Ddi_ReadCacheLookUps(Ddi_DdMgr *dd);
EXTERN double Ddi_ReadCacheHits(Ddi_DdMgr *dd);
EXTERN unsigned int Ddi_ReadMinHit(Ddi_DdMgr *dd);
EXTERN unsigned int Ddi_ReadMaxCacheHard(Ddi_DdMgr *dd);
EXTERN unsigned int Ddi_ReadMaxCache(Ddi_DdMgr* dd);
EXTERN int Ddi_BddOperation(Ddi_DdMgr *defaultDdMgr, Ddi_Dd_t **bddP, char *string, Pdtutil_MgrOp_t operationFlag, void **voidPointer, Pdtutil_MgrRet_t *returnFlag);
EXTERN int Ddi_BddArrayOperation(Ddi_DdMgr *defaultDdMgr, Ddi_DdArray_t **bddArrayP, char *string, Pdtutil_MgrOp_t operationFlag, void **voidPointer, Pdtutil_MgrRet_t *returnFlag);
EXTERN Ddi_Dd_t * Ddi_ReadCube(Ddi_DdMgr *dd, FILE *fp, int id_or_name);
EXTERN Ddi_VarSet_t * Ddi_ReadVarSet(Ddi_DdMgr *dd, FILE *fp, int id_or_name);
EXTERN Ddi_Dd_t * Ddi_VarSubst(Ddi_Dd_t *f, Ddi_VarArray_t *x, Ddi_VarArray_t *y);
EXTERN void Ddi_PrintVarArray(Ddi_VarArray_t *array);
EXTERN void Ddi_PrintDdArray(Ddi_DdArray_t *array);
EXTERN Cuplus_PruneHeuristic_e Ddi_ProfileHeuristicString2Enum(char *string);
EXTERN char * Ddi_ProfileHeuristicEnum2String(Cuplus_PruneHeuristic_e enumType);
EXTERN Cudd_ReorderingType Ddi_ReorderingMethodString2Enum(char *string);
EXTERN char * Ddi_ReorderingMethodEnum2String(Cudd_ReorderingType enumType);
EXTERN Ddi_DenseMethod_e Ddi_DenseMethodString2Enum(char *string);
EXTERN char * Ddi_DenseMethodEnum2String(Ddi_DenseMethod_e enumType);
EXTERN void Ddi_PrintVersion(FILE *fp);
EXTERN int Ddi_VarIndex(Ddi_Var_t *var);
EXTERN Ddi_Var_t * Ddi_IthVar(Ddi_DdMgr *dd, int index);
EXTERN Ddi_Var_t * Ddi_DdNewVar(Ddi_DdMgr *dd);
EXTERN Ddi_Var_t * Ddi_DdNewVarAtLevel(Ddi_DdMgr *dd, int lev);
EXTERN int Ddi_VarCurrPos(Ddi_DdMgr *dd, Ddi_Var_t *var);
EXTERN char * Ddi_VarId2Name(Ddi_DdMgr *dd, int index);
EXTERN char * Ddi_VarName(Ddi_DdMgr *dd, Ddi_Var_t *var);
EXTERN void Ddi_VarSetName(Ddi_DdMgr *dd, Ddi_Var_t *var, char *name);
EXTERN int Ddi_VarId2Auxid(Ddi_DdMgr *dd, int index);
EXTERN int Ddi_VarAuxid(Ddi_DdMgr *dd, Ddi_Var_t *var);
EXTERN void Ddi_VarSetAuxid(Ddi_DdMgr *dd, Ddi_Var_t *var, int auxid);
EXTERN int Ddi_VarName2Id(Ddi_DdMgr *dd, char *name);
EXTERN int Ddi_VarAuxid2Id(Ddi_DdMgr *dd, int auxid);
EXTERN DdNode * Ddi_VarToCU(Ddi_Var_t * v);
EXTERN Ddi_Var_t * Ddi_VarFromCU(DdNode * v);
EXTERN Ddi_Var_t * Ddi_VarCopy(Ddi_DdMgr *dd2, Ddi_Var_t *v);
EXTERN void Ddi_VarMakeGroup(Ddi_DdMgr *dd, Ddi_Var_t *v, int grpSize);
EXTERN Ddi_VarArray_t * Ddi_VarArrayFromCU(DdNode **array, int n);
EXTERN DdNode ** Ddi_VarArrayToCU(Ddi_VarArray_t *array);
EXTERN Ddi_VarArray_t * Ddi_VarArrayAlloc(int size);
EXTERN void Ddi_VarArrayFree(Ddi_VarArray_t *array);
EXTERN int Ddi_VarArray_n(Ddi_VarArray_t *array);
EXTERN int Ddi_VarArrayInsert(Ddi_VarArray_t *array, int pos, Ddi_Var_t *var);
EXTERN Ddi_Var_t * Ddi_VarArrayFetch(Ddi_VarArray_t *array, int i);
EXTERN Ddi_VarArray_t * Ddi_VarArrayDup(Ddi_VarArray_t *old);
EXTERN Ddi_VarArray_t * Ddi_VarArrayCopy(Ddi_DdMgr *ddAux, Ddi_VarArray_t *vars);
EXTERN int Ddi_VarArrayAppend(Ddi_VarArray_t *array1, Ddi_VarArray_t *array2);
EXTERN Ddi_VarArray_t * Ddi_VarArrayJoin(Ddi_VarArray_t *array1, Ddi_VarArray_t *array2);
EXTERN int * Ddi_VarArrayToInt(Ddi_VarArray_t *array);
EXTERN Ddi_VarSet_t * Ddi_VarArrayToVarSet(Ddi_DdMgr *dd, Ddi_VarArray_t *array);
EXTERN int Ddi_VarArrayIndex(Ddi_VarArray_t *array, int i);
EXTERN Ddi_VarSet_t * Ddi_VarSetVoid(Ddi_DdMgr *dd);
EXTERN void Ddi_VarSetFree(Ddi_VarSet_t *src);
EXTERN int Ddi_VarSetIsVoid(Ddi_VarSet_t *varset);
EXTERN Ddi_VarSet_t * Ddi_VarSetNext(Ddi_VarSet_t *varset);
EXTERN Ddi_VarSet_t * Ddi_VarSetAdd(Ddi_VarSet_t *varset, Ddi_Var_t *var);
EXTERN Ddi_VarSet_t * Ddi_VarSetDup(Ddi_VarSet_t *src);
EXTERN Ddi_VarSet_t * Ddi_VarSetCopy(Ddi_DdMgr *dd2, Ddi_VarSet_t *src);
EXTERN Ddi_VarSet_t * Ddi_VarSetEvalFree(Ddi_VarSet_t *f, Ddi_VarSet_t *g);
EXTERN int Ddi_VarSet_n(Ddi_VarSet_t *vars);
EXTERN void Ddi_VarSetPrint(Ddi_VarSet_t *vars);
EXTERN Ddi_VarSet_t * Ddi_VarSetUnion(Ddi_VarSet_t *v1, Ddi_VarSet_t *v2);
EXTERN Ddi_VarSet_t * Ddi_VarSetIntersect(Ddi_VarSet_t *varset1, Ddi_VarSet_t *varset2);
EXTERN Ddi_VarSet_t * Ddi_VarSetDifference(Ddi_VarSet_t *varset, Ddi_VarSet_t *vars);
EXTERN Ddi_VarSet_t * Ddi_VarSetRemove(Ddi_VarSet_t *varset, Ddi_Var_t *var);
EXTERN int Ddi_VarSetEqual(Ddi_VarSet_t *varset1, Ddi_VarSet_t *varset2);
EXTERN Ddi_Var_t * Ddi_VarSetTop(Ddi_VarSet_t *varset);
EXTERN Ddi_Var_t * Ddi_VarSetBottom(Ddi_VarSet_t *varset);
EXTERN Ddi_DdNode * Ddi_VarSetToCU(Ddi_VarSet_t *v);
EXTERN Ddi_VarSet_t * Ddi_VarSetSwapVariables(Ddi_VarSet_t *f, Ddi_VarArray_t *x, Ddi_VarArray_t *y);

/**AutomaticEnd***************************************************************/

#endif /* _DDI */
