/**CFile***********************************************************************

  FileName    [ddiDd.c]

  PackageName [ddi]

  Synopsis    [Functions working on Boolean functions (Ddi_Dd_t)]

  Description [This file includes the set of operators working on Boolean
    functions represented by the Ddi_Dd_t type. Internally,
    functions are implemented using handles (wrappers) pointing
    to BDD roots. Externally, they are accessed only by pointers.
    All the results of operations are <B>implicitly</B> 
    referenced, while explicit free-ing (dereferencing) is required.<br>
    External procedures in this module include
    <ul> 
    <li> Basic Boolean operations: And, Or, ..., ITE
    <li> Specific BDD operators: Constrain, Restrict
    <li> Quantification operators: Exist, And-Exist (relational product)
    <li> Comparison operators: Equality, Inclusion, Tautology checks 
    <li> Manipulation of (disjunctively and/or conjunctively) 
         partitioned BDDs: create, add/remove partitions, 
    <li> translation from/to CUDD BDDs.
    <li> Dumping on file, based on the DDDMP format, distributed 
         with CUDD.
    </ul>
    ]

  SeeAlso   []  

  Author    [Gianpiero Cabodi and Stefano Quer]

  Copyright [This file was created at the Politecnico di Torino,
    Torino, Italy. 
    The  Politecnico di Torino makes no warranty about the suitability of 
    this software for any purpose.  
    It is presented on an AS IS basis.]
  
  Revision  []

******************************************************************************/

#include "ddiInt.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

#define CPU_TIME_DEBUG 0

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/
  
/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/
  
/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static Ddi_DdArray_t * GenPartDdRoots(Ddi_Dd_t *f);
static void GenPartDdRootsRecur(Ddi_Dd_t *f, Ddi_DdArray_t *roots);

/**AutomaticEnd***************************************************************/

/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Computes an operation on Boolean functions]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

Ddi_Dd_t *
DdiDdOp (
  ddi_op_t  opcode        /* operation code */,
  Ddi_Dd_t  *f            /* first dd operand */,
  Ddi_Dd_t  *g            /* second dd operand */,
  Ddi_Dd_t  *h            /* third dd operand */,
  Ddi_VarSet_t *vs        /* variable set */
  )
{
  int i;
  DdNode *r_cu = NULL;                 /* CUDD result */
  Ddi_Dd_t *r;                         /* result */
  Ddi_Dd_t *p;                         /* partition */
  Ddi_DdMgr *dd = Ddi_DdReadMgr(f);    /* dd manager */
  DdManager *mgrCU = 
    Ddi_MgrReadMgrCU(dd);              /* CUDD manager */

  Pdtutil_Assert (f!=NULL, "NULL operand");
  Pdtutil_Assert (g->code==DDI_MONO_BOOLF, "Invalid operand");

  switch(f->code) {
    case DDI_MONO_BOOLF:
      break;
    case DDI_CPART_BOOLF:
      /*
       * Restrictions Exist
       */

      Pdtutil_Warning (1,
        "Smooth a ConjPartArray: Correct only if smooth variables in 1 part");

    case DDI_DPART_BOOLF:
      r = Ddi_DdDup (f);
      for (i=0; i<Ddi_DdReadPartNum(r); i++) {
        p = Ddi_DdReadIthPart(r,i);
        p = DdiDdOp(opcode,p,g,h,vs);
        Ddi_DdReplaceIthPart(r,i,p);
      }
      return (r);
  }

  switch (opcode) {
    case DDI_OP_AND:
      if ((f->profileInfo!=NULL)||(g->profileInfo!=NULL)) {  
        Ddi_VarSet_t *no_vars = Ddi_VarSetVoid(dd);
        r = Ddi_DdAndExist(f,g,no_vars);
        Ddi_VarSetFree(no_vars);
        return (r);
      } 
      r_cu = Cudd_bddAnd(mgrCU,Ddi_DdToCU(f),Ddi_DdToCU(g));
      break;
    case DDI_OP_OR:
      r_cu = Cudd_bddOr(mgrCU,Ddi_DdToCU(f),Ddi_DdToCU(g));
      break;
    case DDI_OP_NOT:
      r_cu = Cudd_Not(Ddi_DdToCU(f));
      break;
    case DDI_OP_EXIST:
      if (f->profileInfo!=NULL) {
        return(Ddi_DdAndExist(f,Ddi_MgrReadOne(dd),vs));
      } 
      else {
        r_cu = Cudd_bddExistAbstract (mgrCU,
                 Ddi_DdToCU(f),Ddi_VarSetToCU(vs));
      }
      break;
    case DDI_OP_ANDEXIST:
      /*@@*/
      break;
    case DDI_OP_XOR:
      break;
    case DDI_OP_XNOR:
      break;
    case DDI_OP_ITE:
      break;
    case DDI_OP_CONSTRAIN:
      break;
    case DDI_OP_RESTRICT:
      break;
    default:
      break;
  }

  r = Ddi_DdCreateFromCU (dd, r_cu);

  return (r);
}

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/


/**Function********************************************************************

  Synopsis    [Computes the Dense Super or Subset of a Boolean functions]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

Ddi_Dd_t *
Ddi_DdDenseSet (
  Ddi_DenseMethod_e method      /* Operation Code */,
  Ddi_Dd_t *f                   /* Operand */,
  int threshold,
  int safe,
  int quality,
  double hardlimit
 )
{
  int suppSize;
  DdNode *sourceCudd, *resultCudd;
  Ddi_Dd_t *result;
  Ddi_DdMgr *dd = Ddi_DdReadMgr(f);
  DdManager *cuddMgr = Ddi_MgrReadMgrCU (dd);
  Ddi_VarSet_t *supp;

  supp = Ddi_DdSupport (f);
  suppSize = Ddi_VarSet_n (supp);
  sourceCudd = Ddi_DdToCU (f);

  switch (method) {
    case Ddi_DenseNone_c:
      resultCudd = sourceCudd;
      break;
    case Ddi_DenseSubHeavyBranch_c:
      resultCudd = Cudd_SubsetHeavyBranch (cuddMgr, sourceCudd, suppSize,
        threshold);
      break;
    case Ddi_DenseSupHeavyBranch_c:
      resultCudd = Cudd_SupersetHeavyBranch (cuddMgr, sourceCudd, suppSize,
        threshold);
      break;
    case Ddi_DenseSubShortPath_c:
      resultCudd = Cudd_SubsetShortPaths (cuddMgr, sourceCudd, suppSize,
        threshold, hardlimit);
      break;
    case Ddi_DenseSupShortPath:
      resultCudd = Cudd_SupersetShortPaths (cuddMgr, sourceCudd, suppSize,
        threshold, hardlimit);
      break;
    case Ddi_DenseUnderApprox_c:
      resultCudd = Cudd_UnderApprox (cuddMgr, sourceCudd, suppSize, threshold,
        safe, quality);
      break;
    case Ddi_DenseOverApprox_c:
      resultCudd = Cudd_OverApprox (cuddMgr, sourceCudd, suppSize, threshold,
        safe, quality);
      break;
    case Ddi_DenseRemapUnder_c:
      resultCudd = Cudd_RemapUnderApprox (cuddMgr, sourceCudd, suppSize,
        threshold, quality);
      break;
    case Ddi_DenseRemapOver_c:
      resultCudd = Cudd_RemapOverApprox (cuddMgr, sourceCudd, suppSize,
        threshold, quality);
      break;
    case Ddi_DenseSubCompress_c:
      resultCudd = Cudd_SubsetCompress (cuddMgr, sourceCudd, suppSize,
        threshold);
      break;
    case Ddi_DenseSupCompress_c:
      resultCudd = Cudd_SupersetCompress (cuddMgr, sourceCudd, suppSize,
        threshold);
      break;
    default:
      resultCudd = sourceCudd;
      break;
  }

  result = Ddi_DdCreateFromCU (dd, resultCudd);

  Ddi_VarSetFree (supp);
  return (result);
}

/**Function********************************************************************

  Synopsis    [Reads the code.]

  Description []

  SideEffects []

  SeeAlso     []

******************************************************************************/

int
Ddi_DdReadCode(
  Ddi_Dd_t * f
)
{
  return f->code;
}

/**Function********************************************************************

  Synopsis    [Sets the code of the boolean function.]

  Description []

  SideEffects []

  SeeAlso     []

******************************************************************************/

void
Ddi_DdSetCode(
  Ddi_Dd_t * f,
  int code
)
{
  f->code = code;
  return;
}

/**Function********************************************************************

  Synopsis    [Reads the DD Manager of the boolean function f.]

  Description []

  SideEffects []

  SeeAlso     []

******************************************************************************/

Ddi_DdMgr *
Ddi_DdReadMgr(
  Ddi_Dd_t * f
)
{
  return f->mgr;
}

/**Function********************************************************************

  Synopsis    [Sets the DD Manager of the boolean function f.]

  Description []

  SideEffects []

  SeeAlso     []

******************************************************************************/

void
Ddi_DdSetMgr(
  Ddi_Dd_t * f,
  Ddi_DdMgr * dd
)
{
  f->mgr = dd;
  return;
}

/**Function********************************************************************

  Synopsis    [Reads the array of partitions of the boolean function f.]

  Description []

  SideEffects []

  SeeAlso     []

******************************************************************************/

Ddi_DdArray_t *
Ddi_DdReadPart(
  Ddi_Dd_t * f
)
{
  return f->boolf.part;
}

/**Function********************************************************************

  Synopsis    [Sets the array of partitions of the boolean function f.]

  Description []

  SideEffects []

  SeeAlso     []

******************************************************************************/

void
Ddi_DdSetPart(
  Ddi_Dd_t * f,
  Ddi_DdArray_t *p  
)
{
  f->boolf.part=p;
  return;
}

/**Function********************************************************************

  Synopsis    [Reads the number of partitions of the boolean function f.]

  Description []

  SideEffects []

  SeeAlso     []

******************************************************************************/

int
Ddi_DdReadPartNum(
  Ddi_Dd_t * f
)
{
  return Ddi_DdArray_n(f->boolf.part);
}

/**Function********************************************************************

  Synopsis    [Reads the i-th partition of the boolean function f.]

  Description []

  SideEffects []

  SeeAlso     []

******************************************************************************/

Ddi_Dd_t *
Ddi_DdReadIthPart(
  Ddi_Dd_t * f,
  int i
)
{
  return Ddi_DdArrayFetch(f->boolf.part,i);
}

/**Function********************************************************************

  Synopsis    [Builds a Ddi_Dd_t from a given CUDD node.]

  Description [Builds the Ddi_Dd_t structure from manager and node.
    The reference count of the node is increased.]

  SeeAlso     []

  SideEffects []

******************************************************************************/

Ddi_Dd_t *
Ddi_DdCreateFromCU (
  Ddi_DdMgr *mgr,
  DdNode * f
)
{
  Ddi_Dd_t *result;

  Cudd_Ref (f);
  result = Pdtutil_Alloc (Ddi_Dd_t, 1);
  if (result == NULL) {
    Cudd_RecursiveDeref(mgr->mgrCU,f);
    return(NULL);
  }
  mgr->ddNum++;
  result->code = DDI_MONO_BOOLF;
  result->mgr = mgr;
  result->profileInfo = NULL;
  result->boolf.mono = f;
  return(result);

}

/**Function********************************************************************

  Synopsis     [Prints a BDD]

  Description  [Calls cuddP. Used to free a BDD that is no longer needed.] 

  SideEffects  [none]

  SeeAlso      []

******************************************************************************/

void
Ddi_DdPrint (
  Ddi_Dd_t  *f 
)
{
  switch(f->code) {
    case DDI_MONO_BOOLF:
      printf("Monolithic Boolf\n");
      cuddP(f->mgr->mgrCU,Ddi_DdToCU(f));
      break;
    case DDI_CPART_BOOLF:
      printf("Conj part Boolf\n");
      Ddi_DdArrayPrint(f->boolf.part);
      break;
    case DDI_DPART_BOOLF:
      printf("Disj part Boolf\n");
      Ddi_DdArrayPrint(f->boolf.part);
      break;
  }

  return;
}

/**Function********************************************************************

  Synopsis     [Prints the support of a BDD]

  Description  [Calls cuddP. Used to free a BDD that is no longer needed.] 

  SideEffects  [none]

  SeeAlso      []

******************************************************************************/

void
Ddi_DdSupportPrint (
  Ddi_Dd_t  *f 
)
{
  int i;
  Ddi_VarSet_t *supp;

  switch(f->code) {
    case DDI_MONO_BOOLF:
      fprintf (stdout, "Monolithic BDD Support:\n");
      supp = Ddi_DdSupport (f);
      Ddi_VarSetPrint (supp); 
      printf ("\n");
      Ddi_VarSetFree (supp);
      break;
    case DDI_CPART_BOOLF:
    case DDI_DPART_BOOLF:
      fprintf (stdout, "Partitioned BDD Support:\n");
      for (i=0; i<Ddi_DdReadPartNum(f); i++) {
        Ddi_DdSupportPrint(Ddi_DdReadIthPart(f,i));
      }
      break;
  }

  return;
}

/**Function********************************************************************

  Synopsis     [Gets the size of a BDD]

  Description  [] 

  SideEffects  [none]

  SeeAlso      []

******************************************************************************/

int
Ddi_DdSize(
  Ddi_Dd_t  *f 
)
{
  Ddi_DdArray_t *roots;
  int size;

  if (f==NULL)
    return (0);

  switch(f->code) {
    case DDI_MONO_BOOLF:
      return (Cudd_DagSize (Ddi_DdToCU (f)));
      break;
    case DDI_CPART_BOOLF:
    case DDI_DPART_BOOLF:
      roots = GenPartDdRoots (f);
      size = Ddi_DdArraySize (roots);
      Ddi_DdArrayFree (roots);
      return (size);
      break;
  }

  return (0); 
}

/**Function********************************************************************

  Synopsis     [Prints Statistics of a BDD]

  Description  [] 

  SideEffects  [none]

  SeeAlso      []

******************************************************************************/

void
Ddi_DdPrintStats (
  Ddi_Dd_t  *f 
)
{
  Ddi_DdMgr *dd;
  DdManager *cuddMgr;
  Ddi_VarSet_t *supp;
  int i, totalSize, partSize, suppSize;

  if (f==NULL) {
    return;
  }

  dd = Ddi_DdReadMgr (f);
  cuddMgr = Ddi_MgrReadMgrCU (dd);

  switch (f->code) {
    case DDI_MONO_BOOLF:
      supp = Ddi_DdSupport (f);
      suppSize = Ddi_VarSet_n (supp);
      Ddi_VarSetFree (supp);
      fprintf (stdout, "Monolitic BDD\n");
      fprintf (stdout, "Size:    %d\n", Ddi_DdSize (f));
      fprintf (stdout, "Density: %8.4g\n",
        Cudd_Density (cuddMgr, Ddi_DdToCU (f), suppSize));
      break;
    case DDI_CPART_BOOLF:
      fprintf (stdout, "Conjunctive Partitioned BDD\n");
      fprintf (stdout, "Number of Partitions : %d\n", Ddi_DdReadPartNum (f));
      fprintf (stdout, "Sharing Size         : %d\n", Ddi_DdSize (f));
      fprintf (stdout, "Partition Size       : ");
      totalSize = 0;
      for (i=0; i<Ddi_DdReadPartNum (f); i++) {
        partSize = Ddi_DdSize (Ddi_DdReadIthPart (f, i));
        totalSize += partSize;
        fprintf (stdout, "%d+", partSize);
      }
      fprintf (stdout, "=%d\n", totalSize);
      break;
    case DDI_DPART_BOOLF:
      fprintf (stdout, "Disjunctive Partitioned BDD\n");
      fprintf (stdout, "Number of Partitions : %d\n", Ddi_DdReadPartNum(f));
      fprintf (stdout, "Sharing Size         : %d\n", Ddi_DdSize (f));
      fprintf (stdout, "Partition Size       : ");
      totalSize = 0;
      for (i=0; i<Ddi_DdReadPartNum(f); i++) {
        partSize = Ddi_DdSize (Ddi_DdReadIthPart (f, i));
        totalSize += partSize;
        fprintf (stdout, "%d+", partSize);
      }
      fprintf (stdout, "=%d\n", totalSize);
      break;
  }

  /*
  double nstates;
  nstates = Ddi_CountMinterm(reached,Ddi_ReadSize(dd)) /
    pow(2.0,(double) (Ddi_ReadSize(dd) -
    Ddi_VarArray_n(Trav_MgrReadPS(travMgr))));
  fprintf(stdout, "# REACHED states : %lg\n",nstates);
  */
  fflush (stdout);

  return;
}

/**Function********************************************************************

  Synopsis     [Frees a BDD by De-referencing it]

  Description  [Calls Cudd_RecursiveDeref. Used to free a BDD
    that is no longer needed.] 

  SideEffects  [none]

  SeeAlso      []

******************************************************************************/

void
Ddi_DdFree (
  Ddi_Dd_t  *f     /* BDD to be freed */ 
)
{
  Ddi_DdMgr *dd;  /* dd manager */

  if (f == NULL) {
    return;
  }

  /*
  if (f->profileInfo!=NULL)
    Ddi_ProfileInfoFree(f);
  */

  dd = Ddi_DdReadMgr (f);    

  if (dd==NULL) {
    return;
  }

  switch (f->code) {
    case DDI_MONO_BOOLF:
      if ((f==Ddi_MgrReadOne (dd)) || (f==Ddi_MgrReadZero (dd))) {
        return;
      }
      Cudd_RecursiveDeref (dd->mgrCU, Ddi_DdToCU (f));
      dd->ddNum--;
      break;
    case DDI_CPART_BOOLF:
    case DDI_DPART_BOOLF:
      Ddi_DdArrayFree(f->boolf.part);
  }
  Pdtutil_Free(f);

  return;
}

/**Function********************************************************************

  Synopsis     [Duplicates a BDD]

  Description  [Returns the input BDD after increasing its reference count.] 

  SideEffects  [none]

  SeeAlso      []

******************************************************************************/

Ddi_Dd_t *
Ddi_DdDup(
  Ddi_Dd_t *f    /* BDD to be duplicated */ 
)
{
  Ddi_Dd_t *result;

  result = Pdtutil_Alloc( Ddi_Dd_t, 1);
  if (result == NULL) {
    return(NULL);
  }
  *result = *f;
  switch(f->code) {
    case DDI_MONO_BOOLF:
      f->mgr->ddNum++;
      Cudd_Ref(Ddi_DdToCU(f));
      break;
    case DDI_CPART_BOOLF:
    case DDI_DPART_BOOLF:
      result->boolf.part = Ddi_DdArrayDup(f->boolf.part);
      break;
  }

  return(result);
}

/**Function********************************************************************

  Synopsis     [Copies a BDD to a destination dd manager]

  Description  [] 

  SideEffects  [none]

  SeeAlso      []

******************************************************************************/

Ddi_Dd_t *
Ddi_DdCopy (
  Ddi_DdMgr *dd2   /* destination manager */,
  Ddi_Dd_t *f      /* BDD to be copied */ 
  )
{
  Ddi_Dd_t *result = NULL;
  Ddi_DdMgr *dd = Ddi_DdReadMgr(f); /* source manager */

  /* Same dd manager: call dup */
  if (dd2 == dd) {
    return (Ddi_DdDup(f));
  }

  switch(f->code) {
    case DDI_MONO_BOOLF:
      result = Ddi_DdCreateFromCU(dd2,
                 Cudd_bddTransfer(dd->mgrCU,dd2->mgrCU,Ddi_DdToCU(f))
               );
      break;
    case DDI_CPART_BOOLF:
    case DDI_DPART_BOOLF:
      result = Pdtutil_Alloc (Ddi_Dd_t, 1);
      if (result == NULL) {
        return(NULL);
      }
      *result = *f;
      result->mgr = dd2;
      result->boolf.part = Ddi_DdArrayCopy(dd2,f->boolf.part);
      break;
  }

  return (result);
}

/**Function********************************************************************

  Synopsis     [Copies a BDD to a destination dd manager using a third manager]

  Description  [] 

  SideEffects  [none]

  SeeAlso      []

******************************************************************************/

Ddi_Dd_t *
Ddi_DdCopyWithAuxMgr (
  Ddi_DdMgr *dd2         /* Destination Manager */,
  Ddi_Dd_t *f            /* BDD to be Copied */,
  Ddi_DdMgr *ddAux       /* Auxiliary Temporary Manager */
)
{
  Ddi_Dd_t *result, *auxf;
  Ddi_DdMgr *dd = Ddi_DdReadMgr(f); /* source manager */
#if CPU_TIME_DEBUG
  static long totTime = 0;
  long startTime, currTime;
#endif

#if CPU_TIME_DEBUG
  startTime = util_cpu_time();
#endif

  /* Same dd manager: call dup */
  if (dd2 == dd) {
    result = Ddi_DdDup (f);
  } else {
    if (ddAux == NULL) {
      result = Ddi_DdCopy (dd2, f);
    } else {
      Ddi_MgrAlign (ddAux,dd);
      auxf = Ddi_DdCopy (ddAux,f);
      Ddi_MgrAlign (ddAux,dd2);
      result = Ddi_DdCopy (dd2,auxf);
      Ddi_DdFree (auxf);
    }
  }

#if CPU_TIME_DEBUG
  currTime = util_cpu_time();
  totTime = totTime + (currTime - startTime);
  fprintf (stdout, "(CopyDdTime: %s)", util_print_time(currTime - startTime));
  fprintf (stdout, "(CopyDdTotTime: %s)\n", util_print_time (totTime));
#endif

  return (result);
}


/**Function********************************************************************

  Synopsis     [Copies a BDD to a destination dd manager using a third manager]

  Description  [] 

  SideEffects  [none]

  SeeAlso      []

******************************************************************************/

Ddi_Dd_t *
Ddi_DdSwapVariablesWithAuxMgr (
  Ddi_Dd_t       *f    /* input function */,
  Ddi_VarArray_t *src  /* source array of variables */ ,
  Ddi_VarArray_t *dst  /* dest array of variables */,
  Ddi_DdMgr *ddAux     /* Auxiliary Temporary Manager */
)
{
  Ddi_Dd_t *result, *auxf;
  Ddi_VarArray_t *aux_src, *aux_dst;
  int *src_ord, *dst_ord, *src_ordTot, *dst_ordTot, *aux_ord;
  int nv, nvTot, i, j, pos;

  Ddi_DdMgr *dd = Ddi_DdReadMgr(f); /* source manager */


#if CPU_TIME_DEBUG
  static long totTime = 0;
  long startTime, currTime;
#endif

#if CPU_TIME_DEBUG
  startTime = util_cpu_time();
#endif

  if (ddAux == NULL) {
    result = Ddi_DdSwapVariables (f,src,dst);
  } else {

    Ddi_MgrAutodynSuspend(ddAux);

    nvTot = Ddi_ReadNumVars(dd);
    nv    = Ddi_VarArray_n(src);
    Pdtutil_Assert(Ddi_ReadNumVars(ddAux)==nvTot,
      "Wrong # of variables in aux manager");
    Pdtutil_Assert(Ddi_VarArray_n(dst)==nv,
      "Var arrays have wrong variable numbers in swap var");

    aux_ord = Pdtutil_Alloc(int,nvTot); 
    src_ord = Pdtutil_Alloc(int,nv); 
    dst_ord = Pdtutil_Alloc(int,nv); 
    src_ordTot = Pdtutil_Alloc(int,nvTot); 
    dst_ordTot = Pdtutil_Alloc(int,nvTot); 

    for(i=0; i<nvTot; i++)
      aux_ord[i] = -1;
    for(i=0; i<nv; i++) {
      pos = Ddi_VarCurrPos(dd,Ddi_VarArrayFetch(src,i));
      aux_ord[pos] = i;
    }
    for(i=0,j=0; i<nvTot; i++)
      if (aux_ord[i] != -1)
        src_ord[j++] = aux_ord[i];
    Pdtutil_Assert(j==nv,"Wrong number of src variables in swap var");

    for(i=0; i<nvTot; i++)
      aux_ord[i] = -1;
    for(i=0; i<nv; i++) {
      pos = Ddi_VarCurrPos(dd,Ddi_VarArrayFetch(dst,i));
      aux_ord[pos] = i;
    }
    for(i=0,j=0; i<nvTot; i++)
      if (aux_ord[i] != -1)
        dst_ord[j++] = aux_ord[i];
    Pdtutil_Assert(j==nv,"Wrong number of src variables in swap var");

    for(i=0; i<nv; i++) {
      j = src_ord[i];
      src_ordTot[2*i] = Ddi_VarIndex(Ddi_VarArrayFetch(src,j));
      src_ordTot[2*i+1] = Ddi_VarIndex(Ddi_VarArrayFetch(dst,j));
    }
    
    for(i=0; i<nv; i++) {
      j = dst_ord[i];
      dst_ordTot[2*i] = Ddi_VarIndex(Ddi_VarArrayFetch(src,j));
      dst_ordTot[2*i+1] = Ddi_VarIndex(Ddi_VarArrayFetch(dst,j));
    }

    Ddi_MgrShuffle (ddAux, src_ordTot, 2*nv);

    auxf = Ddi_DdCopy (ddAux,f);
    aux_src = Ddi_VarArrayCopy (ddAux, src);
    aux_dst = Ddi_VarArrayCopy (ddAux, src);

    auxf = Ddi_DdEvalFree(Ddi_DdSwapVariables (auxf,src,dst), auxf);

    Ddi_MgrShuffle (ddAux, dst_ordTot, 2*nv);

    result = Ddi_DdCopy (dd,auxf);

    Ddi_DdFree (auxf);
    Ddi_VarArrayFree (aux_src);
    Ddi_VarArrayFree (aux_dst);

    Pdtutil_Free(aux_ord);
    Pdtutil_Free(src_ord);
    Pdtutil_Free(dst_ord);
    Pdtutil_Free(src_ordTot);
    Pdtutil_Free(dst_ordTot);

    Ddi_MgrAutodynResume(ddAux);

  }

#if CPU_TIME_DEBUG
  currTime = util_cpu_time();
  totTime = totTime + (currTime - startTime);
  fprintf (stdout, "(CopyDdTime: %s)", util_print_time(currTime - startTime));
  fprintf (stdout, "(CopyDdTotTime: %s)\n", util_print_time (totTime));
#endif

  return (result);
}

/**Function********************************************************************

  Synopsis     [Evaluates an expression and frees a function]

  Description  [Useful when computing a new value for a variable
    and the old value must be freed. Avoids using temporary
    variables. Since the f expression is evalued before
    passing actual parameters, freing of g occurs as last
    operation.
    ] 

  SideEffects  [none]

  SeeAlso      []

******************************************************************************/

Ddi_Dd_t *
Ddi_DdEvalFree(
  Ddi_Dd_t *f    /* expression */,
  Ddi_Dd_t *g    /* function to be freed */
)
{
  Ddi_DdFree (g);

  return (f);
}

/**Function********************************************************************

  Synopsis    [Returns the variable index]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

int
Ddi_DdTopvarIndex (
  Ddi_Dd_t *f
)
{
  return ((int)Cudd_NodeReadIndex(Ddi_DdToCU(f)));
}

/**Function********************************************************************

  Synopsis    [Returns a literal from a variable]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

Ddi_Dd_t *
Ddi_DdLiteral (
  Ddi_DdMgr *dd,
  Ddi_Var_t *v
)
{
  return (Ddi_DdCreateFromCU(dd,Ddi_VarToCU(v)));
}

/**Function********************************************************************

  Synopsis    [Returns a complemented literal from a variable]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

Ddi_Dd_t *
Ddi_DdComplLiteral (
  Ddi_DdMgr *dd,
  Ddi_Var_t *v
)
{
  return (Ddi_DdCreateFromCU(dd,Cudd_Not(Ddi_VarToCU(v))));
}

/**Function********************************************************************

  Synopsis    [The function returns 1 if the two DDs are equal]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

int
Ddi_DdEqual (
  Ddi_Dd_t *f1  /* first dd */,
  Ddi_Dd_t *f2  /* second dd */)
{
  return (Ddi_DdToCU(f1)==
          Ddi_DdToCU(f2));
}

/**Function********************************************************************

  Synopsis    [The function returns 1 if the DD f is the zero constant.]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

int
Ddi_DdIsZero (
  Ddi_Dd_t *f
)
{
  return Ddi_DdEqual(f,Ddi_MgrReadZero(Ddi_DdReadMgr(f)));
}

/**Function********************************************************************

  Synopsis    [The function returns 1 if the DD f is the one constant.]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

int
Ddi_DdIsOne (
  Ddi_Dd_t *f
)
{
  return Ddi_DdEqual(f,Ddi_MgrReadOne(Ddi_DdReadMgr(f)));
}


/**Function********************************************************************

  Synopsis     [Returns the complement of a function]

  Description  [calls Cudd_Not. Result automatically referenced by conversion
                function Ddi_DdCreateFromCU]

  SideEffects  [none]

  SeeAlso      []

******************************************************************************/

Ddi_Dd_t *
Ddi_DdNot( 
  Ddi_Dd_t *f )
{
  Ddi_Dd_t *r;
 
  r = Ddi_DdCreateFromCU(
          Ddi_DdReadMgr (f), Cudd_Not (Ddi_DdToCU (f))
      );

  return (r);
}

/**Function********************************************************************

  Synopsis     [Computes the AND of two BDDs]

  Description  []

  SideEffects  [none]

  SeeAlso      []

******************************************************************************/

Ddi_Dd_t *
Ddi_DdAnd (
  Ddi_Dd_t  *f,
  Ddi_Dd_t  *g )
{
  Ddi_Dd_t *r, *r_i;
  Ddi_DdMgr *dd;

  if ((f == NULL) || (g == NULL))
    return NULL;

  dd = Ddi_DdReadMgr(f);    /* dd manager */

  Pdtutil_Assert(g->code==DDI_MONO_BOOLF, "Invalid partitioned operand");

  switch(f->code) {
    case DDI_MONO_BOOLF:
      if ((f->profileInfo!=NULL)||(g->profileInfo!=NULL)) {  
        Ddi_VarSet_t *no_vars = Ddi_VarSetVoid(dd);
        r = Ddi_DdAndExist(f,g,no_vars);
        Ddi_VarSetFree(no_vars);
      } 
      else 
        r = Ddi_DdCreateFromCU(dd,
          Cudd_bddAnd(dd->mgrCU,Ddi_DdToCU(f),Ddi_DdToCU(g)));
      break;
    case DDI_DPART_BOOLF:
      r = Ddi_DdDup(f);
      Ddi_DdArrayAnd (r->boolf.part,g);
      break;
    case DDI_CPART_BOOLF:
      /* and only first partition */
      r = Ddi_DdDup(f);
      r_i = Ddi_DdReadIthPart(r,0); 
      r_i = Ddi_DdEvalFree(Ddi_DdAnd(r_i,g),r_i);
      Ddi_DdArrayInsert(r->boolf.part,0,r_i);
      break;
    default:
      fprintf (stderr, "Error : wrong dd node code -> %d\n",f->code);
      r = NULL;
  }
  return r;
}


/**Function********************************************************************

  Synopsis    [Computes the OR of two BDDs]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

Ddi_Dd_t *
Ddi_DdOr (
  Ddi_Dd_t  *f,
  Ddi_Dd_t  *g )
{
  Ddi_Dd_t *r;
  Ddi_DdMgr *dd = Ddi_DdReadMgr(f);    /* dd manager */

  r = Ddi_DdCreateFromCU(dd,
          Cudd_bddOr(dd->mgrCU,Ddi_DdToCU(f),Ddi_DdToCU(g))
      );
  return r;
}

/**Function********************************************************************

  Synopsis     [Computes the XOR of two BDDs]

  Description []

  SideEffects  [none]

  SeeAlso      []

******************************************************************************/

Ddi_Dd_t *
Ddi_DdXor (
  Ddi_Dd_t  *f,
  Ddi_Dd_t  *g )
{
  Ddi_Dd_t *r;
  Ddi_DdMgr *dd = Ddi_DdReadMgr(f);    /* dd manager */

  r = Ddi_DdCreateFromCU(dd,
          Cudd_bddXor(dd->mgrCU,Ddi_DdToCU(f),Ddi_DdToCU(g))
      );
  return r;
}

/**Function********************************************************************

  Synopsis     [Computes the XNOR of two BDDs]

  Description []

  SideEffects  [none]

  SeeAlso      []

******************************************************************************/

Ddi_Dd_t *
Ddi_DdXnor (
  Ddi_Dd_t  *f,
  Ddi_Dd_t  *g )
{
  Ddi_Dd_t *r;
  Ddi_DdMgr *dd = Ddi_DdReadMgr(f);    /* dd manager */

  r = Ddi_DdCreateFromCU(dd,
          Cudd_bddXnor(dd->mgrCU,Ddi_DdToCU(f),Ddi_DdToCU(g))
      );
  return r;
}

/**Function********************************************************************

  Synopsis     [ITE operator]

  Description []

  SideEffects  [none]

  SeeAlso      []

******************************************************************************/

Ddi_Dd_t *
Ddi_DdIte (
  Ddi_Dd_t *f,
  Ddi_Dd_t *g, 
  Ddi_Dd_t *h)
{
  Ddi_Dd_t *r;
  Ddi_DdMgr *dd = Ddi_DdReadMgr(f);    /* dd manager */

  r = Ddi_DdCreateFromCU(dd,
          Cudd_bddIte(dd->mgrCU,Ddi_DdToCU(f),Ddi_DdToCU(g),Ddi_DdToCU(h))
      );
  return r;
}


/**Function********************************************************************

  Synopsis    [Computes the AND-Abstract of two BDDs] 

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

Ddi_Dd_t *
Ddi_DdAndExist(
  Ddi_Dd_t  *f,
  Ddi_Dd_t  *g,
  Ddi_VarSet_t *supp )
{
  Ddi_Dd_t *r;
  DdNode *rCu;
  Ddi_DdMgr *dd = Ddi_DdReadMgr(f);    /* dd manager */

  if ((f->profileInfo==NULL)&&(g->profileInfo==NULL)) 
    rCu = Cudd_bddAndAbstract (dd->mgrCU, Ddi_DdToCU (f), Ddi_DdToCU (g),
      Ddi_DdToCU(supp));
  else {
    if ((f->profileInfo!=NULL)&&(g->profileInfo!=NULL)) {
      printf("Warning: two operands cannot be profiled together.\n");
      printf("First one not profiled.\n");
    }
    if (f->profileInfo!=NULL) { /* second operand must be profiled */
      Ddi_Dd_t *tmp = g;
      g=f; f=tmp;
    }
    rCu = Cuplus_ProfileAndAbstract (dd->mgrCU, Ddi_DdToCU (f), Ddi_DdToCU (g),
      Ddi_DdToCU (supp), g->profileInfo);
  }
  r = Ddi_DdCreateFromCU(dd,rCu);

  return r;
}

/**Function********************************************************************

  Synopsis    [Computes the existential abstraction of a BDD with respect to
    a support]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

Ddi_Dd_t *
Ddi_DdExist(
  Ddi_Dd_t  *f,
  Ddi_VarSet_t *supp)
{
  Ddi_Dd_t *r;
  Ddi_DdMgr *dd = Ddi_DdReadMgr(f);    /* dd manager */

  switch(f->code) {
    case DDI_MONO_BOOLF:
      if (f->profileInfo!=NULL) {  
        r = Ddi_DdAndExist(f,Ddi_MgrReadOne(dd),supp);
      } 
      else 
        r = Ddi_DdCreateFromCU(dd,
          Cudd_bddExistAbstract(dd->mgrCU,Ddi_DdToCU(f),Ddi_VarSetToCU(supp))
      );
      break;
    case DDI_CPART_BOOLF:
      {
      int i;
      Ddi_VarSet_t *suppF, *suppTot;
      /* each supp variable must belong to a single partition */
      suppTot = Ddi_VarSetVoid (dd);
      for (i=0; i<Ddi_DdReadPartNum(f); i++) {
        /* take the support of i-th function */
        suppF = Ddi_DdSupport (Ddi_DdReadIthPart(f,i));
        suppTot = Ddi_VarSetEvalFree( 
          Ddi_VarSetIntersect(suppF,suppTot),suppTot);
        Ddi_VarSetFree(suppF);
      }
      suppTot = Ddi_VarSetEvalFree( 
        Ddi_VarSetIntersect(supp,suppTot),suppTot);
      Pdtutil_Assert(Ddi_VarSetIsVoid(suppTot),
        "smoothing a conj.part.array: some partitions share smoothing vars\n");
        Ddi_VarSetFree(suppTot);
      }
    case DDI_DPART_BOOLF:
      r = Ddi_DdDup(f);
      Ddi_DdArrayExist(r->boolf.part,supp);
      break;
    default:
      fprintf (stderr, "Error : wrong dd node code -> %d\n",f->code);
      r = NULL;
  }

  return r;
}


/**Function********************************************************************

  Synopsis    [Computes the constrain cofactor between F and C]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

Ddi_Dd_t *
Ddi_DdConstrain(
  Ddi_Dd_t  *f,
  Ddi_Dd_t  *c
  )
{
  Ddi_Dd_t *r;
  Ddi_DdMgr *dd = Ddi_DdReadMgr(f);    /* dd manager */

  switch(f->code) {
    case DDI_MONO_BOOLF:
      r = Ddi_DdCreateFromCU(dd,
        Cudd_bddConstrain(dd->mgrCU,Ddi_DdToCU(f),Ddi_DdToCU(c))
        );
      break;
    case DDI_CPART_BOOLF:
    case DDI_DPART_BOOLF:
      r = Ddi_DdDup(f);
      Ddi_DdArrayConstrain (r->boolf.part,c);
      break;
    default:
      fprintf (stderr, "Error : wrong dd node code -> %d\n",f->code);
    r = NULL;
  }

  return r;
}

/**Function********************************************************************

  Synopsis    [Computes the restrict cofactor between F and C]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

Ddi_Dd_t *
Ddi_DdRestrict(
  Ddi_Dd_t  *f,
  Ddi_Dd_t  *c )
{
  Ddi_Dd_t *r;
  Ddi_DdMgr *dd = Ddi_DdReadMgr(f);    /* dd manager */

  switch(f->code) {
    case DDI_MONO_BOOLF:
        r = Ddi_DdCreateFromCU(dd,
          Cudd_bddRestrict(dd->mgrCU,Ddi_DdToCU(f),Ddi_DdToCU(c))
        );
      break;
    case DDI_CPART_BOOLF:
    case DDI_DPART_BOOLF:
      r = Ddi_DdDup(f);
      Ddi_DdArrayRestrict (r->boolf.part,c);
      break;
    default:
      fprintf (stderr, "Error : wrong dd node code -> %d\n",f->code);
      r = NULL;
  }
  return r;
}


/**Function********************************************************************

  Synopsis    [Swaps two sets of variables in a function]

  Description [Returns a new function with the variables swapped (x replaces y and
    viceversa.]

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

Ddi_Dd_t *
Ddi_DdSwapVariables(
  Ddi_Dd_t       *f    /* input function */,
  Ddi_VarArray_t *x    /* 1-st array of variables */,
  Ddi_VarArray_t *y    /* 2-nd array of variables */ )
{
  Ddi_Dd_t *r;
  DdNode **Xv,**Yv;
  int n;
  Ddi_DdMgr *dd = Ddi_DdReadMgr(f);    /* dd manager */

  /* the length of the two arrays of variables must be the same */
  n = Ddi_VarArray_n(x);
  if (Ddi_VarArray_n(y)!=n) {
    fprintf(stderr, "Error : The variables in function are not swapped.\n");
    return NULL;
  }
  
  Xv = Ddi_VarArrayToCU(x);
  if (Xv==NULL) {
    fprintf(stderr, "Error : The variables in function are not swapped.\n");
    return NULL;
  }

  Yv = Ddi_VarArrayToCU(y);
  if (Yv==NULL) {
    fprintf(stderr, "Error : The variables in function are not swapped.\n");
    return NULL;
  }

  r = Ddi_DdCreateFromCU(dd,
          Cudd_bddSwapVariables(dd->mgrCU,Ddi_DdToCU(f),Yv,Xv,n)
      );

  Pdtutil_Free(Xv);
  Pdtutil_Free(Yv);
  return r;
}

/**Function********************************************************************

  Synopsis    [Returns the support of BDD]

  Description [The support of a BDD is a cube where each node corresponds to
    a variable.]

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

Ddi_VarSet_t *
Ddi_DdSupport(
  Ddi_Dd_t *f      /* BDD */ )
{
  Ddi_VarSet_t *supp;
  Ddi_DdMgr *dd = Ddi_DdReadMgr(f);    /* dd manager */

  if (f==NULL)
    return Ddi_VarSetVoid(dd);

  switch(f->code) {
    case DDI_MONO_BOOLF:
      supp = (Ddi_VarSet_t *) Ddi_DdCreateFromCU(dd,
                               Cudd_Support(dd->mgrCU,Ddi_DdToCU(f))
        );    
      break;
    case DDI_CPART_BOOLF:
    case DDI_DPART_BOOLF:
      supp = Ddi_DdArraySupport(f->boolf.part);
      break;
    default:
      fprintf (stderr, "Error : wrong dd node code -> %d\n",f->code);
      supp = NULL;
  }
  return supp;
}

/**Function********************************************************************

  Synopsis    [Stores BDD on file]

  Description [This function stores only a BDD (not a BDDs'array).
    The BDD is stored in the DDDMP format. The parameter "mode" 
    can be DDDMP_MODE_TEXT, DDDMP_MODE_BINARY or
    DDDMP_MODE_DEFAULT.<br>
    The function returns 1 if succefully stored, 0 otherwise.
    ]

  SideEffects [None]

  SeeAlso     [Ddi_DdLoad,Dddmp_cuddBddStore]

******************************************************************************/

int
Ddi_DdStore (
  Ddi_Dd_t  *f       /* BDD */,
  char *ddname       /* dd name (or NULL) */,
  char mode          /* storing mode */,
  char *filename     /* file name */,
  FILE *fp           /* pointer to the store file */ )
{
  int close_fp = 0;
  int ret;
  Ddi_DdMgr *dd = Ddi_DdReadMgr(f);    /* dd manager */
  char **vnames = Ddi_MgrReadVarnames(dd);
  int *vauxids = Ddi_MgrReadVarauxids(dd);


  if (fp == NULL) {
    fp = fopen(filename, "w");
    if (fp == NULL)
      return 0;
    close_fp = 1;
  }

  switch(f->code) {
    case DDI_MONO_BOOLF:
      fprintf(fp,"# MONO\n");
      ret = Dddmp_cuddBddStore(dd->mgrCU,ddname,Ddi_DdToCU(f),
            vnames,vauxids,mode,DDDMP_VARNAMES,filename,fp);
      break;
    case DDI_CPART_BOOLF:
      fprintf(fp,"# CPART\n");
      ret = Ddi_DdArrayStore (f->boolf.part,ddname,vnames,
               NULL,vauxids,mode,filename,fp);
      break;
    case DDI_DPART_BOOLF:
      fprintf(fp,"# DPART\n");
      ret = Ddi_DdArrayStore (f->boolf.part,ddname,vnames,
               NULL,vauxids,mode,filename,fp);
      break;
    default:
      fprintf (stderr, "Error : wrong dd node code -> %d\n",f->code);
      ret = 0;
  }

  if (close_fp)
    fclose(fp);

  return ret;

}

/**Function********************************************************************

  Synopsis    [Loads BDD from file]

  Description [This function loads only a BDD. If the file contain a BDDs'
    array, then will be load only the first BDD.<br>
    The BDD on file must be in the DDDMP format. The parameter 
    "mode" can be DDDMP_MODE_TEXT, DDDMP_MODE_BINARY or
    DDDMP_MODE_DEFAULT.<br>
    The function returns the pointer of BDD root if succefully
    loaded, NULL otherwise.]

  SideEffects [None]

  SeeAlso     [Ddi_DdStore,Dddmp_cuddBddLoad]

******************************************************************************/

Ddi_Dd_t *
Ddi_DdLoad (
  Ddi_DdMgr *dd      /* dd manager */,
  int varmatchmode   /* variable matching mode */,
  char mode          /* loading mode */,
  char *filename     /* file name */,  
  FILE *fp           /* file pointer */ )             
{
  Ddi_Dd_t *f, *p;
  int close_fp = 0;
  char code[200],auxname[200];
  Ddi_DdArray_t *array;    /* descriptor of array */
  char **vnames = Ddi_MgrReadVarnames(dd);
  int *vauxids = Ddi_MgrReadVarauxids(dd);

  f = NULL;

  if (fp == NULL) {
    fp = fopen(filename, "r");
    if (fp == NULL)
      return 0;
    close_fp = 1;
  }

  if (fgets(code,199,fp)==NULL)
    return NULL;
  if ((strncmp(code,"# MONO",6)==0)
     || (strncmp(code,"MONO",4)==0)) {
    f = Ddi_DdCreateFromCU(dd,
              Dddmp_cuddBddLoad(dd->mgrCU,varmatchmode,vnames,vauxids,
                            NULL,mode,filename,fp));
  }
  if ((strncmp(code,"# CPART",7)==0)
     || (strncmp(code,"CPART",5)==0)) {
    array = Ddi_DdArrayLoad(dd,
             vnames,vauxids,mode,filename,fp);
    f = Ddi_DdPartFromArray(array,DDI_CPART_BOOLF);
    Ddi_DdArrayFree(array);
  }
  if ((strncmp(code,"# DPART",7)==0)
    || (strncmp(code,"DPART",5)==0)) {
    array = Ddi_DdArrayLoad(dd,
             vnames,vauxids,mode,filename,fp);
    f = Ddi_DdPartFromArray(array,DDI_DPART_BOOLF);
    Ddi_DdArrayFree(array);
  }
  if ((strncmp(code,"# FILEDPART",11)==0)
    || (strncmp(code,"FILEDPART",9)==0)) {
    f = Ddi_MgrReadZero(dd);
    while (fscanf(fp,"%s",auxname)!=EOF) {
      p = Ddi_DdLoad(dd,varmatchmode,mode,auxname,NULL);
      f = Ddi_DdEvalFree(Ddi_DdAddDpart(f,p),f);
      Ddi_DdFree(p);
    }
  }

  if (close_fp)
    fclose(fp);

  return f;
}

/**Function********************************************************************

  Synopsis    [Adds conjoined partition]

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

Ddi_Dd_t *
Ddi_DdAddCpart (
  Ddi_Dd_t *f        /* input function */,
  Ddi_Dd_t *new      /* new partition */
)
{
  Ddi_Dd_t *result;
  Ddi_DdArray_t *partitions;
  Ddi_DdMgr *dd;                  /* dd manager */

  /* if f is void or 1 return new */
  if ((f == NULL)||(Ddi_DdIsOne(f))) {
    return (Ddi_DdDup(new));
  }

  dd = Ddi_DdReadMgr(f);    

  if (f->code==DDI_CPART_BOOLF) {
    partitions = Ddi_DdArrayDup(f->boolf.part);
  }
  else {
    partitions = Ddi_DdArrayAlloc(1);
    if (!Ddi_DdIsOne(f))
      Ddi_DdArrayInsert(partitions,0,f);
  }
  Ddi_DdArrayInsert(partitions,Ddi_DdArray_n(partitions),new);
  result = Pdtutil_Alloc( Ddi_Dd_t, 1);
  if (result == NULL) {
    return(NULL);
  }
  result->code = DDI_CPART_BOOLF;
  result->boolf.part = partitions;
  result->profileInfo = NULL;
  result->mgr = dd;

  return(result);

}

/**Function********************************************************************

  Synopsis    [Adds disjoined partition]

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

Ddi_Dd_t *
Ddi_DdAddDpart (
  Ddi_Dd_t *f        /* input function */,
  Ddi_Dd_t *new      /* new partition */
)
{
  Ddi_Dd_t *result;
  Ddi_DdArray_t *partitions;
  Ddi_DdMgr *dd;                  /* dd manager */

  /* if f is void or 0 return new */
  if ((f == NULL)||(Ddi_DdIsZero(f))) {
    return (Ddi_DdDup(new));
  }

  dd = Ddi_DdReadMgr(f);    

  if (f->code==DDI_DPART_BOOLF) {
    partitions = Ddi_DdArrayDup(f->boolf.part);
  }
  else {
    partitions = Ddi_DdArrayAlloc(1);
    if (!Ddi_DdIsZero(f))
      Ddi_DdArrayInsert(partitions,0,f);
  }
  Ddi_DdArrayInsert(partitions,Ddi_DdArray_n(partitions),new);
  result = Pdtutil_Alloc( Ddi_Dd_t, 1);
  if (result == NULL) {
    return(NULL);
  }
  result->code = DDI_DPART_BOOLF;
  result->boolf.part = partitions;
  result->profileInfo = NULL;
  result->mgr = dd;

  return(result);

}

/**Function********************************************************************

  Synopsis    [Replaces i-th partition with a new one]

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

void
Ddi_DdReplaceIthPart (
  Ddi_Dd_t *f        /* input function */,
  int       i        /* index */,
  Ddi_Dd_t *p        /* new partition */
)
{

  Pdtutil_Assert(f->code!=DDI_MONO_BOOLF, "Conjunction or disjunction of BDDs expected");
  Ddi_DdFree(Ddi_DdReadIthPart(f,i));
  Ddi_DdArrayInsert(Ddi_DdReadPart(f),i,p);

}

/**Function********************************************************************

  Synopsis    [Removes a partition]

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

Ddi_Dd_t *
Ddi_DdRemPart (
  Ddi_Dd_t *f        /* input function */,
  int pi             /* partition index */
)
{
  Ddi_Dd_t *result, *p;
  Ddi_DdMgr *dd = Ddi_DdReadMgr(f);    /* dd manager */
  int i;

  switch(f->code) {
    case DDI_MONO_BOOLF:
        printf ("Warning: removing a partition from a monolithic BDD. NOT executed!\n");
        return (Ddi_DdDup(f));
      break;
    case DDI_CPART_BOOLF:
      result = Ddi_DdDup(Ddi_MgrReadOne(dd));
      for (i=0; i<Ddi_DdReadPartNum(f); i++) {
        if (i!=pi) {
          p = Ddi_DdReadIthPart(f,i);
          result = Ddi_DdEvalFree(Ddi_DdAddCpart(result,p),result);
        }
      }
      break;
    case DDI_DPART_BOOLF:
      result = Ddi_DdDup(Ddi_MgrReadZero(dd));
      for (i=0; i<Ddi_DdReadPartNum(f); i++) {
        if (i!=pi) {
          p = Ddi_DdReadIthPart(f,i);
          result = Ddi_DdEvalFree(Ddi_DdAddDpart(result,p),result);
        }
      }
      break;
    default:
      fprintf (stderr, "Error : wrong dd node code -> %d\n",f->code);
      result = NULL;
  }

  return(result);

}

/**Function********************************************************************

  Synopsis    [Creates a monolithic BDD from a partitioned one]

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

Ddi_Dd_t *
Ddi_DdPart2Mono (
  Ddi_Dd_t *f        /* input function */
)
{
  Ddi_Dd_t *result, *part;
  int i;

  if (f->code==DDI_MONO_BOOLF) { 
    return (Ddi_DdDup(f));
  }

  result = Ddi_DdPart2Mono(Ddi_DdReadIthPart(f,0));

  for (i=1;i<Ddi_DdReadPartNum(f);i++) {
    part = Ddi_DdPart2Mono(Ddi_DdReadIthPart(f,i));
    if (f->code==DDI_CPART_BOOLF) {
      result = Ddi_DdEvalFree(Ddi_DdAnd(result,part),result);
    } else {
      if (f->code==DDI_DPART_BOOLF) {
        result = Ddi_DdEvalFree(Ddi_DdOr(result,part),result);
      }
    }
    Ddi_DdFree(part);
  }

  return (result);
}

/**Function********************************************************************

  Synopsis    [Builds a partitioned function]

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

Ddi_Dd_t *
Ddi_DdPartFromArray (
  Ddi_DdArray_t *array /* input array */,
  int code             /* code: disj/conj */
)
{
  Ddi_Dd_t *result;
  Ddi_DdArray_t *partitions;
  int i;

  Pdtutil_Assert((code==DDI_CPART_BOOLF)||(code==DDI_DPART_BOOLF), 
    "Partition code invalid");

  partitions = Ddi_DdArrayDup(array);
  result = Pdtutil_Alloc( Ddi_Dd_t, 1);
  if (result == NULL) {
    return(NULL);
  }
  result->code = code;
  result->profileInfo = NULL;
  result->boolf.part = partitions;
  Pdtutil_Assert(Ddi_DdArray_n(array)>=1, "No partitions");
  result->mgr = Ddi_DdReadMgr(Ddi_DdArrayFetch(array,0));

  for (i=0;i<Ddi_DdArray_n(array);i++)
   if (Ddi_DdReadMgr(Ddi_DdArrayFetch(array,i))!=result->mgr) {
     printf ("single manager for a dd array still required!\n");
     return NULL;
   }

  return(result);
}

/**Function********************************************************************

  Synopsis    []

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

void
Ddi_ProfileInfoInit (
  Ddi_Dd_t *f          /* input function */,
  cuplus_profile_info_t *profileInfo /* profile info */,
  unsigned char dac99_compatible /* enabled for settings of paper */
)
{
  int i;

  if (profileInfo != NULL) {
    f->profileInfo = profileInfo;
    return;
  }

  switch(f->code) {
    case DDI_MONO_BOOLF:
      /* monolitic BDD: one profile generated */
      f->profileInfo = Cuplus_ProfileInfoGen(dac99_compatible);
      break;
    case DDI_CPART_BOOLF:
    case DDI_DPART_BOOLF:
      /* partitioned BDD: single profile generated for the whole
         BDD, pointed to by all partitions */
      f->profileInfo = profileInfo = 
        Cuplus_ProfileInfoGen (dac99_compatible);
      for (i=0; i<Ddi_DdReadPartNum(f); i++) {
        Ddi_ProfileInfoInit (Ddi_DdReadIthPart (f, i), 
          profileInfo, dac99_compatible);
      }
      break;
    default:
      fprintf (stderr, "Error : wrong dd node code -> %d\n",f->code);
  }

  return;
}

/**Function********************************************************************

  Synopsis    []

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

void
Ddi_ProfileInfoFree (
  Ddi_Dd_t *f          /* input function */
)
{
  int i;
  Ddi_Dd_t *part;

  switch(f->code) {
    case DDI_MONO_BOOLF:
      /* free profile */
      Cuplus_ProfileInfoFree(f->profileInfo);
      f->profileInfo = NULL;
      break;
    case DDI_CPART_BOOLF:
    case DDI_DPART_BOOLF:
      if (f->profileInfo != NULL)
        Cuplus_ProfileInfoFree(f->profileInfo);
      f->profileInfo = NULL;

      for (i=0; i<Ddi_DdReadPartNum(f); i++) {
        part = Ddi_DdReadIthPart (f, i);
        /* profile already free-ed. The following statement
           prevents multiple freeings */
        part->profileInfo = NULL;
        /* recursion is done in case of nested partitionings */
        Ddi_ProfileInfoFree (part);
      }
      break;
    default:
      fprintf (stderr, "Error : wrong dd node code -> %d\n",f->code);
  }

  return;
}

/**Function********************************************************************

  Synopsis    []

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

void
Ddi_ProfileInfoPrint (
  Ddi_Dd_t *f          /* input function */
)
{
  switch(f->code) {
    case DDI_MONO_BOOLF:
      Pdtutil_ChrPrint (stdout, '*', 50);
      fprintf (stdout, "\n");
      fprintf (stdout, "BDD Profile Summary:\n");
      fprintf (stdout, "Total Nodes                   : %d\n", Ddi_DdSize (f));

      Cuplus_ProfileInfoPrint (f->profileInfo);

      Pdtutil_ChrPrint (stdout, '*', 50);
      fprintf (stdout, "\n");
      break;
    case DDI_CPART_BOOLF:
    case DDI_DPART_BOOLF:
#if 0
      for (i=0; i<Ddi_DdReadPartNum(f); i++) {
        fprintf (stdout, "Print-Profile-Info for PARTITION %d\n", i);
        Ddi_ProfileInfoPrint (Ddi_DdReadIthPart (f, i));
      }
#else
      Pdtutil_ChrPrint (stdout, '*', 50);
      fprintf (stdout, "\n");
      fprintf (stdout, "BDD Profile Summary:\n");
      fprintf (stdout, "Total Nodes                   : %d\n", Ddi_DdSize (f));
      fprintf (stdout, "Number of partitions          : %d\n", Ddi_DdReadPartNum (f));

      Cuplus_ProfileInfoPrint (f->profileInfo);

      Pdtutil_ChrPrint (stdout, '*', 50);
      fprintf (stdout, "\n");
#endif
      break;
    default:
      fprintf (stderr, "Error : wrong dd node code -> %d\n",f->code);
  }

  return;
}

/**Function********************************************************************

  Synopsis    []

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

Ddi_Dd_t *
Ddi_PruneProfiled (
  Ddi_Dd_t *f                                /* input function */,
  Cuplus_PruneHeuristic_e prune_heuristic,
  int threshold
)
{
  int i;
  Ddi_Dd_t *res, *p;
  Ddi_DdMgr *dd = f->mgr;
  char enable[101];

  switch(f->code) {
    case DDI_MONO_BOOLF:
      res = Ddi_DdCreateFromCU (dd,
              Cuplus_ProfilePrune (dd->mgrCU,
                Ddi_DdToCU (f), f->profileInfo, prune_heuristic, threshold)
            );
      fprintf (stdout, "Size before->after pruning : %d->%d\n", Ddi_DdSize (f),
        Ddi_DdSize (res));
      if (Ddi_DdIsZero (res)) {
        /* abort because of error */
        Ddi_DdFree (res);
        fprintf (stdout, "Pruning aborted because 0!\n");
        res=Ddi_DdDup(f);
      }
      return (res);
      break;
    case DDI_CPART_BOOLF:
      Pdtutil_ChrPrint (stdout, '*', 50);
      fprintf (stdout, "\nPruned Transition Relation Summary:\n");
      fprintf (stdout, "Partitions %d, Threshold %d\n",
        Ddi_DdReadPartNum (f), threshold);
      /*
      fprintf (stdout, "enable mask: ");
      fflush(stdout);
      gets (enable);
      */
      res = Ddi_MgrReadOne(dd);
      for (i=0; i<Ddi_DdReadPartNum(f); i++) {
        /*if (enable[i] == '1')*/ 
        if (1) {
          p = Ddi_PruneProfiled (Ddi_DdReadIthPart (f, i),
            prune_heuristic, threshold);
        } else {
          p = Ddi_DdDup (Ddi_DdReadIthPart (f, i));
        }
        res = Ddi_DdEvalFree (Ddi_DdAddCpart (res, p), res);
      }
      Pdtutil_ChrPrint (stdout, '*', 50);
      fprintf (stdout, "\n");
      return (res);
      break;
    case DDI_DPART_BOOLF:
      Pdtutil_ChrPrint (stdout, '*', 50);
      fprintf (stdout, "\nPruned Transition Relation Summary:\n");
      fprintf (stdout, "Partitions %d, Threshold %d\n",
        Ddi_DdReadPartNum (f), threshold);
      /*
      fprintf (stdout, "enable mask: ");
      fflush(stdout);
      fgets (enable, 100, stdin);
      */
      res = Ddi_MgrReadZero (dd);
      for (i=0; i<Ddi_DdReadPartNum (f); i++) {
        if (enable[i] == '1') {
          p = Ddi_PruneProfiled (Ddi_DdReadIthPart (f, i), prune_heuristic,
            threshold);
	} else {
          p = Ddi_DdDup (Ddi_DdReadIthPart (f, i));
        }
        res = Ddi_DdEvalFree (Ddi_DdAddDpart (res, p), res);
      }
      Pdtutil_ChrPrint (stdout, '*', 50);
      fprintf (stdout, "\n");
      return (res);
      break;
    default:
      fprintf (stderr, "Error: wrong dd node code -> %d\n", f->code);
  }

  Pdtutil_Assert (0, "Unreachable statement reached!");

  return (NULL);
}


/**Function********************************************************************

  Synopsis    [generates a common ordering given 2 input BDDs]

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

void
Ddi_MulvarInterleave(
  Ddi_Dd_t *f          /* first input function */,
  Ddi_Dd_t *g          /* second input function */
)
{
  /*
   * Mulvar disabled
   */

#if 0
  int i;
  Ddi_Dd_t *p;
  Ddi_DdArray_t *froots, *groots;
  DdNode **proots0, **proots1;
  int *resultord0, *resultord1, *adj0, *adj1;
  int method, verbosity, result;
  Ddi_DdMgr *dd0 = f->mgr;
  Ddi_DdMgr *dd1 = g->mgr;

  fprintf (stdout, "Mulvar disabled\n");

#if 0
  if (dd0 == dd1) return;
#endif

#if 0
  p = f;
  f = g;
  g = p;
#endif

  /* generate arrays of root pointers */

  froots = GenPartDdRoots(f);
  groots = GenPartDdRoots(g);

  proots0 = Pdtutil_Alloc( DdNode*,Ddi_DdArray_n(froots)+1);
  proots1 = Pdtutil_Alloc( DdNode*,Ddi_DdArray_n(groots)+1);
  if ((proots0 == NULL) || (proots1 == NULL)) {
    printf("Allocation error in Ddi_MulvarInterleave\n");
    return;
  }

  for (i=0; i<Ddi_DdArray_n(froots); i++) {
    p = Ddi_DdArrayFetch(froots,i);
    proots0[i] = Ddi_DdToCU(p);
  }
  proots0[i] = NULL;

  for (i=0; i<Ddi_DdArray_n(groots); i++) {
    p = Ddi_DdArrayFetch(groots,i);
    proots1[i] = Ddi_DdToCU(p);
  }
  proots1[i] = NULL;

  /* generate result order arrays */ 

  resultord0=Pdtutil_Alloc( int,Ddi_ReadSize(dd0));
  resultord1=Pdtutil_Alloc( int,Ddi_ReadSize(dd1));

  /* variable correspondence arrays */

  adj0=Pdtutil_Alloc( int,Ddi_ReadSize(dd0));
  adj1=Pdtutil_Alloc( int,Ddi_ReadSize(dd1));

  for (i=0; i<Ddi_ReadSize(dd0); i++) {
    adj0[i] = adj1[i] = i;
  }

  /* select method and verbosity */

  method = 3;
  verbosity = 2;

  /* call the procedure doing job */

  fprintf (stdout, "Initial sizes - f: %d - g: %d\n", Ddi_DdSize(f),
    Ddi_DdSize(g));
  fprintf (stdout, "Calling Mulvar_interleave...\n");
  result = Mulvar_interleave (Ddi_MgrReadMgrCU (dd0), proots0,
    Ddi_MgrReadMgrCU (dd1), proots1, adj0, adj1, NULL, method, NULL, 0,
    NULL, NULL, verbosity, &resultord0, &resultord1);
  fprintf (stdout, "...done Mulvar_interleave. Return value: %d\n", result);

  Ddi_DdArrayFree (froots);
  Ddi_DdArrayFree (groots);

  Pdtutil_Free (proots0);
  Pdtutil_Free (proots1);
  Pdtutil_Free (resultord0);
  Pdtutil_Free (resultord1);
  Pdtutil_Free (adj0);
  Pdtutil_Free (adj1);
#endif

  return;
}

/**Function********************************************************************

  Synopsis    [Converts a DD into the correspondent Cudd Node]

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

Ddi_DdNode *
Ddi_DdToCU(
  Ddi_Dd_t *f         
)
{
  return ((f->code==DDI_MONO_BOOLF)?((f)->boolf.mono):NULL);
}


/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis     [Generates root pointers for partitioned DDs]

  Description  [] 

  SideEffects  [none]

  SeeAlso      []

******************************************************************************/

static Ddi_DdArray_t *
GenPartDdRoots(
  Ddi_Dd_t  *f 
)
{
  Ddi_DdArray_t *roots;

  roots = Ddi_DdArrayAlloc (0);
  if (roots == NULL) {
    return (NULL);
  }

  GenPartDdRootsRecur(f, roots);

  return (roots);
}


/**Function********************************************************************

  Synopsis     [Generates root pointers for partitioned DDs (recursive)]

  Description  [] 

  SideEffects  [none]

  SeeAlso      []

******************************************************************************/

static void
GenPartDdRootsRecur(
  Ddi_Dd_t  *f,
  Ddi_DdArray_t *roots 
)
{
  int i;

  switch(f->code) {
    case DDI_MONO_BOOLF:
      Ddi_DdArrayInsert(roots,Ddi_DdArray_n(roots),f);
      break;
    case DDI_CPART_BOOLF:
    case DDI_DPART_BOOLF:
      for (i=0; i<Ddi_DdReadPartNum(f); i++)
        GenPartDdRootsRecur(Ddi_DdReadIthPart(f,i),roots);
      break;
    default:
      fprintf (stderr, "Error : wrong dd node code -> %d\n",f->code);
  }

}
