/**CFile***********************************************************************

  FileName    [ddiDdArray.c]

  PackageName [ddi]

  Synopsis    [Functions to manage <em>arrays of BDDs</em>]

  Description [The basic manipulation allowed for BDD arrays are:<BR> 
               <OL>
               <LI><B>Alloc</B> and <B>free</B> an array
               <LI><B>Insert</B>, <B>fetch</B>, <B>extract</B> and
                   <B>include</B> a BDD into/from array
               <LI><B>Dup</B>licate an array
               <LI><B>Join</B> and <B>append</B> a new array
               <LI><B>Store</B> and <B>load</B> to/from file
               <LI>Other features as <B>resize</B>, and <B>size</B>  
               </OL>
               For each element of array can be apply any boolean operator
               as AND,OR,XOR,XNOR,Restrict and Constrain.]

  SeeAlso     []  

  Author    [Gianpiero Cabodi and Stefano Quer]

  Copyright [This file was created at the Politecnico di Torino,
  Torino, Italy. 
  The  Politecnico di Torino makes no warranty about the suitability of 
  this software for any purpose.  
  It is presented on an AS IS basis.]
  
  Revision    []

******************************************************************************/

#include "ddiInt.h"

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */ 
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */ 
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */ 
/*---------------------------------------------------------------------------*/

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static int DdArrayResize(Ddi_DdArray_t *array, int nSizeNew);

/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of internal function                                           */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/


/**Function*******************************************************************
  Synopsis    [Allocates a new array of BDDs]
  Description [The function returns a pointer to a newly created DD array
               of the required length. 
               The new array is initialized with NULL elements.
               NULL is returned in case of failure.]
  SideEffects [none]
  SeeAlso     [Ddi_DdArrayFree]
******************************************************************************/
Ddi_DdArray_t *
Ddi_DdArrayAlloc (
  int length        /* array length */ )
{
  Ddi_DdArray_t *array;
  int i;

  array = Pdtutil_Alloc(Ddi_DdArray_t,1);
  if (array == NULL) {
    fprintf(stderr, "Error : Out of memory\n"); 
    return NULL;
  }

  array->num = length;
  array->nSize = DDI_MAX(length, DDI_DDARRAY_INIT_SIZE);
  array->data = Pdtutil_Alloc(Ddi_Dd_t *, array->nSize);
  if (array->data == NULL) {
    Pdtutil_Free(array);
    fprintf(stderr, "Error : Out of memory\n"); 
    return NULL;
  }
  for (i=0;i<array->nSize;i++)
    array->data[i] = NULL;

  return array;
}


/**Function*******************************************************************
  Synopsis    [Frees an array of BDDs]
  Description [Frees all BDDs in the array and the
               dynamic memory allocated.]
  SideEffects [none]
  SeeAlso     [Ddi_DdArrayAlloc]
******************************************************************************/
void
Ddi_DdArrayFree (
  Ddi_DdArray_t *array   /* array to be freed */ )
{
  int i;
  Ddi_Dd_t *f;

  if (array==NULL)
    return;

  for (i=0; i<Ddi_DdArray_n (array); i++) {
    f = Ddi_DdArrayFetch (array, i);
    Ddi_DdFree (f);
  }

  Pdtutil_Free(array->data);
  Pdtutil_Free(array);
}

/**Function*******************************************************************
  Synopsis    [Inserts a BDD in the array]
  Description [The function inserts the BDD <em>f</em> in the <em>idx</em>
               position of the array. The inserted inserted BDD is 
               duplicated (referenced).<br>
               If the i-th array slot already contained a BDD, it is NOT
               freed, this should be done explicitely.
               The function returns 1 for success, 0 for failure.]
  SideEffects [none]
  SeeAlso     [Ddi_DdArrayInclude]
******************************************************************************/
int
Ddi_DdArrayInsert(
  Ddi_DdArray_t *array    /* array of BDDs */,
  int idx                 /* position of new element */,
  Ddi_Dd_t *f             /* BDD to be inserted */ )
{
  if (array==NULL)
    return 0;

  if (idx < 0) {
    fprintf(stderr, "Error : Access out of bounds\n");
    return 0;
  }
  if (idx >= array->nSize) {
    if (! DdArrayResize(array, idx + 1))
      return 0;
  }
  if (idx >= array->num)
    array->num = idx + 1;

  array->data[idx] = Ddi_DdDup(f);
  return 1;
}


/**Function*******************************************************************
  Synopsis    [Inserts an element at the end of array]
  Description [The function inserts the BDD <em>f</em> after the last BDD
               inserted in the array. The function returns 1 if successfully
               inserted, 0 otherwise.]
  SideEffects [none]
  SeeAlso     [Ddi_DdArrayInsert]
******************************************************************************/
int
Ddi_DdArrayInsertLast(
  Ddi_DdArray_t *array   /* BDDs'array */,
  Ddi_Dd_t *f            /* element to be inserted */ )
{
  return Ddi_DdArrayInsert(array,array->num,f);
}



/**Function*******************************************************************
  Synopsis    [Includes a BDD into an array at the i-th position.]
  Description [The BDD <em>f</em> is assigned to the <em>idx</em>
               position of the array and it is referenced. 
               All BDDs with higher indexes are moved upward.
               The function returns 1 for success, 0 otherwise]
  SideEffects [none]
  SeeAlso     [Ddi_DdArrayInsert]
******************************************************************************/
int
Ddi_DdArrayInclude(
  Ddi_DdArray_t *array    /* array of BDDs */,
  int idx                 /* position of new element */,
  Ddi_Dd_t *f             /* BDD to be included */ )
{
  int i;

  if (array==NULL)
    return 0;

  if (idx < 0) {
    fprintf(stderr, "Error : Access out of bounds\n");
    return 0;
  }

  /* if index of element to be included is greater than the greatest index */
  if (idx >= array->num)
    return Ddi_DdArrayInsert(array,idx,f);

  /* if there is not room to shift elements */
  if (array->num==array->nSize) {
    if (!DdArrayResize(array, array->num + 1))
      return 0;
  }

  array->num++;
  /* shift toward the left all the element after idx-th position */
  for (i=Ddi_DdArray_n(array)-1;i>=idx+1;i--)
    array->data[i]=array->data[i-1];

  array->data[idx] = Ddi_DdDup(f);  /* replace idx-th element */
  return 1;
}


/**Function*******************************************************************
  Synopsis    [Returns an element of the array]
  Description [The function returns the pointer to the BDD at the <em>i-th</em>
               position of the array. NULL is returned on failure.]
  SideEffects [none]
  SeeAlso     []
******************************************************************************/
Ddi_Dd_t *
Ddi_DdArrayFetch (
  Ddi_DdArray_t *array    /* BDDs' array */,
  int i                   /* index */ )
{
  if (array==NULL)
    return NULL;

  if ((i<0)||(i>array->nSize)) {
    fprintf(stderr, "Error : Access out of bounds\n");
    return NULL;
  }
  return array->data[i];
}


/**Function*******************************************************************
  Synopsis    [Extracts an element of the array]
  Description [The function returns the pointer to the BDD at the <em>i-th</em>
               position of the array, and replaces it with a NULL element.
               The BDD is thus removed from the array.
               If it fails, then returns NULL.]
  SideEffects [none]
  SeeAlso     []
******************************************************************************/
Ddi_Dd_t *
Ddi_DdArrayExtract (
  Ddi_DdArray_t *array    /* BDDs' array */,
  int i                   /* index */ )
{
  Ddi_Dd_t *F;

  if (array==NULL)
    return NULL;

  if ((i<0)||(i>=array->nSize)) {
    fprintf(stderr, "Error : Access out of bounds\n");
    return NULL;
  }
  F = array->data[i];
  array->data[i]=NULL;

  /* update last position of a element */
  if (i==array->num-1) {
    while (array->data[i]==NULL) i--;
    array->num = i+1;
  }
  return F;
}


/**Function*******************************************************************
  Synopsis    [Returns the number of elements of the array]
  SideEffects [none]
  SeeAlso     []
******************************************************************************/
int
Ddi_DdArray_n(
  Ddi_DdArray_t *array)
{
  if (array==NULL)
    return 0;
  return array->num;
}


/**Function*******************************************************************
  Synopsis    [Returns the last element of the array]
  Description [The function returns the pointer to the BDD at the last
               position of the array.]
  SideEffects [none]
  SeeAlso     [Ddi_DdArrayFetch]
******************************************************************************/
Ddi_Dd_t *
Ddi_DdArrayFetchLast(
  Ddi_DdArray_t *array)
{
  if (array==NULL)
    return NULL;
  return array->data[array->num-1];
}


/**Function*******************************************************************
  Synopsis    [Duplicates a BDD array.]
  Description [The function duplicates an array. 
               The unused room of the first array is not copied.
               On error, it returns NULL.]
  SideEffects [none]
  SeeAlso     []
******************************************************************************/
Ddi_DdArray_t *
Ddi_DdArrayDup(
  Ddi_DdArray_t *old  /* array to be duplicated */ )
{
  Ddi_DdArray_t *new;
  int i;

  new = Ddi_DdArrayAlloc (Ddi_DdArray_n(old));
  if (new == NULL) {
    fprintf(stderr ,"Error : Out of memory\n");
    return NULL;
  }

  for (i=0; i<Ddi_DdArray_n(new) ; i++) {
    new->data[i] = Ddi_DdDup(old->data[i]);
  }
  return new;
}


/**Function*******************************************************************
  Synopsis    [Copy a BDD array to a destination manager.]
  Description []
  SideEffects [none]
  SeeAlso     []
******************************************************************************/
Ddi_DdArray_t *
Ddi_DdArrayCopy(
  Ddi_DdMgr     *dd2 /* destination manager */,
  Ddi_DdArray_t *F  /* array to be duplicated */ )
{
  Ddi_DdArray_t *new;
  int i;

  new = Ddi_DdArrayAlloc (Ddi_DdArray_n(F));
  if (new == NULL) {
    fprintf(stderr ,"Error : Out of memory\n");
    return NULL;
  }

  for (i=0; i<Ddi_DdArray_n(new) ; i++) {
    new->data[i] = Ddi_DdCopy(dd2,F->data[i]);
  }
  return new;
}



/**Function*******************************************************************
  Synopsis    [Appends array2 at the end of array1]
  Description [The function appends the elements of array2 after the last
               element of array1. The function returns 1 if successfully
               appended, 0 otherwise.]
  SideEffects [none]
  SeeAlso     [Ddi_DdArrayJoin]
******************************************************************************/
int
Ddi_DdArrayAppend (
  Ddi_DdArray_t *array1   /* first array */,
  Ddi_DdArray_t *array2   /* array to be appended */ )
{
  int i;

  /* make sure array1 has enough room */
  if (array1->nSize < array1->num + array2->num) {
    if (! DdArrayResize(array1, array1->num + array2->num))
      return 0;
  }
  
  for (i=0; i<Ddi_DdArray_n(array2) ; i++)
    Ddi_DdArrayInsertLast(array1,array2->data[i]);
 
  return 1;
}


/**Function*******************************************************************
  Synopsis    [Joins two arrays]
  Description [The function returns the pointer of the new array if
               successfully joined, NULL otherwise]
  SideEffects [none]
  SeeAlso     [Ddi_DdArrayAppend]
*****************************************************************************/
Ddi_DdArray_t *
Ddi_DdArrayJoin(
  Ddi_DdArray_t *array1   /* first array */,
  Ddi_DdArray_t *array2   /* second array */ )
{
  Ddi_DdArray_t *array;   /* new array */
  int i;

  /* create a new array */
  array = Pdtutil_Alloc(Ddi_DdArray_t, 1);
  if (array == NULL) {
    fprintf(stderr ,"Error : Out of memory\n");
    return NULL;
  }

  array->num = array1->num + array2->num;
  array->nSize = array->num;
  array->data = Pdtutil_Alloc(Ddi_Dd_t *,array->nSize);
  if (array->data == NULL) {
    Pdtutil_Free(array);
    fprintf(stderr ,"Error : Out of memory\n");  
    return NULL;
  }

  /* copy array1 into array */
  for (i=0; i<Ddi_DdArray_n(array1) ; i++) {
    array->data[i] = Ddi_DdDup(array1->data[i]);
  }

  /* copy array2 into array */
  for (i=0; i<Ddi_DdArray_n(array2) ; i++) {
    array->data[array1->num + i] = Ddi_DdDup(array2->data[i]);
  }

  return array;
}



/**Function*******************************************************************
  Synopsis    [Converts Ddi_DdArray_t * into an array of Ddi_Dd_t **]
  Description [The function converts a DD array 
               (Ddi_DdArray_t *) into an array of BDDs (Ddi_Dd_t **).<BR>
               In case of error, it returns NULL.]
  SideEffects [none]
  SeeAlso     []
*****************************************************************************/
Ddi_Dd_t **
Ddi_DdArrayReadData (
  Ddi_DdArray_t * array )
{
  return (array->data);
}


/**Function*******************************************************************
  Synopsis    [Converts Ddi_DdArray_t * into an array of Ddi_Dd_t **]
  Description [The function converts a BDD array 
               (Ddi_DdArray_t *) into a BDDs'array (Ddi_Dd_t **).<BR>
               In case of error, it returns NULL.]
  SideEffects [none]
  SeeAlso     []
*****************************************************************************/
DdNode **
Ddi_DdArrayToCU (
  Ddi_DdArray_t *array)
{
  DdNode **data;
  int i;

  data = Pdtutil_Alloc(DdNode *,Ddi_DdArray_n(array));
  if (data == NULL) {
    fprintf(stderr ,"Error : Out of memory\n");
    return NULL;
  }
  for (i=0; i<array->num; i++)
    data[i] = Ddi_DdToCU(Ddi_DdArrayFetch(array,i));
  return data;
}

/**Function*******************************************************************
  Synopsis    [Converts Ddi_DdArray_t * into an array of Ddi_Dd_t **]
  Description [The function converts a BDD array 
               (Ddi_DdArray_t *) into a BDDs'array (Ddi_Dd_t **).<BR>
               In case of error, it returns NULL.]
  SideEffects [none]
  SeeAlso     []
*****************************************************************************/
Ddi_DdArray_t *
Ddi_DdArrayFromCU (
  Ddi_DdMgr *dd,
  DdNode **array,
  int n
)
{
  Ddi_DdArray_t *data;
  int i;

  data = Ddi_DdArrayAlloc(n);
  if (data == NULL) {
    fprintf(stderr ,"Error : Out of memory\n");
    return NULL;
  }
  for (i=0; i<n; i++)
    Ddi_DdArrayInsert(data,i,Ddi_DdCreateFromCU(dd,array[i])); 

  return data;
}


/**Function*******************************************************************
  Synopsis    [From the i0-th to last BDD in the array , take the constrain]
  Description [The function compute the constrain operator between <em>C</em>
               and each element of array from i0-th position to last position.
               Afterwards, the element of array are replaced with the result.]
  SideEffects [none]
  SeeAlso     [Ddi_DdArrayConstrain]
*****************************************************************************/
void
Ddi_DdArrayConstrainIthToN (
  Ddi_DdArray_t *array  /* BDDs'array */,
  Ddi_Dd_t *C     /* constrain */,
  int i0                /* first index of BDD to be constrained */,
  int n                 /* length of array */ )
{
 int i;
 Ddi_Dd_t *F,   /* an element of array */
          *FC;  /* element after constrain */

 for (i=0 ; i<i0+n ; i++) {
   if (i>=Ddi_DdArray_n(array))
     break;
   F = Ddi_DdArrayFetch (array,i);
   FC = Ddi_DdConstrain(F,C);
   Ddi_DdFree(F);
   Ddi_DdArrayInsert(array,i,FC);
 }
}


/**Function*******************************************************************
  Synopsis    [For each BDDs in the array takes the constrain]
  Description [The function computes the constrain operator between <em>C</em>
               and each element of array. 
               Afterwards, the element of array are replaced with the result.]
  SideEffects [none]
  SeeAlso     [Ddi_DdArrayConstrainIthToN]
*****************************************************************************/
void
Ddi_DdArrayConstrain (
  Ddi_DdArray_t *array  /* BDDs'array */,
  Ddi_Dd_t *C     /* constrain */ )
{
 Ddi_DdArrayConstrainIthToN (array,C,0,Ddi_DdArray_n(array));
}



/**Function*******************************************************************
  Synopsis    [From the i0-th to last BDD in the array , take the restrict]
  Description [The function compute the restrict operator between <em>C</em>
               and each element of array from i0-th position to last position.
               Afterwards, the element of array are replaced with the result.]
  SideEffects [none]
  SeeAlso     [Ddi_DdArrayRestrict]
*****************************************************************************/
void
Ddi_DdArrayRestrictIthToN (
  Ddi_DdArray_t *array, /* BDDs'array */
  Ddi_Dd_t *C,    /* constrain */
  int i0,               /* first index of BDD to be constrained */
  int n                 /* length of array */ )
{
 int i;
 Ddi_Dd_t *F,   /* an element of array */
	  *FC;  /* element after restrict */

 for (i=0 ; i<i0+n ; i++)
   {
    if (i>=Ddi_DdArray_n(array))
      break;
    F = Ddi_DdArrayFetch (array,i);
    FC = Ddi_DdRestrict(F,C);
    Ddi_DdFree(F);
    Ddi_DdArrayInsert(array,i,FC);
   }
}


/**Function*******************************************************************
  Synopsis    [For each BDDs in the array takes the restrict]
  Description [The function compute the restrict operator between <em>C</em>
               and each element of array. 
               Afterwards, the element of array are replaced with the result.]
  SideEffects [none]
  SeeAlso     [Ddi_DdArrayRestrictIthToN]
*****************************************************************************/
void
Ddi_DdArrayRestrict (
  Ddi_DdArray_t *array  /* BDDs'array */,
  Ddi_Dd_t *C     /* constrain */ )
{
 Ddi_DdArrayRestrictIthToN (array,C,0,Ddi_DdArray_n(array));
}



/**Function*******************************************************************
  Synopsis    [prints the BDDs'array]
  SideEffects [none]
  SeeAlso     []
*****************************************************************************/
void
Ddi_DdArrayPrint (
  Ddi_DdArray_t *array   /* array to be printed */ )
{
 int i;
 Ddi_Dd_t *F;

 for ( i=0; i<Ddi_DdArray_n(array); i++ ) {
   F = Ddi_DdArrayFetch ( array,i );
     Ddi_DdPrint(F);  
#if 0
     printf("%d ",Ddi_DdSize(F));
#endif
 }
}


/**Function*******************************************************************
  Synopsis    [Returns the numbers of nodes in a BDDs'array]
  Description [Counts the  numbers of nodes in an array of BDDs. Shared
               nodes are counted only once.]
  SideEffects [none]
  SeeAlso     []
*****************************************************************************/

int
Ddi_DdArraySize (
  Ddi_DdArray_t *array
  )
{
  int n;
  DdNode **data;

  data = Ddi_DdArrayToCU (array);
  if (data == NULL) {
    return (0);
  }

  n = Cudd_SharingSize (data, Ddi_DdArray_n (array));
  Pdtutil_Free (data);   

  return (n);
}


/**Function********************************************************************
  Synopsis     [Prints Statistics of a BDD]
  Description  []
  SideEffects  [none]
  SeeAlso      []
******************************************************************************/

void
Ddi_DdArrayPrintStats (
  Ddi_DdArray_t *array
)
{
  int size, componentSize, totalSize, i;
  Ddi_Dd_t *f;

  size = Ddi_DdArraySize (array);

  fprintf (stdout, "Number of Components:    %d\n", Ddi_DdArray_n (array));
  fprintf (stdout, "Sharing Size:    %d\n", size);
  fprintf (stdout, "Component Size:    ");
  totalSize = 0;
  for (i=0; i<Ddi_DdArray_n (array); i++) {
    f = Ddi_DdArrayFetch (array, i);
    componentSize = Ddi_DdSize (f);
    totalSize += componentSize;
    fprintf (stdout, "%d+", componentSize);
  }
  fprintf (stdout, "=%d\n", totalSize);

  return;
}

/**Function*******************************************************************
  Synopsis    [From the i0-th to last BDD in the array , computes AND]
  Description [The function compute the logic AND between <em>C</em> and
    each element of array from i0-th position to last position.
    Afterwards, the element of array are replaced with the result.]
  SideEffects [none]
  SeeAlso     [Ddi_DdArrayAnd]
*****************************************************************************/

void
Ddi_DdArrayAndIthToN (
  Ddi_DdArray_t *array   /* BDDs' array */,
  Ddi_Dd_t *c            /* Constrain */,
  int i0                 /* First index of BDDs' array */,
  int n                  /* Length of array */
  )
{
 int i;
 Ddi_Dd_t *f,      /* An element of array */
	  *fAndC;  /* Element after AND */

 for (i=0 ; i<i0+n ; i++) {
   if (i>=Ddi_DdArray_n (array)) {
     break;
   }

   f = Ddi_DdArrayFetch (array, i);
   fAndC = Ddi_DdAnd (f, c);
   Ddi_DdFree (f);
   Ddi_DdArrayInsert (array, i, fAndC);
 }

  return;
}


/**Function*******************************************************************
  Synopsis    [For each BDD in the array , computes AND]
  Description [The function compute the logic AND between <em>C</em> and
               each element of array.
               Afterwards, the element of array are replaced with the result.]
  SideEffects [none]
  SeeAlso     [Ddi_DdArrayAndIthToN]
*****************************************************************************/
void
Ddi_DdArrayAnd (
  Ddi_DdArray_t *array    /* array */,
  Ddi_Dd_t *C             /* constrain */ )
{
  Ddi_DdArrayAndIthToN (array,C,0,Ddi_DdArray_n(array));
}

/**Function*******************************************************************
  Synopsis    [From the i0-th to last BDD in the array , applies Exist]
  Description [The function abstracts existentially the variables belonging
               to the smoothing varset from each element of array 
               from i0-th position to the last.
               Afterwards, the element of array are replaced with the result.]
  SideEffects [none]
  SeeAlso     [Ddi_DdArrayAnd]
*****************************************************************************/
void
Ddi_DdArrayExistIthToN (
  Ddi_DdArray_t *array  /* BDDs'array */,
  Ddi_VarSet_t *V     /* smoothing varset */,
  int i0                /* first index of BDDs'array */,
  int n                 /* length of array */ )
{
 int i;
 Ddi_Dd_t *F,   /* an element of array */
	  *FV;  /* element after Exist */

 for (i=0 ; i<i0+n ; i++) {
   if (i>=Ddi_DdArray_n(array))
     break;
   F = Ddi_DdArrayFetch (array,i);
   FV = Ddi_DdExist(F,V);
   Ddi_DdFree(F);
   Ddi_DdArrayInsert(array,i,FV);
 }
}


/**Function*******************************************************************
  Synopsis    [For each BDD in the array , applies Exist]
  Description [The function abstracts existentially the variables belonging
               to the smoothing varset from each element of array 
               from i0-th position to the last.
               Afterwards, the element of array are replaced with the result.]
  SideEffects [none]
  SeeAlso     [Ddi_DdArrayAndIthToN]
*****************************************************************************/
void
Ddi_DdArrayExist (
  Ddi_DdArray_t *array    /* array */,
  Ddi_VarSet_t *V             /* smoothing varset */ )
{
  Ddi_DdArrayExistIthToN (array,V,0,Ddi_DdArray_n(array));
}




/**Function*******************************************************************
  Synopsis    [Computes OR among two BDDs]
  Description [The function, for all elements of the two arrays,
               computes logic OR between the <em>i-th</em>
               element of first array and the <em>i-th</em> element of
               second array.<br> The function returns the pointer of new
               array. Be careful! The two arrays must be same length.]
  SideEffects [none]
  SeeAlso     []
*****************************************************************************/
Ddi_DdArray_t *
Ddi_DdArrayOrDdArray (
  Ddi_DdArray_t *array1  /* first array */,
  Ddi_DdArray_t *array2  /* second array */ )
{
 int i;
 Ddi_DdArray_t *array3;
 Ddi_Dd_t *F,*G,*R;

 if (Ddi_DdArray_n(array1) != Ddi_DdArray_n(array2)) {
   fprintf(stderr, 
     "Error : OR among BDDs not definited for array of different length\n");
   return NULL;
 } 
 array3 = Ddi_DdArrayAlloc (Ddi_DdArray_n(array1));

 for (i=0 ; i<Ddi_DdArray_n(array1) ; i++) {
   F = Ddi_DdArrayFetch (array1,i);
   G = Ddi_DdArrayFetch (array2,i);
   R = Ddi_DdOr(F,G);
   Ddi_DdArrayInsert(array3,i,R);
 }
 return array3;
}


/**Function*******************************************************************
  Synopsis    [Writes array of BDDs in a dump file]
  Description [This function stores a BDDs'array.
               The BDD is stored in the DDDMP format. The parameter "mode" 
               can be DDDMP_MODE_TEXT, DDDMP_MODE_COMPRESSED or
               DDDMP_MODE_AUTOMATIC.<br>
               The function returns 1 if succefully stored, 0 otherwise.]
  SideEffects [none]
  SeeAlso     [Dddmp_cuddBddStore,Ddi_DdStore]
*****************************************************************************/
int
Ddi_DdArrayStore (
  Ddi_DdArray_t *array    /* array to be stored */,
  char *ddname            /* dd name (or NULL) */,
  char **vnames           /* array of variable names (or NULL) */,
  char **rnames           /* array of root names (or NULL) */,
  int *auxids             /* array of aux var ids (or NULL) */,
  int mode                /* storing mode selector */,
  char *fname             /* file name */,
  FILE *fp                /* pointer to the store file */ )
{
 DdNode **f;   /* array of BDD roots to be stored */
 int i;
 Ddi_DdMgr *dd;

 Pdtutil_Assert(Ddi_DdArray_n(array)>=1, "No partitions");
 dd = Ddi_DdReadMgr(Ddi_DdArrayFetch(array,0));    /* dd manager */
 for (i=0;i<Ddi_DdArray_n(array);i++)
   if (Ddi_DdReadMgr(Ddi_DdArrayFetch(array,i))!=dd) {
     printf ("single manager for a dd array still required!\n");
     return 0;
   }

 f = Ddi_DdArrayToCU(array);
 if (f==NULL)
   return 0;

 i = Dddmp_cuddBddArrayStore (dd->mgrCU,ddname,Ddi_DdArray_n(array),f,rnames,
       vnames,auxids,mode,DDDMP_VARIDS,fname,fp);

 Pdtutil_Free(f);   

 return i;
}


/**Function*******************************************************************
  Synopsis    [Reads array of BDDs from a dump file]
  Description [This function loads a BDDs'array.<br>
               The BDD on file must be in the DDDMP format. The parameter 
               "mode" can be DDDMP_MODE_TEXT, DDDMP_MODE_COMPRESSED or
               DDDMP_MODE_AUTOMATIC.<br>
               The function returns the pointer of array if succefully
               loaded, NULL otherwise.]
  SideEffects [none]
  SeeAlso     [Dddmp_cuddBddLoad,Ddi_DdLoad]
*****************************************************************************/
Ddi_DdArray_t *
Ddi_DdArrayLoad (
  Ddi_DdMgr *dd           /* dd manager */,
  char **vnames           /* variable names */,
  int *vauxids            /* variable auxids */,
  int mode                /* storing mode selector */,
  char *file              /* name file */,
  FILE *fp                /* file pointer */ )             
{
  DdNode   **roots;        /* array of BDD roots to be loaded */
  Ddi_DdArray_t *array;    /* descriptor of array */
  int nroots,              /* number of BDD roots */
      i;
 
  nroots = Dddmp_cuddBddArrayLoad(dd->mgrCU,DDDMP_ROOT_MATCHLIST,NULL,
             DDDMP_VAR_MATCHNAMES,vnames,vauxids,NULL,
             mode,file,fp,&roots);

  if (nroots<=0)
    return NULL;

  array = Ddi_DdArrayFromCU(dd,roots,nroots);
  if (array==NULL)
    return NULL;

  for (i=0;i<nroots;i++)
    Cudd_RecursiveDeref(dd->mgrCU,roots[i]);

  Pdtutil_Free(roots);   

  return array; 

}


/**Function*******************************************************************
  Synopsis    [Returns the support of BDDs'array from i0-th to last element]
  Description [The function computes the support of all BDDs from the
               <em>i-th</em> to last BDD of array.<br>
               The function return the support as a <em>varset</em> type.]
  SideEffects [none]
  SeeAlso     [Ddi_DdArraySupport]
*****************************************************************************/
Ddi_VarSet_t *
Ddi_DdArraySupportIthToN (
  Ddi_DdArray_t *array   /* BDDs'array */,
  int i0                 /* position of first BDD */
)
{
 int i;
 Ddi_VarSet_t *supp,*suppF,*tmp;
 Ddi_Dd_t *F;

 Ddi_DdMgr *dd = Ddi_DdReadMgr(Ddi_DdArrayFetch(array,0));

 supp = Ddi_VarSetVoid(dd);

 for (i=i0; i<Ddi_DdArray_n(array); i++) {
   F = Ddi_DdArrayFetch(array,i);    
   suppF = Ddi_DdSupport(F);    
   tmp = Ddi_VarSetUnion(supp,suppF); 
   Ddi_VarSetFree(supp);
   Ddi_VarSetFree(suppF);
   supp = tmp;
 }

 return supp;  
} 


/**Function*******************************************************************
  Synopsis    [Returns the support of BDDs'array]
  Description [The function returns a var-set representing the set of
               variables (=support) of all the BDDs in the array]
  SideEffects [none]
  SeeAlso     [Ddi_DdArraySupportArray]
*****************************************************************************/
Ddi_VarSet_t *
Ddi_DdArraySupport (
  Ddi_DdArray_t *array   /* BDDs'array */ )
{
  return Ddi_DdArraySupportIthToN(array,0);  
} 


/**Function*******************************************************************
  Synopsis    [Returns a vector of supports of BDDs'array]
  Description [The function return a pointer of an array of Ddi_VarSet_t * ,
               or NULL if it fails]
  SideEffects [none]
  SeeAlso     [Ddi_DdArraySupport]
*****************************************************************************/
Ddi_VarSet_t **
Ddi_DdArraySupportArray (
  Ddi_DdArray_t *fArray   /* array of function */ )
{
  Ddi_VarSet_t **fArraySupp;  /* array of function supports */
  Ddi_Dd_t *F;              /* function */
  Ddi_VarSet_t *S;        /* support of function */
  int i,n;                    

  n = Ddi_DdArray_n(fArray);    /* length of array */

  fArraySupp = Pdtutil_Alloc(Ddi_VarSet_t *,n);
  if (fArraySupp==NULL) {
    fprintf(stderr, "Error : Out of memory\n");
    return NULL;
  }

  for (i=0;i<n;i++) {
    F = Ddi_DdArrayFetch(fArray,i);
    S = Ddi_DdSupport(F);
    fArraySupp[i]=S;
  }

  return (fArraySupp); 
}


/*----------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*----------------------------------------------------------------------------*/


/**Function*******************************************************************
  Synopsis    [Resize a BDD array] 
  Description [The function returns 1 for success, 0 othewise.
               The minimum length of new array is twice the old one]
  SideEffects [none]
  SeeAlso     []
*****************************************************************************/
static int
DdArrayResize (
  Ddi_DdArray_t  *array      /* array to be resized */,
  int            nSizeNew    /* new size of array */ )
{
  int nSizeOld;
  Ddi_Dd_t **newdata;
  int i;

  nSizeOld = array->nSize;
  array->nSize = DDI_MAX(array->nSize * 2, nSizeNew);
  newdata = Pdtutil_Realloc(Ddi_Dd_t *, array->data, array->nSize);
  
  if (newdata == NULL) {
    array->nSize = nSizeOld;
    fprintf(stderr, "Error : Out of memory\n");
    return 0;
  } else {
    array->data = newdata;
  }
  
  for (i=nSizeOld; i<array->nSize ; i++)
     array->data[i] = NULL;

  return 1;
}

