/**CFile***********************************************************************

  FileName    [ddiVarArray.c]

  PackageName [ddi]

  Synopsis    [Functions to manage <em>arrays of variables</em>]

  Description [Array of variables is implemented as an array of BDD nodes
               (called "projection funtions" in CUDD package).<BR> 
               The basic manipulation allowed for variable arrays are:<BR> 
               <OL>
               <LI><B>Alloc</B> and <B>free</B> an array
               <LI><B>Insert</B> and <B>fetch</B> a variable
               <LI><B>Duplicate</B> an array
               <LI><B>Join</B> and <B>append</B> a new array
               <LI>Convert into array of integer
               <LI>Convert into a set of variables
               </OL>]

  SeeAlso     []  

  Author    [Gianpiero Cabodi and Stefano Quer]

  Copyright [This file was created at the Politecnico di Torino,
  Torino, Italy. 
  The  Politecnico di Torino makes no warranty about the suitability of 
  this software for any purpose.  
  It is presented on an AS IS basis.]
  
  Revision    []

******************************************************************************/

#include "ddiInt.h"

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */ 
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */ 
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */ 
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */ 
/*---------------------------------------------------------------------------*/

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static int VarArrayResize(Ddi_VarArray_t *array, int nSizeNew);

/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/

/**Function*******************************************************************
  Synopsis    [Generates a variable array from CU vars]
  Description []
  SideEffects [none]
  SeeAlso     []
*****************************************************************************/
Ddi_VarArray_t *
Ddi_VarArrayFromCU (
  DdNode **array,
  int n
)
{
  Ddi_VarArray_t *data;
  int i;

  data = Ddi_VarArrayAlloc(n);
  if (data == NULL) {
    fprintf(stderr, "Error : Out of memory\n");
    return NULL;
  }
  for (i=0; i<n; i++)
    Ddi_VarArrayInsert(data,i,Ddi_VarFromCU(array[i])); 

  return data;
}



/**Function*******************************************************************
  Synopsis    [Duplicates the data field of an array of variables]
  Description [The VarArray type is a struct pointing to the array of variables
               whose length can be read using Ddi_VarArray_n()]
  SideEffects [none]
  SeeAlso     []
*****************************************************************************/
DdNode **
Ddi_VarArrayToCU (
  Ddi_VarArray_t *array )
{
  DdNode **vett;
  int i;

  if (array==NULL)
    return NULL;
  vett = Pdtutil_Alloc(DdNode *,Ddi_VarArray_n(array));
  if (vett==NULL) {
    fprintf(stderr, "Error : Out of memory\n");
    return NULL;
  }

  for(i=0;i<Ddi_VarArray_n(array);i++) {
    vett[i] = Ddi_VarToCU(Ddi_VarArrayFetch(array,i));
  } 

  return vett;
}


/**Function*******************************************************************
  Synopsis    [Returns the data field of an array of variables]
  Description [The VarArray type is a struct pointing to the array of variables
               whose length can be read using Ddi_VarArray_n()]
  SideEffects [none]
  SeeAlso     []
*****************************************************************************/
Ddi_Var_t **
DdiVarArrayReadData (
  Ddi_VarArray_t *array )
{
  if (array==NULL)
    return NULL;

  return array->data;
}

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function*******************************************************************
  Synopsis    [Allocates a new array of variables of length "size"]
  Description [The function returns the pointer to the new array or NULL on error]
  SideEffects [none]
  SeeAlso     [Ddi_VarArrayFree]
******************************************************************************/
Ddi_VarArray_t *
Ddi_VarArrayAlloc (
  int size           /* array length */ )
{
  Ddi_VarArray_t *array;
  int i;

  array = Pdtutil_Alloc(Ddi_VarArray_t,1);
  if (array == NULL) {
    fprintf(stderr, "Error : Out of memory\n");
    return NULL;
  }  
  
  array->nSize = size;
  array->data = Pdtutil_Alloc(Ddi_Var_t *, array->nSize);
  if (array->data == NULL) {
    Pdtutil_Free(array);
    fprintf(stderr, "Error : Out of memory\n");
    return NULL;
  }

  for (i=0;i<array->nSize;i++) 
    array->data[i] = NULL; 
  return array;
}


/**Function*******************************************************************
  Synopsis    [Frees an array of variables]
  SideEffects [none]
  SeeAlso     [Ddi_VarArrayAlloc]
******************************************************************************/
void
Ddi_VarArrayFree (
  Ddi_VarArray_t *array )
{
  if (array==NULL)
    return;

  Pdtutil_Free(array->data); 
  Pdtutil_Free(array); 
}


/**Function*******************************************************************
  Synopsis    [Returns the number of elements in a variable array]
  SideEffects [none]
  SeeAlso     []
******************************************************************************/
int                      /* number of elements */
Ddi_VarArray_n(
  Ddi_VarArray_t *array
)
{
  if (array!=NULL)
    return array->nSize;
  return 0;
}


/**Function*******************************************************************
  Synopsis    [Inserts a variable into an array at a given position]
  Description [Returns 1 on success, 0 on failure. If the position 
               is greater than the array, the array is resized]
  SideEffects [none]
  SeeAlso     [Ddi_VarArrayFetch]
******************************************************************************/
int                        /* true or false */
Ddi_VarArrayInsert(
  Ddi_VarArray_t *array    /* array of variables */,
  int pos                  /* position of new element */,
  Ddi_Var_t *var         /* variable to be inserted */ )
{
  if (pos < 0) {
    fprintf(stderr, "Error : Access out of bounds\n");
    return 0;
  }

  if (pos >= array->nSize) {
    if (! VarArrayResize(array,pos+1))
      return 0;
  }
  
  array->data[pos] = var;
  return 1;
}


/**Function*******************************************************************
  Synopsis    [Returns a variable of the array at the i-th position]
  Description [The function returns NULL if the index is out of range]
  SideEffects [none]
  SeeAlso     [Ddi_VarArrayInsert]
******************************************************************************/
Ddi_Var_t *
Ddi_VarArrayFetch (
  Ddi_VarArray_t *array    /* variables'array */,
  int i                    /* position of variable */ )
{
  if ((i<0)||(i>=array->nSize)) {
    fprintf(stderr, "Error : Access out of bounds\n");
    return NULL;
  }
  return array->data[i];  
}


/**Function*******************************************************************
  Synopsis    [Duplicates an array of variables]
  Description [The function returns the pointer to the new array 
               or NULL if it fails]
  SideEffects [none]
  SeeAlso     []
******************************************************************************/

Ddi_VarArray_t *
Ddi_VarArrayDup(
  Ddi_VarArray_t *old  /* array to be duplicated */ )
{
  Ddi_VarArray_t *new;
  int i;

  new = Pdtutil_Alloc(Ddi_VarArray_t, 1);
  if (new == NULL) {
    fprintf(stderr, "Error : Out of memory\n");
    return NULL;
  }  
  
  new->nSize = old->nSize;
  new->data = Pdtutil_Alloc(Ddi_Var_t *, new->nSize);
  if (new->data == NULL) {
    Pdtutil_Free(new);
    fprintf(stderr, "Error : Out of memory\n");
    return NULL;
  }

  for (i=0; i<Ddi_VarArray_n(new) ; i++)
    new->data[i] = old->data[i];
  
  return new;
}


/**Function*******************************************************************
  Synopsis    [Duplicates an array of variables]
  Description [The function returns the pointer to the new array 
               or NULL if it fails]
  SideEffects [none]
  SeeAlso     []
******************************************************************************/

Ddi_VarArray_t *
Ddi_VarArrayCopy (
  Ddi_DdMgr *ddAux        /* dd Manager */,
  Ddi_VarArray_t  *vars   /* array of variables */
)
{
  Ddi_VarArray_t *new;
  Ddi_Var_t *v, *vnew;
  int i, nvars;

  nvars = Ddi_VarArray_n(vars);

  new = Ddi_VarArrayAlloc(nvars);

  for (i=0; i<nvars; i++) {
    v = Ddi_VarArrayFetch(vars,i);
    vnew = Ddi_VarCopy(ddAux,v);
    Ddi_VarArrayInsert(new,i,vnew);
  }

  return (new);
}




/**Function*******************************************************************
  Synopsis    [Appends the elements of array2 at the end of array1]
  Description [The function returns 1 if successfully appended, otherwise 0]
  SideEffects [none]
  SeeAlso     [Ddi_VarArrayJoin]
******************************************************************************/
int                        /* true or false */
Ddi_VarArrayAppend (
  Ddi_VarArray_t *array1   /* first array */,
  Ddi_VarArray_t *array2   /* array to be appended */ )
{
  int i;

  if (! VarArrayResize(array1, array1->nSize + array2->nSize))
    return 0;  
  
  for (i=0; i<Ddi_VarArray_n(array2) ; i++)
    array1->data[array1->nSize + i] = array2->data[i];
 
  array1->nSize += array2->nSize;
  return 1;
}


/**Function*******************************************************************
  Synopsis    [Joins two arrays, returning a new array]
  Description [Returns a pointer of the new array or NULL if it fails]
  SideEffects [none]
  SeeAlso     [Ddi_VarArrayAppend]
*****************************************************************************/
Ddi_VarArray_t *           /* new array */
Ddi_VarArrayJoin(
  Ddi_VarArray_t *array1,  /* first array */
  Ddi_VarArray_t *array2   /* second array */ )
{
  Ddi_VarArray_t *array;   /* new array */
  int i;

  array = Pdtutil_Alloc(Ddi_VarArray_t, 1);
  if (array == NULL) {
    fprintf(stderr, "Error : Out of memory\n");
    return NULL;
  }

  array->nSize = array1->nSize + array2->nSize;
  array->data = Pdtutil_Alloc(Ddi_Var_t *, array->nSize);
  if (array->data == NULL) {
    Pdtutil_Free(array);
    fprintf(stderr, "Error : Out of memory\n");
    return NULL;
  }

  /* copy first array */
  for (i=0; i<Ddi_VarArray_n(array1) ; i++)
     array->data[i] = array1->data[i];

  /* copy second array */
  for (i=0; i<Ddi_VarArray_n(array2) ; i++)
     array->data[array1->nSize + i] = array2->data[i];

  return array;
}



/**Function*******************************************************************
  Synopsis    [Returns the indices of variables]
  Description [The function returns a pointer to an array of integers, 
               where each element is the index of a variable in VarArray.
               If it fails, then return NULL.]
  SideEffects [none]
  SeeAlso     [Ddi_VarArrayToVarSet]
*****************************************************************************/
int *                    /* array of integer */
Ddi_VarArrayToInt (
  Ddi_VarArray_t *array  /* array of variables */ )
{
  int i;
  int *vett;
  Ddi_Var_t *var;

  vett = Pdtutil_Alloc(int,Ddi_VarArray_n(array));
  if (vett==NULL) {
    fprintf(stderr, "Error : Out of memory\n");
    return NULL;
  }

  for (i=0; i<Ddi_VarArray_n(array); i++) {
    var = Ddi_VarArrayFetch(array,i);
    vett[i] = Ddi_VarIndex(var);
  }
 return vett;
}


/**Function*******************************************************************
  Synopsis    [Converts an array of variables into a set of variables]
  Description [The function returns a pointer to a set of variables.
               From array to set of variables the information about ordering 
               is lost.]
  SideEffects [none]
  SeeAlso     [Ddi_VarArrayToInt]
*****************************************************************************/
Ddi_VarSet_t *         /* set of variables */
Ddi_VarArrayToVarSet (
  Ddi_DdMgr *dd          /* dd manager */,
  Ddi_VarArray_t *array  /* array of variables */ )
{
  Ddi_VarSet_t *varset;
  Ddi_Var_t *var;
  int i;

  varset = Ddi_VarSetVoid(dd);

  for(i=0;i<Ddi_VarArray_n(array);i++) {
    var = Ddi_VarArrayFetch(array,i);
    varset = Ddi_VarSetAdd(varset,var);
  }
  return varset;
}

/**Function*******************************************************************
  Synopsis    [Returns the index of the i-th element of the array]
  Description [The function is equivalent at the call of the instruction
               below:<BR>
               Ddi_VarIndex( Ddi_VarArrayFetch(array,i) );<BR>
               If the index i is out of bounds, then return -1]
  SideEffects [none]
  SeeAlso     [Ddi_VarArrayFetch,Ddi_VarIndex]
*****************************************************************************/
int                      /* index of variable */
Ddi_VarArrayIndex (
  Ddi_VarArray_t *array  /* array of variables */,
  int i                  /* position of variable */ )
{
  Ddi_Var_t *var;
 
  var = Ddi_VarArrayFetch(array,i);
  if (var==NULL) 
    return -1;
  return Ddi_VarIndex(var);
}

/*---------------------------------------------------------------------------*/
/* Definition of static functions                                          */
/*---------------------------------------------------------------------------*/

/**Function*******************************************************************
  Synopsis    [Resizes an array]
  Description [The function returns 1 if successfully resized, otherwise 0]
  SideEffects [none]
  SeeAlso     []
*****************************************************************************/
static int                   /* true or false */
VarArrayResize (
  Ddi_VarArray_t  *array,    /* array to be resized */
  int             nSizeNew   /* nre size of array */ )
{
  int nSizeOld;
  Ddi_Var_t **newdata;
  int i;

  nSizeOld = array->nSize;
  array->nSize = nSizeNew;
  newdata = Pdtutil_Realloc(Ddi_Var_t *,array->data, array->nSize);
  if (newdata == NULL) {
    array->nSize = nSizeOld;
    fprintf(stderr, "Error : Out of memory\n");
    return 0;
  } else {
    array->data = newdata;
  }
  
  for (i=0; i< (array->nSize-nSizeOld) ; i++)
     array->data[nSizeOld+i] = NULL;

  return 1;
}

