/**CFile***********************************************************************

  FileName    [ddiVarSet.c]

  PackageName [ddi] 

  Synopsis    [Functions to manage <em>sets of variables</em>]

  Description [Sets of variables (var-set) are implemented as cubes.<BR>
               The basic manipulation allowed on a var-set are:<BR> 
               <OL>
               <LI>Create an empty var-set (<B>New</B>) 
	             and destroy one (<B>Free</B>)
               <LI><B>Add</B> a new variable or <B>remove</B> one
               <LI><B>Dup</B>licate a var-set
               <LI><B>Union</B>, <B>intersect</B> and <B>difference</B>
                   of two var-sets
               <LI>Other specific operations
               </OL>]

  SeeAlso     []  

  Author    [Gianpiero Cabodi and Stefano Quer]

  Copyright [This file was created at the Politecnico di Torino,
  Torino, Italy. 
  The  Politecnico di Torino makes no warranty about the suitability of 
  this software for any purpose.  
  It is presented on an AS IS basis.]
  
  Revision    []

******************************************************************************/

#include "ddiInt.h"


/*---------------------------------------------------------------------------*/
/* Structure declaration                                                     */ 
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declaration                                                          */ 
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declaration                                                      */ 
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declaration                                                         */ 
/*---------------------------------------------------------------------------*/

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/


/**AutomaticEnd***************************************************************/

/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/
	
/*---------------------------------------------------------------------------*/
/* Definition of exported function                                           */ 
/*---------------------------------------------------------------------------*/


/**Function********************************************************************
 Synopsis    [Returns an empty var-set]
 Description [Returns a pointer to an empty var-set]
 SideEffects [none]
 SeeAlso     [Ddi_VarSetFree]
******************************************************************************/
Ddi_VarSet_t *
Ddi_VarSetVoid (
  Ddi_DdMgr *dd      /* dd manager */ )
{
  return ((Ddi_VarSet_t *)Ddi_MgrReadOne(dd));  /* the constant "one" */
}


/**Function********************************************************************
 Synopsis    [Frees a var-set, dereferencing all BDD nodes]
 SideEffects [none]
 SeeAlso     [Ddi_VarSetNew]
******************************************************************************/
void
Ddi_VarSetFree (
  Ddi_VarSet_t *src      /* varset to be freed */ )
{
  Ddi_DdFree((Ddi_Dd_t *)src);
}


/**Function********************************************************************
 Synopsis    [The function returns 1 if the var-set is empty]
 SideEffects [none]
 SeeAlso     []
******************************************************************************/
int
Ddi_VarSetIsVoid (
  Ddi_VarSet_t *varset )
{
  return Ddi_DdIsConstant((Ddi_Dd_t *)varset);
}


/**Function********************************************************************
 Synopsis    [The function returns a var-set excluding the top variable]
 SideEffects [none]
 SeeAlso     []
******************************************************************************/
Ddi_VarSet_t *
Ddi_VarSetNext (
  Ddi_VarSet_t *varset )
{
  return (Ddi_VarSet_t *)Ddi_DdCreateFromCU(
    ((Ddi_Dd_t *)varset)->mgr,Cudd_T(Ddi_VarSetToCU(varset))
  ); 
}


/**Function********************************************************************
 Synopsis    [Adds a variable to the var-set]
 Description [The function return the pointer to the new var-set]
 SideEffects [none]
 SeeAlso     [Ddi_VarSetRemove]
******************************************************************************/
Ddi_VarSet_t *
Ddi_VarSetAdd (
  Ddi_VarSet_t *varset   /* var-set */,
  Ddi_Var_t *var         /* the new variable added to var-set */ )
{
  Ddi_VarSet_t *result;
  Ddi_Dd_t *tmp;
  Ddi_DdMgr *dd=((Ddi_Dd_t *)varset)->mgr;
                                                       
  result = (Ddi_VarSet_t *)
    Ddi_DdAnd((Ddi_Dd_t *)varset,(tmp=Ddi_DdLiteral(dd,var)));
  Ddi_DdFree(tmp);
  return result;
}


/**Function********************************************************************
 Synopsis    [Copies a var-set into another one]
 Description [The function returns the new pointer of the var-set]
 SideEffects [none]
 SeeAlso     []
******************************************************************************/
Ddi_VarSet_t *
Ddi_VarSetDup (  
  Ddi_VarSet_t *src   /* var-set to be copied */ )
{
  return ((Ddi_VarSet_t *)Ddi_DdDup((Ddi_Dd_t *)src));
}


/**Function********************************************************************
 Synopsis    [Copies a var-set to another manager]
 Description []
 SideEffects [none]
 SeeAlso     []
******************************************************************************/
Ddi_VarSet_t *
Ddi_VarSetCopy (  
  Ddi_DdMgr *dd2 /* destination manager */,
  Ddi_VarSet_t *src   /* var-set to be copied */ 
)
{
  return ((Ddi_VarSet_t *)Ddi_DdCopy(dd2,(Ddi_Dd_t *)src));
}

/**Function********************************************************************
  Synopsis     [Evaluates an expression and frees first argument]
  SideEffects  [none]
  SeeAlso      []
******************************************************************************/
Ddi_VarSet_t *
Ddi_VarSetEvalFree(
  Ddi_VarSet_t *f  /* expression */,
  Ddi_VarSet_t *g  /* varset to be freed */
)
{
 Ddi_VarSetFree(g);
 return f;
}


/**Function********************************************************************
 Synopsis    [Returns the number of variables in the varset]
 SideEffects [none]
 SeeAlso     []
******************************************************************************/
int
Ddi_VarSet_n (
  Ddi_VarSet_t *vars  /* varset */ )
{
  int count = 0;
  Ddi_VarSet_t *aux, *src;

  src = Ddi_VarSetDup(vars);
  while ( !Ddi_VarSetIsVoid(src) ) {
    aux = Ddi_VarSetNext(src);  
    Ddi_VarSetFree(src);
    src = aux;
    count++;
  }
 return count;
}

/**Function********************************************************************
 Synopsis    [Prints the varset]
 SideEffects [none]
 SeeAlso     []
******************************************************************************/
void
Ddi_VarSetPrint (
  Ddi_VarSet_t *vars  /* varset */ )
{
  Ddi_VarSet_t *aux, *src;
  Ddi_Var_t *var;
  char *name;
  Ddi_DdMgr *dd = Ddi_DdReadMgr(vars);    /* dd manager */
  int i;
  char *buf;

  buf = Pdtutil_Alloc(char,(Ddi_ReadNumVars(dd)+1));
  if (buf == NULL) {
    printf("Unable to allocate memory for varset print\n");
    return;
  }
  for (i=0;i<Ddi_ReadNumVars(dd);i++)
    buf[i]='.';
  buf[i]='\0';

  src = Ddi_VarSetDup(vars);
  while ( !Ddi_VarSetIsVoid(src) ) {
    var = Ddi_VarSetTop (src);
    if ((name = Ddi_VarId2Name(dd,Ddi_VarIndex(var))) != NULL)
      printf ("%s ", name);
    else 
      printf ("<ID:%d> ", Ddi_VarIndex(var));
    buf[Ddi_VarIndex(var)]='*';
    aux = Ddi_VarSetNext(src); 
    Ddi_VarSetFree(src);
    src = aux;
  }

printf("\n%s\n",buf);

Pdtutil_Free(buf);

}


/**Function********************************************************************
 Synopsis    [Computes the union of two var-sets]
 Description [The function returns the pointer to the union of two var-sets]
 SideEffects [none]
 SeeAlso     [Ddi_VarSetIntersect,Ddi_VarSetDifference]
******************************************************************************/
Ddi_VarSet_t *
Ddi_VarSetUnion (
  Ddi_VarSet_t *v1    /* first var-set */ ,
  Ddi_VarSet_t *v2    /* second var-set */ )
{
  return ((Ddi_VarSet_t *)Ddi_DdAnd((Ddi_Dd_t *)v1,(Ddi_Dd_t *)v2));
}


/**Function********************************************************************
 Synopsis    [Computes the intersection of two var-sets]
 Description [The function return the new pointer of intersection of two var-sets]
 SideEffects [none]
 SeeAlso     [Ddi_VarSetUnion,Ddi_VarSetDifference]
******************************************************************************/
Ddi_VarSet_t *
Ddi_VarSetIntersect (
  Ddi_VarSet_t *varset1   /* first var-set */,
  Ddi_VarSet_t *varset2   /* second var-set */ )
{
  Ddi_VarSet_t *dest;
  Ddi_DdMgr *dd=((Ddi_Dd_t *)varset1)->mgr;

  dest = (Ddi_VarSet_t *)Ddi_DdCreateFromCU(dd,
            Cudd_bddLiteralSetIntersection(dd->mgrCU,
            Ddi_VarSetToCU(varset1),Ddi_VarSetToCU(varset2))
         );
  return dest;
}


/**Function********************************************************************
 Synopsis    [Computes the difference of two var-sets]
 Description [The difference of two var-sets is defined as the union of first
              var-set with the negate of second.<BR>
              It is equivalent to exclude a group of variable from var-set.]
 SideEffects [none]
 SeeAlso     [Ddi_VarSetIntersect,Ddi_VarSetUnion]
******************************************************************************/
Ddi_VarSet_t *
Ddi_VarSetDifference (
  Ddi_VarSet_t *varset    /* var-set */,
  Ddi_VarSet_t *vars      /* excluded variables */ )
{
  return ((Ddi_VarSet_t *)Ddi_DdExist((Ddi_Dd_t *)varset,(Ddi_Dd_t *)vars));
}


/**Function********************************************************************
 Synopsis    [Removes a variable from the var-set]
 Description [The function return a new pointer of var-set without the
	      removed variables]
 SideEffects [none]
 SeeAlso     [Ddi_VarSetAdd]
******************************************************************************/
Ddi_VarSet_t *
Ddi_VarSetRemove (
  Ddi_VarSet_t *varset /* var-set */,
  Ddi_Var_t    *var    /* variable to be removed */ )
{
  Ddi_Dd_t *tmp, *result;
  Ddi_DdMgr *dd=((Ddi_Dd_t *)varset)->mgr;

  result = (Ddi_VarSet_t *)Ddi_DdExist((Ddi_Dd_t *)varset,
          (tmp=Ddi_DdLiteral(dd,var)));
  Ddi_DdFree(tmp);

  return result;
}


/**Function********************************************************************
 Synopsis    [The function returns 1 if the two var-sets are equal]
 SideEffects [none]
 SeeAlso     []
******************************************************************************/
int
Ddi_VarSetEqual (
  Ddi_VarSet_t *varset1  /* first var-set */,
  Ddi_VarSet_t *varset2  /* second var-set */)
{
  return (Ddi_VarSetToCU(varset1)==
          Ddi_VarSetToCU(varset2));
}


/**Function********************************************************************
 Synopsis    [Returns the top variable in the ordering]
 SideEffects [none]
 SeeAlso     []
******************************************************************************/
Ddi_Var_t *
Ddi_VarSetTop (
  Ddi_VarSet_t *varset   /* var-set */ )
{
  int index;
  Ddi_DdMgr *dd=((Ddi_Dd_t *)varset)->mgr;

  if (Ddi_VarSetIsVoid (varset))
    return (NULL);

  index = Ddi_DdTopvarIndex((Ddi_Dd_t *)varset);
  return Ddi_IthVar(dd,index);
}


/**Function********************************************************************
 Synopsis    [Returns the bottom variable in the ordering]
 SideEffects [none]
 SeeAlso     []
******************************************************************************/
Ddi_Var_t *
Ddi_VarSetBottom (
  Ddi_VarSet_t *varset   /* var-set */ )
{
  int index;
  Ddi_VarSet_t *prev, *next;
  Ddi_DdMgr *dd=((Ddi_Dd_t *)varset)->mgr;

  if (Ddi_VarSetIsVoid (varset))
    return (NULL);

  prev = Ddi_VarSetDup(varset);
  next = Ddi_VarSetNext(prev);  
  while ( !Ddi_VarSetIsVoid(next) ) {
    Ddi_VarSetFree(prev);
    prev = next;
    next = Ddi_VarSetNext(prev);  
  }

  index = Ddi_DdTopvarIndex((Ddi_Dd_t *)prev);
  Ddi_VarSetFree(prev);
  Ddi_VarSetFree(next);

  return Ddi_IthVar(dd,index);
}


/**Function********************************************************************
  Synopsis    [Converts a varset into the correspondent Cudd Node]
  Description []
  SideEffects [None]
  SeeAlso     []
******************************************************************************/
Ddi_DdNode *
Ddi_VarSetToCU(
  Ddi_VarSet_t *v         
)
{
  Pdtutil_Assert(v->code == DDI_MONO_BOOLF, "Monolithic BDD expected");

  return ((v)->boolf.mono);
}

/**Function********************************************************************
  Synopsis    [Swaps two sets of variables in a varset]
  Description [Returns a new varset with the variables swapped 
               (x replaces y and viceversa.]
  SideEffects [none]
  SeeAlso     []
******************************************************************************/
Ddi_VarSet_t *
Ddi_VarSetSwapVariables(
  Ddi_VarSet_t   *f    /* input varset */,
  Ddi_VarArray_t *x    /* 1-st array of variables */,
  Ddi_VarArray_t *y    /* 2-nd array of variables */ )
{
  return ((Ddi_VarSet_t *)Ddi_DdSwapVariables((Ddi_Dd_t *)f,x,y));
}
