/**CFile***********************************************************************

  FileName    [fsmPort.c]

  PackageName [fsm]

  Synopsis    [Functions to guarantee portability]

  Description [In this file are declarated function to deal with 
    BLIF format (as package Nanotrav in CUDD).
    ]

  SeeAlso     []  

  Author    [Gianpiero Cabodi and Stefano Quer]

  Copyright [This file was created at the Politecnico di Torino,
    Torino, Italy. 
    The  Politecnico di Torino makes no warranty about the suitability of 
    this software for any purpose.  
    It is presented on an AS IS basis.]
  
  Revision    []

******************************************************************************/

#include "fsmInt.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static FsmPortNtrOptions_t * mainInit();

/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of external functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Reads description of a fsmMgr (BLIF format) from file]

  Description [The function gives a compatible DDI:
    <ul>
    <li>number of primary input
    <li>number of latches
    <li>array of primary input variables
    <li>array of present state variables
    <li>array of next state variables
    <li>array of next state functions
    <li>array of partitioned transition relation
    </ul>]

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

int
Fsm_MgrLoadFromBlif (
  Fsm_Mgr_t **fsmMgrP     /* FSM Pointer */,
  Ddi_DdMgr *dd           /* dd manager */,
  char *fileFsmName       /* FSM File Name */,
  char *fileOrdName       /* ORD File Name */,
  int bddFlag             /* Not Used For Now */
  )
{
  Fsm_Mgr_t *fsmMgr;
  FsmPortBnetNetwork_t *net;    /* network */
  FsmPortNtrOptions_t *option;  /* options */
  FILE *fp = NULL;
  Ddi_Dd_t *temp;
  Ddi_Var_t *var;
  int i,result;
  int xlevel;
  FsmPortBnetNode_t *node;
  char *name, *nameext;
  int index, ntotv, value;
  long fsmTime;

  fsmTime = util_cpu_time();

  /*-------------------------- Check For FSM File  --------------------------*/

  nameext = Pdtutil_FileName (fileFsmName, "", "blif", 0);
  fp = Pdtutil_OpenFile (fp, nameext, "r");
  if (fp==NULL) {
    Pdtutil_Warning (1, "NULL Finite State Machine Pointer.");
    return (1);
  }

  /*------------------------- FSM Manager Allocation ------------------------*/

  fsmMgr = *fsmMgrP;
  if (fsmMgr == NULL) {
    Pdtutil_Warning (1, "Allocate a new FSM structure");
    fsmMgr = Fsm_MgrInit (NULL);
  }

  if (fsmMgr->settings.verbosity > Pdtutil_VerbLevelDevMin_c) {
    fprintf (stdout, "(... loading %s ...)", fileFsmName) ;
    fflush (stdout);
  }

  /*------------------------------ Init Options -----------------------------*/

  option = mainInit();

  if (fileOrdName != NULL) {
    if (strcmp (fileOrdName, "file") == 0) {
      /* Enable Using Order From Blif File (default) */
      option->ordering = PI_PS_FROM_FILE;
    } else {
      if (strcmp (fileOrdName, "dfs") == 0) {
        /* Enable Using Depth First Search */
        option->ordering = PI_PS_DFS;
      } else {
        /* Enable Reading Order From File */
        option->ordering = PI_PS_GIVEN;
        option->orderPiPs = Pdtutil_StrDup (fileOrdName);
      }
    }
  }

  /*------------------ Load description of the FSM Manager ------------------*/

  net = Fsm_PortBnetReadNetwork (fp, option->verb);  
  Pdtutil_Warning (net==NULL, "Error : Description of fsmMgr not loaded.\n");
  if (net==NULL) {
    return (1);
  }

  Pdtutil_CloseFile (fp);

  /*-------------- Build the BDDs for the Nodes of the Network --------------*/

  result = Fsm_PortNtrBuildDDs (net, Ddi_MgrReadMgrCU(dd), option, NULL);
  Pdtutil_Warning (result==0, "Impossible to build the BDDs for the network");
  if (result == 0) {
    return (1);
  }

  i =  Ddi_CheckZeroRef(dd);
  if (i!=0) {
    fprintf(stderr, "%d non-zero DD reference counts after building DDs\n",i);
  }

  /*------- Translate CUDD format in our internal FSM Manager Format --------*/
  
  Fsm_MgrSetDdManager (fsmMgr, dd);

  /* Number of State Variable */
  Fsm_MgrSetNL (fsmMgr, net->nlatches);
  /* Number of Primary Input */
  Fsm_MgrSetNI (fsmMgr, net->npis);
  /* Number of Primary Outputs */
  Fsm_MgrSetNO (fsmMgr, net->npos);

  /* BDD of Initial State Set */
  fsmMgr->init.bdd =
    Ddi_DdCreateFromCU (dd, Fsm_PortNtrInitState (Ddi_MgrReadMgrCU (dd),
      net, option));

  /* BDD of Reached State Set */
  fsmMgr->reached.bdd = Ddi_DdDup (fsmMgr->init.bdd);

  /*
   *  Init Temporary Set and Array
   */

  ntotv = Ddi_ReadSize(dd)+2*Fsm_MgrReadNL (fsmMgr);

  fsmMgr->var.i = Ddi_VarArrayAlloc (Fsm_MgrReadNI (fsmMgr));
  fsmMgr->var.ps = Ddi_VarArrayAlloc (Fsm_MgrReadNL (fsmMgr));
  fsmMgr->var.ns = Ddi_VarArrayAlloc (Fsm_MgrReadNL (fsmMgr));
  fsmMgr->delta.bdd = Ddi_DdArrayAlloc (Fsm_MgrReadNL (fsmMgr));
  fsmMgr->lambda.bdd = Ddi_DdArrayAlloc (Fsm_MgrReadNO (fsmMgr));
  fsmMgr->varnames = Pdtutil_Alloc (char *,ntotv); 
  fsmMgr->varauxids = Pdtutil_Alloc (int, ntotv);

  for (i=0; i<ntotv; i++) {
    fsmMgr->varnames[i] = NULL;
    fsmMgr->varauxids[i] = (-1);
  }

  /*------------------ Translate Array of Input Variables -------------------*/

  for (i=0; i<net->npis; i++) {
    if (!st_lookup(net->hash,net->inputs[i],(char **)&node)) { 
      return (1);
    }

    Cudd_Ref (node->dd);
    var = Ddi_VarFromCU (node->dd);
    Ddi_VarArrayInsert (fsmMgr->var.i, i, var);

    index = Ddi_VarIndex(var);
    fsmMgr->varauxids[index] = i;
    name = net->inputs[i];
    fsmMgr->varnames[index] = Pdtutil_StrDup (name);
  }

  /*-------- Translate Array of Present State Variables and Delta -----------*/

  for (i=0; i<net->nlatches; i++) {
    if (!st_lookup(net->hash,net->latches[i][1],(char **)&node)) {
      return (1);
    }

    Cudd_Ref (node->dd);
    var = Ddi_VarFromCU (node->dd);
    Ddi_VarArrayInsert (fsmMgr->var.ps, i, var);

    index = Ddi_VarIndex (var);
    fsmMgr->varauxids[index] = i+net->npis;
    fsmMgr->varnames[index] = Pdtutil_StrDup (net->latches[i][1]);

    /* create array of next state var. */
    xlevel = Ddi_VarCurrPos(dd,var)+1;
#if 1
    var = Ddi_DdNewVarAtLevel(dd,xlevel);
#else
    var = Ddi_DdNewVar(dd);
#endif
    Ddi_VarArrayInsert (fsmMgr->var.ns, i, var);

    index = var->index;
    fsmMgr->varauxids[index] = i+net->npis+net->nlatches;
    fsmMgr->varnames[index] = Pdtutil_Alloc (char,
      strlen(net->latches[i][1])+4);
    sprintf (fsmMgr->varnames[index],"%s$NS",net->latches[i][1]);

    /* translate array of next state function (delta) */

    if (!st_lookup(net->hash,net->latches[i][0],(char **)&node)) {
      return (1);
    }

    temp = Ddi_DdCreateFromCU (dd,node->dd);
    Ddi_DdArrayInsert (fsmMgr->delta.bdd, i, temp);
    Ddi_DdFree (temp);
  }

  /*---------------- Translate Array of Output Function ---------------------*/

  for (i=0;i<Fsm_MgrReadNO (fsmMgr);i++) {
    if (!st_lookup(net->hash,net->outputs[i],(char **)&node)) { 
      return (1);
    }
    
    temp = Ddi_DdCreateFromCU (dd,node->dd);
    Ddi_DdArrayInsert (fsmMgr->lambda.bdd, i, temp); 
    Ddi_DdFree (temp);
  } 
   
  /* create variable array of PI+PS */
  fsmMgr->sx =
    Ddi_VarArrayJoin (Fsm_MgrReadVarPS (fsmMgr),Fsm_MgrReadVarI (fsmMgr));

  /* Dispose of network */
  for (node = net->nodes; node != NULL; node = node->next) {
    if (node->dd != NULL) {
      Cudd_RecursiveDeref(Ddi_MgrReadMgrCU(dd),node->dd);
    }
  }

  Fsm_PortBnetFreeNetwork (net);

  if (Fsm_MgrReadVarnames (fsmMgr) != NULL) {
    fsmMgr->name.i = Pdtutil_Alloc (char *,Fsm_MgrReadNI (fsmMgr));
    fsmMgr->index.i  = Pdtutil_Alloc (int,Fsm_MgrReadNI (fsmMgr));
    for (i=0; i<Fsm_MgrReadNI (fsmMgr); i++) {
      value = Ddi_VarIndex (Ddi_VarArrayFetch
        (Fsm_MgrReadVarI (fsmMgr), i));
      fsmMgr->name.i[i] = Pdtutil_StrDup (fsmMgr->varnames[value]);
      fsmMgr->index.i[i] = fsmMgr->varauxids[value];
    }

    fsmMgr->name.ps = Pdtutil_Alloc (char *, Fsm_MgrReadNL (fsmMgr));
    fsmMgr->index.ps = Pdtutil_Alloc (int, Fsm_MgrReadNL (fsmMgr));
    fsmMgr->name.ns = Pdtutil_Alloc (char *, Fsm_MgrReadNL (fsmMgr));
    fsmMgr->index.ns =  Pdtutil_Alloc (int, Fsm_MgrReadNL (fsmMgr));

    for (i=0; i<Fsm_MgrReadNL (fsmMgr); i++) {
      value = Ddi_VarIndex (Ddi_VarArrayFetch (
        Fsm_MgrReadVarPS (fsmMgr), i));
      fsmMgr->name.ps[i] = Pdtutil_StrDup (fsmMgr->varnames[value]);
      fsmMgr->index.ps[i] = fsmMgr->varauxids[value];

      value = Ddi_VarIndex(Ddi_VarArrayFetch (
        Fsm_MgrReadVarNS (fsmMgr),i));
      fsmMgr->name.ns[i] = Pdtutil_StrDup (fsmMgr->varnames[value]);
      fsmMgr->index.ns[i] = fsmMgr->varauxids[value];
    }

    for (i=0; i<Ddi_ReadNumVars(dd); i++) {
      Ddi_VarSetName(dd,Ddi_IthVar(dd,i),fsmMgr->varnames[i]);
      Ddi_VarSetAuxid(dd,Ddi_IthVar(dd,i),fsmMgr->varauxids[i]);
    }

  }

  /*
   *  return 1 = Something went wrong
   */

  if (Fsm_MgrReadNL (fsmMgr) == 0 || Fsm_MgrReadNI (fsmMgr) == 0
    || Fsm_MgrReadNO (fsmMgr) == 0) {
    Fsm_MgrFree(fsmMgr);
    return (1);  
  }

  /*
   *  return 0 = All OK
   */

  fsmTime = util_cpu_time() - fsmTime;
  fsmMgr->fsmTime += fsmTime;

  *fsmMgrP = fsmMgr;
  return (0);
}

/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                            */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Allocates the option structure for BLIF format and initializes
    it.]

  Description [This function was taken from main.c in <b>nanotrav</b> package]

  SideEffects [none]

  SeeAlso     [Fsm_MgrLoadFromBlif]

******************************************************************************/

static FsmPortNtrOptions_t *
  mainInit (
  )
{
  FsmPortNtrOptions_t	*option;

  /* Initialize option structure. */
  option = ALLOC(FsmPortNtrOptions_t,1);

  option->initialTime    = util_cpu_time();
  option->verify         = FALSE;
  option->second         = FALSE;
  option->file1          = NULL;
  option->file2          = NULL;
  option->traverse       = FALSE;
  option->depend         = FALSE;
  option->image          = PORT_NTRIMAGE_MONO;
  option->imageClip      = 1.0;
  option->approx         = PORT_NTRUNDER_APPROX;
  option->threshold      = -1;
  option->from           = PORT_NTRFROM_NEW;
  option->groupnsps      = PORT_NTRGROUP_NONE;
  option->closure        = FALSE;
  option->closureClip    = 1.0;
  option->envelope       = FALSE;
  option->scc            = FALSE;
  option->maxflow        = FALSE;
  option->zddtest        = FALSE;
  option->sinkfile       = NULL;
  option->partition      = FALSE;
  option->char2vect      = FALSE;
  option->density        = FALSE;
  option->quality        = 1.0;
  option->decomp         = FALSE;
  option->cofest         = FALSE;
  option->clip           = -1.0;
  option->noBuild        = FALSE;
  option->stateOnly      = FALSE;
  option->node           = NULL;
  option->locGlob        = PORT_BNETGLOBAL_DD;
  option->progress       = FALSE;
  option->cacheSize      = DDI_CACHE_SLOTS;
  option->maxMemory      = 0; /* set automatically */
  option->slots          = CUDD_UNIQUE_SLOTS;

  /*
   * Create Variable Order while Building the BDDs
   */

  option->ordering       = PI_PS_FROM_FILE;
  option->orderPiPs      = NULL;
  option->reordering     = CUDD_REORDER_NONE;
  option->autoMethod     = CUDD_REORDER_SIFT;
  option->autoDyn        = 0;
  option->treefile       = NULL;
  option->firstReorder   = DD_FIRST_REORDER;
  option->countDead      = FALSE;
  option->maxGrowth      = 20;
  option->groupcheck     = CUDD_GROUP_CHECK7;
  option->arcviolation   = 10;
  option->symmviolation  = 10;
  option->recomb         = DD_DEFAULT_RECOMB;
  option->nodrop         = TRUE;
  option->signatures     = FALSE;
  option->verb           = 0;
  option->gaOnOff        = 0;
  option->populationSize = 0; /* use default */
  option->numberXovers   = 0; /* use default */
  option->bdddump        = FALSE;
  option->dumpFmt        = 0; /* dot */
  option->dumpfile       = NULL;
  option->store          = -1; /* do not store */
  option->storefile      = NULL;
  option->load           = FALSE;
  option->loadfile       = NULL;

  return(option);
}

