/**CFile**********************************************************************

  FileName    [partConj.c]

  PackageName [part]

  Synopsis    [Conjunctively partitioned functions]

  Description []

  SeeAlso     []  

  Author    [Gianpiero Cabodi and Stefano Quer]

  Copyright [This file was created at the Politecnico di Torino,
    Torino, Italy. 
    The  Politecnico di Torino makes no warranty about the suitability of 
    this software for any purpose.  
    It is presented on an AS IS basis.]
  
  Revision    []

******************************************************************************/

#include "partInt.h"

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */ 
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */ 
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */ 
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */ 
/*---------------------------------------------------------------------------*/

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/


/**AutomaticEnd***************************************************************/

  
/*----------------------------------------------------------------------------*/
/* Definition of internal function                                            */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/* Definition of External function                                            */
/*----------------------------------------------------------------------------*/

/**Function*******************************************************************

  Synopsis    [Compute the multiway and-exist over a conjunctively partitioned
    function]

  Description []

  SideEffects [none]

  SeeAlso     []

*****************************************************************************/ 

Ddi_Dd_t *
Part_DdMultiwayLinearAndExist (
  Ddi_Dd_t *FPart           /* Input partitioned function */,
  Ddi_VarSet_t *smoothV     /* Var Set */,
  int threshold             /* Size threshold for result factor */,
  int verbosity             /* Level of verbosity */,
  int *sizeMaxP             /* Pointer to Size Max */
)
{
  Ddi_Dd_t *Ris;
  Ddi_Dd_t *F,                /* an element in the BDDs'array */
           *P,*Pnew;          /* product of BDDs */
  Ddi_VarSet_t *smoothF, *smoothPF, *suppF, *suppP, *suppRis,
    *suppNext, *smooth;
  Ddi_VarSet_t **EarlySm;   /* early smoothed variable sets */
  Ddi_DdMgr *dd;
  int i, size, n;

  *sizeMaxP = 0;

  if (Ddi_DdReadCode (FPart)==DDI_MONO_BOOLF) {
    return (Ddi_DdExist (FPart, smoothV));
  }

  n = Ddi_DdReadPartNum (FPart);
  Pdtutil_Assert (n>=1, "Bug: no partitions in multiway and-exist");

  dd = Ddi_DdReadMgr(FPart);    /* dd manager */

  /*
   *  Generate vector of early smoothable variable sets, analyzing
   *  Partitions in reverse order.
   */

  EarlySm = Pdtutil_Alloc (Ddi_VarSet_t *, n);

  /* Check for Dynamic Memory Error */
  if (EarlySm == NULL) {
    return (NULL);
  }

  suppNext = Ddi_VarSetVoid (dd);

  for (i=n-1; i>=0; i--) {

    /* take the support of i-th function */
    suppF = Ddi_DdSupport (Ddi_DdReadIthPart(FPart,i));
    suppF = Ddi_VarSetEvalFree( 
      Ddi_VarSetIntersect(suppF,smoothV),suppF);

    /* Isolate variables not in the following functions for early smoothing */
    EarlySm[i] = Ddi_VarSetDifference(suppF,suppNext);

    /* Update support of following functions */
    suppNext = Ddi_VarSetEvalFree( 
               Ddi_VarSetUnion(suppF,suppNext),suppNext);

    Ddi_VarSetFree (suppF);
  }

  Ddi_VarSetFree (suppNext);

  Ris = Ddi_DdDup(Ddi_MgrReadOne(dd));
  P = Ddi_DdDup(Ddi_MgrReadOne(dd));
  suppP = Ddi_VarSetVoid(dd);
  suppRis = Ddi_VarSetVoid(dd);

  for (i=0; i<n; i++) {

    F = Ddi_DdDup (Ddi_DdReadIthPart (FPart, i));

    /* compute smooth varsets */
    smooth = Ddi_VarSetDifference (EarlySm[i], suppRis);
    smoothF = Ddi_VarSetDifference (smooth, suppP);
    smoothPF = Ddi_VarSetIntersect(smooth, suppP);
    Ddi_VarSetFree (smooth);
    Ddi_VarSetFree (EarlySm[i]);

    size = Ddi_DdSize (F);
    if (size > *sizeMaxP) {
      *sizeMaxP = size;
    }

    if (verbosity >= Pdtutil_VerbLevelAppMed_c) {
      fprintf (stdout, "{%d", size);
      fflush (stdout);
    }

    /* Early smoothing on F */
    if (!(Ddi_VarSetIsVoid (smoothF))) {
      F = Ddi_DdEvalFree (Ddi_DdExist (F, smoothF), F);

      size = Ddi_DdSize (F);
      if (size > *sizeMaxP) {
        *sizeMaxP = size;
      }

      if (verbosity > Pdtutil_VerbLevelAppMed_c) {
        fprintf (stdout, "<%d>%d}", Ddi_DdSize (smoothF)-1, size);
        fflush (stdout);
      }

    } else {
      if (verbosity >= Pdtutil_VerbLevelAppMed_c) {
        fprintf (stdout, "}");
        fflush (stdout);
      }
    }

    Ddi_VarSetFree (smoothF);

    if (!(Ddi_VarSetIsVoid (smoothPF))) {
      /* Early smoothing on P and F */
      if (verbosity >= Pdtutil_VerbLevelAppMed_c) {
        fprintf (stdout, "(%d)", Ddi_DdSize (smoothPF)-1);
        fflush (stdout);
      }

      Pnew = Ddi_DdAndExist (P, F, smoothPF);
    } else {
      /* No early smoothing on P and F */  

      Pnew = Ddi_DdAnd (P, F);
    }

    Ddi_VarSetFree (smoothPF);

    size = Ddi_DdSize (Pnew);

    if ((size>threshold)&&(threshold>=0)) {

      suppRis = Ddi_VarSetEvalFree(Ddi_VarSetUnion(suppP,suppRis),suppRis);

      if (!Ddi_DdIsOne(P)) {
        Ris = Ddi_DdEvalFree(Ddi_DdAddCpart(Ris,P),Ris);
      }

      Ddi_DdFree (Pnew);
      Ddi_DdFree (P);
      P = Ddi_DdDup(F);

      if (verbosity >= Pdtutil_VerbLevelAppMed_c) {
        fprintf (stdout, "(+%d)-", size-threshold);
        fflush (stdout);
      }

    } else { 

      Ddi_DdFree (P);
      P = Pnew;

      if (size > *sizeMaxP ) {
        *sizeMaxP = size;
      }

      if (verbosity >= Pdtutil_VerbLevelAppMed_c) {
        fprintf (stdout, "=%d*", size);
        fflush (stdout);
      }

    }

    suppP = Ddi_VarSetEvalFree(Ddi_DdSupport(P), suppP);

    Ddi_DdFree (F);

  } /* end-for(i) */

  Pdtutil_Free (EarlySm);

  Ris = Ddi_DdEvalFree (Ddi_DdAddCpart (Ris, P), Ris);
  Ddi_DdFree (P);
  Ddi_VarSetFree (suppP);
  Ddi_VarSetFree (suppRis);

  if ((Ddi_DdReadCode(Ris)!=DDI_MONO_BOOLF)&&(Ddi_DdReadPartNum(Ris)==1)) {
    Ris = Ddi_DdEvalFree (Ddi_DdPart2Mono (Ris), Ris);
  }

  return (Ris);
}

/**Function*******************************************************************

  Synopsis    [Computes the product of function and take existential
    abstraction]

  Description [The function computes the following operation:<br>
    <p>
    Exist (smoothV) ( f1*f2*f2* ... *fn )
    <p>
    where f1,f2,..fn are elements of a array's function (DdArray),
    and <em>smoothV</em> are variables that must be abstracted.<br>
    Although existential abstraction does not distribute over
    conjunction, function can be moved out of the scope of an
    existential quantification if they do not depend on any of 
    the variables being quantified.<br> 
    This method is called <em>early quantification</em> or
    <em>early smooth</em>.
    ]

  SideEffects [none]

  SeeAlso     []

*****************************************************************************/ 

Ddi_DdArray_t *
Part_DdArrayMultiwayLinearAndExist (
  Ddi_DdArray_t *Fa         /* Array of BDDs */,
  Ddi_VarSet_t *smoothV     /* Var Set */,
  int threshold             /* Size threshold for result factor */,
  int verbosity             /* Level of verbosity */
)
{
  Ddi_DdArray_t *FaRis;
  Ddi_Dd_t *F,                /* an element in the BDDs'array */
           *P,*Pnew;          /* product of BDDs */
  Ddi_VarSet_t *smoothF, *smoothPF, *suppF, *suppP,
    *suppRis, *suppNext, *smooth;
  Ddi_VarSet_t **EarlySm;   /* early smoothed variable sets */
  Ddi_DdMgr *dd;
  int i,size,
      n;                   /* length of array */

  n = Ddi_DdArray_n(Fa);

  Pdtutil_Assert(n>=1, "DdArray empty");
  dd = Ddi_DdReadMgr(Ddi_DdArrayFetch(Fa,0));    /* dd manager */
  for (i=0;i<n;i++)
    if (Ddi_DdReadMgr(Ddi_DdArrayFetch(Fa,i))!=dd) {
      fprintf (stdout, "single manager for a dd array still required!\n");
      return (NULL);
    }

  /*
   *  Generate vector of early smoothable variable sets, analyzing Fa functions
   *  in reverse order.
   */

  EarlySm = Pdtutil_Alloc(Ddi_VarSet_t *, n);
  if (EarlySm == NULL)
    return NULL; /* dynamic memory error */

  suppNext = Ddi_VarSetVoid(dd);

  for (i=n-1; i>=0; i--) {

    /* Take the support of i-th function */
    suppF = Ddi_DdSupport(Ddi_DdArrayFetch(Fa,i));
    suppF = Ddi_VarSetEvalFree( 
               Ddi_VarSetIntersect(suppF,smoothV),suppF);

    /* Isolate variables not in the following functions for early smoothing */
    EarlySm[i] = Ddi_VarSetDifference(suppF,suppNext);

    /* Update support of following functions */
    suppNext = Ddi_VarSetEvalFree( 
      Ddi_VarSetUnion(suppF,suppNext),suppNext);

    Ddi_VarSetFree (suppF);
  }

  Ddi_VarSetFree (suppNext);

  FaRis = Ddi_DdArrayAlloc (0);
  P = Ddi_DdDup (Ddi_MgrReadOne (dd));
  suppP = Ddi_VarSetVoid (dd);
  suppRis = Ddi_VarSetVoid (dd);

  for (i=0; i<n; i++) {
    F = Ddi_DdDup(Ddi_DdArrayFetch(Fa,i));

    /* compute smooth varsets */
    smooth = Ddi_VarSetDifference(EarlySm[i],suppRis);
    smoothF = Ddi_VarSetDifference(smooth,suppP);
    smoothPF = Ddi_VarSetIntersect(smooth,suppP);
    Ddi_VarSetFree (smooth);
    Ddi_VarSetFree (EarlySm[i]);

    /* early smoothing on F */
    if (!(Ddi_VarSetIsVoid(smoothF))) {
      if (verbosity >= Pdtutil_VerbLevelAppMed_c) {
        fprintf (stdout, "<%d>", Ddi_DdSize (smoothF)-1);
        fflush(stdout);
      }
      F = Ddi_DdEvalFree (Ddi_DdExist(F,smoothF),F);
    }
    Ddi_VarSetFree (smoothF);

    if (!(Ddi_VarSetIsVoid(smoothPF))) {
      /* early smoothing on PF */
      if (verbosity >= Pdtutil_VerbLevelAppMed_c) {
        fprintf (stdout, "(%d)", Ddi_DdSize (smoothPF)-1);
        fflush(stdout);
      }
      if (verbosity >= Pdtutil_VerbLevelAppMed_c) {
        fprintf (stdout, "{%d}", Ddi_DdSize (F));
        fflush (stdout);
      }
      Pnew = Ddi_DdAndExist(P, F, smoothPF);
    } else {
      /* no early smoothing on PF: and */  
      if (verbosity >= Pdtutil_VerbLevelAppMed_c) {
        fprintf (stdout, "{%d}", Ddi_DdSize(F));
        fflush (stdout);
      }
      Pnew = Ddi_DdAnd(P,F);
    }
    Ddi_VarSetFree (smoothPF);

    size = Ddi_DdSize(Pnew);

    if ((size>threshold)&&(threshold>=0)) {

      suppRis = Ddi_VarSetEvalFree(Ddi_VarSetUnion(suppP,suppRis),suppRis);

      if (!Ddi_DdIsOne(P)) {
        Ddi_DdArrayInsertLast(FaRis,P);
      }
      Ddi_DdFree (Pnew);
      Ddi_DdFree (P);
      P = Ddi_DdDup(F);

      if (verbosity >= Pdtutil_VerbLevelAppMed_c) {
        fprintf (stdout, "(+%d)-",size-threshold);
        fflush(stdout);
      }
    } else { 

      Ddi_DdFree (P);
      P = Pnew;

      if (verbosity >= Pdtutil_VerbLevelAppMed_c) {
        fprintf (stdout, "%d-", size);
        fflush(stdout);
      }

    }

    suppP = Ddi_VarSetEvalFree(Ddi_DdSupport(P), suppP);

    Ddi_DdFree (F);

  } /* end-for(i) */

  Pdtutil_Free(EarlySm);

  Ddi_DdArrayInsertLast(FaRis,P);
  Ddi_DdFree (P);
  Ddi_VarSetFree (suppP);
  Ddi_VarSetFree (suppRis);

  return (FaRis);
}

/**Function*******************************************************************

  Synopsis    []

  Description []

  SideEffects [none]

  SeeAlso     []

*****************************************************************************/ 

Ddi_Dd_t *
Part_DdDisjSuppPart (
  Ddi_Dd_t *f              /* a BDD */,
  Ddi_Dd_t *TR             /* a Clustered Transition Relation */,
  Ddi_VarArray_t *psv      /* array of present state variables */,
  Ddi_VarArray_t *nsv      /* array of next state variables */,
  int verbosity            /* level of verbosity */
)
{
  int i, n;
  Ddi_Dd_t *f2, *fh, *fl;
  Ddi_VarSet_t *suppf, *suppl, *supph;

  Pdtutil_Assert (Ddi_DdReadCode(TR)==DDI_CPART_BOOLF,
    "Conjuntion of BDDs expected");

  suppf = Ddi_DdSupport(f);
  suppl = Ddi_VarSetDup(suppf);

  fprintf (stdout, "%d support vars found\nHow many keep ? ",
    Ddi_VarSet_n(suppf));
  fflush(stdout);
  scanf ("%d",&n);
  fprintf (stdout, "-> %d\n", n);
  for(i=0;i<n;i++) {
    suppl = Ddi_VarSetEvalFree(Ddi_VarSetRemove(suppl,
             Ddi_VarSetTop(suppl)),suppl);
  }
  supph = Ddi_VarSetDifference(suppf,suppl);


  fh = Ddi_DdExist(f,suppl);
  fl = Ddi_DdExist(f,supph);
  printf ("FH: [%d] - FL: [%d]\n", Ddi_DdSize(fh),Ddi_DdSize(fl));
  f2 = Ddi_DdConstrain(f,fh);
  printf ("F|FH: [%d]\n", Ddi_DdSize(f2));
  f2 = Ddi_DdEvalFree(Ddi_DdConstrain(f2,fl),f2);
  printf ("F|FL: [%d]\n", Ddi_DdSize(f2));
  Ddi_DdFree (f2);

  f2 = Ddi_DdNot(f);

  fh = Ddi_DdExist(f2,suppl);
  fl = Ddi_DdExist(f2,supph);
  Ddi_DdFree (f2);
  printf ("~FH: [%d] - ~FH: [%d]\n", Ddi_DdSize(fh), Ddi_DdSize(fl));
  f2 = Ddi_DdConstrain(f,fh);
  printf ("F|~FH: [%d]\n", Ddi_DdSize(f2));
  Ddi_DdFree (f2);
  f2 = Ddi_DdConstrain(f,fl);
  printf ("F|~FL: [%d]\n", Ddi_DdSize(f2));
  Ddi_DdFree (f2);

  Ddi_DdFree (f);

#if 0
  f2 = Ddi_DdDup(f);

  for (i=0;i<Ddi_DdReadPartNum(TR);i++) {
    t = Ddi_DdSwapVariables(Ddi_DdReadIthPart(TR,i),nsv,psv);
    supp = Ddi_DdSupport(t);
    smooth = Ddi_VarSetDifference(suppf,supp);
    Ddi_DdFree (t);
    part = Ddi_DdExist(f2,smooth);
    f2 = Ddi_DdEvalFree (Ddi_DdConstrain(f2,part),f2);
    printf ("Partition %d: (%d) - [%d]\n", i,
            Ddi_DdSize(part),Ddi_DdSize(f2));
    Ddi_DdFree (part);
    Ddi_VarSetFree (supp);
    Ddi_VarSetFree (smooth);
  }

  printf ("Trying Disj on ~F\n");

  f2 = Ddi_DdNot(f);

  for (i=0;i<Ddi_DdReadPartNum(TR);i++) {
    t = Ddi_DdSwapVariables(Ddi_DdReadIthPart(TR,i),nsv,psv);
    supp = Ddi_DdSupport(t);
    smooth = Ddi_VarSetDifference(suppf,supp);
    Ddi_DdFree (t);
    part = Ddi_DdExist(f2,smooth);
    f2 = Ddi_DdEvalFree (Ddi_DdConstrain(f2,part),f2);
    printf ("Partition %d: (%d) - [%d]\n", i,
            Ddi_DdSize(part),Ddi_DdSize(f2));
    Ddi_DdFree (part);
    Ddi_VarSetFree (supp);
    Ddi_VarSetFree (smooth);
  }
#endif

  Ddi_VarSetFree (suppf);
  return (NULL);
}



