/**CFile***********************************************************************

  FileName    [partCuPlus.c]

  PackageName [part]

  Synopsis    [Recursive Functions for Disjunctive Partitioning.
    This functions rely on the ddi package.]

  Description [External procedures included in this file:
    <ul>
    <li> Part_EstimateCofactor
    <li> Part_EstimateCofactorComplex
    <li> Part_EstimateCofactorFast
    <li> Part_EstimateCofactorFreeOrder
    </ul>
    Static procedures included in this module:
    <ul>
    <li> partEstimateCofactor
    <li> partEstimateCofactorComplex
    <li> partEstimateCofactorFast
    <li> partEstimateCofactorFreeOrder
    <li> ddFlagClear 
    </ul>]

  SeeAlso   []  

  Author    [Gianpiero Cabodi and Stefano Quer]

  Copyright [This file was created at the Politecnico di Torino,
    Torino, Italy. 
    The  Politecnico di Torino makes no warranty about the suitability of 
    this software for any purpose.  
    It is presented on an AS IS basis.
    ]
  
  Revision  []

******************************************************************************/

#include "partInt.h"                                          

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static int partEstimateCofactor(Ddi_DdNode *f, Ddi_Dd_t *topVar);
static struct fs_struct partEstimateCofactorComplex(Ddi_DdNode *f, Ddi_Dd_t *topVar);
static int partEstimateCofactorFast(Ddi_DdNode *f, int LeftRight, Ddi_VarSet_t *supp, int suppSize, int *totalArray, int *thenArray, int *elseArray);
static int partEstimateCofactorFreeOrder(Ddi_DdNode *f, Ddi_Dd_t *topVar, Ddi_VarSet_t *supp, int *vet);
static void ddFlagClear(Ddi_DdNode *f);

/**AutomaticEnd***************************************************************/

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Counts the number of nodes of the cofactors: 
    Stefano Quer solution.]

  Description []

  SideEffects [None]

  SeeAlso     [partEstimateCofactor]

******************************************************************************/

int
Part_EstimateCofactor (
  Ddi_Dd_t *f        /* Input BDD */,
  Ddi_Dd_t *topVar   /* Top Variable */
  )
{
  int i;	

  i = partEstimateCofactor (Ddi_DdRootNode (f), topVar);
  ddFlagClear (Ddi_DdRootNode (f));
  
  return(i);
} 


/**Function********************************************************************

  Synopsis    [Counts the number of nodes of the cofactors:
    Fabio Somenzi Private Commnunication Solution.]

  Description []

  SideEffects [none]

  SeeAlso     [partEstimateCofactor]

******************************************************************************/

int
Part_EstimateCofactorComplex (
  Ddi_Dd_t *f         /* Input BDD */,
  Ddi_Dd_t *topVar    /* Top Variable */
  )
{
  struct fs_struct s;	
  
  s = partEstimateCofactorComplex (Ddi_DdRootNode (f), topVar);
  ddFlagClear (Ddi_DdRootNode (f));
  
  return (s.n);
} 

/**Function********************************************************************

  Synopsis    [Count the number of nodes of the cofactors using the fast
    heuristic (TCAD'99)]

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

int
Part_EstimateCofactorFast (
  Ddi_Dd_t *f,
  int LeftRight,
  Ddi_VarSet_t *supp,
  int suppSize,
  int *totalArray,
  int *thenArray,
  int *elseArray
  )
{
  int i;

  i = partEstimateCofactorFast (Ddi_DdRootNode (f), LeftRight, supp,
    suppSize, totalArray, thenArray, elseArray);
  ddFlagClear (Ddi_DdRootNode (f));

  return (i);
}

/**Function********************************************************************

  Synopsis    [Count the number of nodes of the cofactors using the Free
    Order heuristic (TCAD'99).]

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

int
Part_EstimateCofactorFreeOrder (
  Ddi_Dd_t *f          /* Input BDD */,
  Ddi_Dd_t *topVar     /* Top Variable */,
  Ddi_VarSet_t *supp,
  int *vet
  )
{
  int i;	

  i = partEstimateCofactorFreeOrder (Ddi_DdRootNode (f), topVar, supp, vet);
  ddFlagClear (Ddi_DdRootNode (f));
  
  return(i);
} 

/**Function********************************************************************

  Synopsis    [Performs the recursive step of Part_EstimateCofactor]

  Description []

  SideEffects [None]

  SeeAlso     [Part_EstimateCofactor]

******************************************************************************/

static int
partEstimateCofactor (
  Ddi_DdNode *f     /* Input BDD */,
  Ddi_Dd_t *topVar  /* Top Variable */
  )
{
  int thenVal, elseVal, val;

  if (Ddi_DdNodeIsVisited (f)) {
    return (0);
  }

  if (Ddi_DdNodeIsConstant (f)) {
    return (0);
  }

  Ddi_DdNodeSetVisited(f);
  
  if (Ddi_DdNodeReadIndex (f) == Ddi_DdNodeReadIndex
    (Ddi_DdRootNode (topVar))) {
    if (Ddi_DdIsComplement(topVar)) {
      val = partEstimateCofactor (Ddi_DdNodeElse (f), topVar);
    } else {
      val = partEstimateCofactor (Ddi_DdNodeThen (f), topVar);
    }
  } else {
    thenVal = partEstimateCofactor (Ddi_DdNodeThen (f), topVar);
    elseVal = partEstimateCofactor (Ddi_DdNodeElse (f), topVar);
    val = thenVal + elseVal +1;
  }
  
  return (val);
} 


/**Function********************************************************************

  Synopsis    [Performs the recursive step of Part_EstimateCofactor]

  Description []

  SideEffects [none]

  SeeAlso     [Part_EstimateCofactor]

******************************************************************************/

static struct fs_struct 
partEstimateCofactorComplex (
  Ddi_DdNode *f     /* Input BDD */,
  Ddi_Dd_t *topVar  /* Top Variable */
  )
{
  struct fs_struct s0, s, st, se;
  
  s0.p = f;
  s0.n = 0;
  
  if (Ddi_DdNodeIsVisited (f)) {
    return(s0);
  }
  if (Ddi_DdNodeIsConstant (f)) {
    return (s0);
  }
  Ddi_DdNodeSetVisited(f);
  
  if (Ddi_DdNodeReadIndex (f) ==
    Ddi_DdNodeReadIndex (Ddi_DdRootNode (topVar))) {
    if (Ddi_DdIsComplement(topVar)) {
      s = partEstimateCofactorComplex (Ddi_DdNodeElse (f), topVar);
    } else {
      s = partEstimateCofactorComplex (Ddi_DdNodeThen (f), topVar);
    }
  } else {
    st = partEstimateCofactorComplex (Ddi_DdNodeThen (f), topVar);
    se = partEstimateCofactorComplex (Ddi_DdNodeElse (f), topVar);
    if (st.p == se.p) {
      s = st;
    } else {
      s.p = f;
      s.n = st.n + se.n + 1;
    }
  }
  
  return (s);
}

/**Function********************************************************************

  Synopsis    [Performs the recursive step of fsCntNodesVar]

  Description []

  SideEffects [none]

  SeeAlso     [fsCntNodesVar]

******************************************************************************/

static int
partEstimateCofactorFast (
  Ddi_DdNode *f       /* input BDD */,
  int LeftRight,
  Ddi_VarSet_t *supp,
  int suppSize,
  int *totalArray,
  int *thenArray,
  int *elseArray
  )
{
  int i, totalVal, thenVal, elseVal;
  Ddi_DdMgr *dd = Ddi_DdReadMgr (supp);

  if (Ddi_DdNodeIsVisited (f)) {
    return (0);
  }

  if (Ddi_DdNodeIsConstant (f)) {
    return (0);
  }

  Ddi_DdNodeSetVisited(f);

#if 0
  i=0;
  scan = Ddi_VarSetDup (supp);
  var = Ddi_VarSetTop (scan);
  while (Ddi_DdNodeReadIndex (f) != Ddi_VarIndex (var)) {
    Pdtutil_Assert (!Ddi_VarSetIsVoid (scan), "Support Empty!");
    tmp = Ddi_VarSetNext (scan);
    Ddi_VarSetFree (scan);
    scan = tmp;
    var = Ddi_VarSetTop (scan);
    i++;
  }
#else 
  i = Ddi_ReadPerm(dd, Ddi_DdNodeReadIndex (f));
#endif

  if (LeftRight==1) {
    elseVal = partEstimateCofactorFast (Ddi_DdNodeElse (f),
      LeftRight, supp, suppSize, totalArray, thenArray, elseArray);
    thenVal = partEstimateCofactorFast (Ddi_DdNodeThen (f), LeftRight,
      supp, suppSize, totalArray, thenArray, elseArray);
  } else {
    thenVal = partEstimateCofactorFast (Ddi_DdNodeThen (f), LeftRight,
      supp, suppSize, totalArray, thenArray, elseArray);
    elseVal = partEstimateCofactorFast (Ddi_DdNodeElse (f),
      LeftRight, supp, suppSize, totalArray, thenArray, elseArray);
  }

  totalVal = thenVal + elseVal + 1;
  totalArray[i] += 1;
  thenArray[i] += thenVal;
  elseArray[i] += elseVal;

  return (totalVal);
}

/**Function********************************************************************

  Synopsis    [Performs the recursive step of fsCntNodesVar]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

static int
partEstimateCofactorFreeOrder (
  Ddi_DdNode *f            /* Input BDD */,
  Ddi_Dd_t *topVar         /* Top Variable */,
  Ddi_VarSet_t *supp,
  int *vet
  )
{
  int i, thenVal, elseVal, val;
  Ddi_VarSet_t *scan, *tmp;
  Ddi_Var_t *var;

  if (Ddi_DdNodeIsVisited (f)) {
    return(0);
  }

  if (Ddi_DdNodeIsConstant (f)) {
    return(0);
  }

  Ddi_DdNodeSetVisited (f);

  i=0;
  scan = Ddi_VarSetDup (supp);
  var = Ddi_VarSetTop (scan);
  while (Ddi_DdNodeReadIndex (f) != Ddi_VarIndex (var)) {
    Pdtutil_Assert (!Ddi_VarSetIsVoid (scan), "Support Empty!");
    tmp = Ddi_VarSetNext (scan);
    Ddi_VarSetFree (scan);
    scan = tmp;
    var = Ddi_VarSetTop (scan);
    i++;
  }
  vet[i]++;

  if (Ddi_DdNodeReadIndex (f) == Ddi_DdNodeReadIndex (Ddi_DdRootNode
    (topVar))) {
    if (Ddi_DdIsComplement (topVar)) {
      val = partEstimateCofactorFreeOrder (Ddi_DdNodeElse (f),
        topVar, supp, vet);
    } else {
      val = partEstimateCofactorFreeOrder (Ddi_DdNodeThen (f),
        topVar, supp, vet);
    }
  } else {
    thenVal = partEstimateCofactorFreeOrder (Ddi_DdNodeThen (f),
      topVar, supp, vet);
    elseVal = partEstimateCofactorFreeOrder (Ddi_DdNodeElse (f),
      topVar, supp, vet);
    val = thenVal + elseVal + 1;
  }

  return (val);
}

/**Function********************************************************************

  Synopsis    [Performs a DFS from f, clearing the LSB of the next
    pointers.]

  Description [Static function copied from cudd_Util.c]

  SideEffects [None]

  SeeAlso     [Part_EstimateCofactor, Part_EstimateCofactor]

******************************************************************************/

static void
ddFlagClear (
  Ddi_DdNode *f                     /* input BDD */
  )
{
  if (!Ddi_DdNodeIsVisited (f)) { 
    return;
  }
  
  /* Clear visited flag. */
  Ddi_DdNodeClearVisited (f);
  if (Ddi_DdNodeIsConstant (f)) { 
    return;
  }
  ddFlagClear (Ddi_DdNodeThen (f));
  ddFlagClear (Ddi_DdNodeElse (f));

  return;
} 
