/**CFile***********************************************************************

  FileName    [partCudd.c]

  PackageName [part]

  Synopsis    [Functions for Disjunctive and Conjunctive Partitioning
    as implemented in the Cudd package.]

  Description [This file contains interface function to linked with the
    standard cudd partitionin routines.]

  SeeAlso     []  

  Author    [Gianpiero Cabodi and Stefano Quer]

  Copyright [This file was created at the Politecnico di Torino,
    Torino, Italy. 
    The  Politecnico di Torino makes no warranty about the suitability of 
    this software for any purpose.  
    It is presented on an AS IS basis.]
  
  Revision    []

******************************************************************************/
                                         
#include "partInt.h"                                          

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static void partPartitionCudd(Ddi_Dd_t *function, Part_Method_e partitionMethod, int threshold, Ddi_DdArray_t *partition, Pdtutil_VerbLevel_e verbosity);

/**AutomaticEnd***************************************************************/

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Partition a state set]

  Description [This function derives from the Part_PartitionDisjSet
    function. It is rewritten here for two reasons:
    <ul>
    <li> this file includes the direct calls to the routine implemented
    in the Cudd package,
    <li> this file includes both disjunctive and conjuctive decompositions.
    </ul>
    ]

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

Ddi_Dd_t *
Part_PartitionSetCudd (
  Ddi_Dd_t *f                      /* Input BDD */,
  Part_Method_e partitionMethod    /* Partition Method */,
  int threshold                    /* Threshold Size */,
  Pdtutil_VerbLevel_e verbosity    /* Verbosity Level */
  )
{
  int i, bddSize=0;
  long startTime=0, currTime;
  Ddi_DdArray_t *ddVet;
  Ddi_Dd_t *fdup;
  Ddi_Dd_t *result, *dpart;
  Ddi_DdMgr *dd = Ddi_DdReadMgr(f);

  ddVet = Ddi_DdArrayAlloc (0);
  
  /* f is freed by partPartitionDisjSet */ 
  fdup = Ddi_DdDup (f);
  
  if (verbosity >= Pdtutil_VerbLevelUsrMax_c) {
    startTime = util_cpu_time();
    bddSize = Ddi_DdSize (fdup);
  }

  /*----------------------------- Partitioning ------------------------------*/

  switch (Ddi_DdReadCode(fdup)) {
    case DDI_MONO_BOOLF:
      partPartitionCudd (fdup, partitionMethod, threshold, ddVet, verbosity);
      break;
    case DDI_CPART_BOOLF:
      Pdtutil_Warning (Ddi_DdReadPartNum(f)>1, "One partition expected");
      partPartitionCudd (Ddi_DdReadIthPart (fdup,0), 
        partitionMethod, threshold, ddVet, verbosity);
      break;
  }
  
  /*--------------------- From BDD-Array to BDD-Disjunctive -----------------*/

  if (verbosity >= Pdtutil_VerbLevelUsrMax_c) {
    fprintf (stdout, "PARTITIONING (%d) ==> [", bddSize);
  }

  result = Ddi_MgrReadZero (dd);

  for (i=0; i<Ddi_DdArray_n (ddVet); i++) {
    dpart = Ddi_DdArrayFetch (ddVet, i);
    bddSize = Ddi_DdSize (dpart);

    if (verbosity >= Pdtutil_VerbLevelUsrMax_c) {
      fprintf (stdout, "%d+", bddSize);
      fflush(stdout);
    }

    /* Add partition to disjoined result */
    result = Ddi_DdEvalFree (Ddi_DdAddDpart (result, dpart), result);
  }

  /*------------------------------ Final Statistics -------------------------*/

  if (verbosity >= Pdtutil_VerbLevelUsrMax_c) {
    bddSize = Ddi_DdArraySize (ddVet);
    currTime = util_cpu_time();

    fprintf (stdout, "=%d]", bddSize);
    fprintf (stdout, "(time:%s)\n", util_print_time (currTime - startTime));
    fflush(stdout);
  }

  Ddi_DdArrayFree (ddVet);
  
  return (result);
}

/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Direct call to cudd partitioning functions.]

  Description [Direct call to cudd partitioning functions. It transforms
    ddi objects to cudd ones before the call and reverse the process after
    it.]

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

static void
partPartitionCudd (
  Ddi_Dd_t *function               /* Function to Partition */,
  Part_Method_e partitionMethod    /* Partition Method */,
  int threshold                    /* Threshold Size */,
  Ddi_DdArray_t *partition         /* Partition Array */,
  Pdtutil_VerbLevel_e verbosity    /* Verbosity Level */
)
{
  Ddi_Dd_t *thenFunction, *elseFunction;
  Ddi_DdMgr *dd;
  DdManager *ddCudd;
  DdNode *functionCudd;
  DdNode **partitionCudd;
  long startTime, currTime;
  int bddSize, thenSize, elseSize, numPart=0;

  bddSize = Ddi_DdSize (function);
  startTime = util_cpu_time();

  /*--------------------- Check NULL method and BDD size --------------------*/

  if (bddSize<threshold) {
    Ddi_DdArrayInsert (partition, Ddi_DdArray_n (partition), function);
    Ddi_DdFree (function);
    return;
  }

  /*--------------------------------- Split ---------------------------------*/

  /*
   *  From ddi to Cudd
   */

  functionCudd = Ddi_DdToCU (function);
  dd = Ddi_DdReadMgr (function);
  ddCudd = Ddi_MgrReadMgrCU (dd);   

  switch (partitionMethod) {
    case Part_MethodAppCon_c:
      numPart = Cudd_bddApproxConjDecomp (ddCudd, functionCudd, &partitionCudd);
      break;
    case Part_MethodAppDis_c:
      numPart = Cudd_bddApproxDisjDecomp (ddCudd, functionCudd, &partitionCudd);
      break;
    case Part_MethodGenCon_c:
      numPart = Cudd_bddGenConjDecomp (ddCudd, functionCudd, &partitionCudd);
      break;
    case Part_MethodGenDis_c:
      numPart = Cudd_bddGenDisjDecomp (ddCudd, functionCudd, &partitionCudd);
      break;
    case Part_MethodIteCon_c:
      numPart = Cudd_bddIterConjDecomp (ddCudd, functionCudd, &partitionCudd);
      break;
    case Part_MethodIteDis_c:
      numPart = Cudd_bddIterDisjDecomp (ddCudd, functionCudd, &partitionCudd);
      break;
    case Part_MethodVarCon_c:
      numPart = Cudd_bddVarConjDecomp (ddCudd, functionCudd, &partitionCudd);
      break;
    case Part_MethodVarDis_c:
      numPart = Cudd_bddVarDisjDecomp (ddCudd, functionCudd, &partitionCudd);
      break;
    default:
      break;
  }
 
  if (numPart<2) {
    Ddi_DdArrayInsert (partition, Ddi_DdArray_n (partition),
      function);
    Ddi_DdFree (function);

    return;
  }

  /*
   *  From Cudd to ddi
   */

  thenFunction = Ddi_DdCreateFromCU (dd, partitionCudd[0]);
  elseFunction = Ddi_DdCreateFromCU (dd, partitionCudd[1]);

  /*------------------------------ Statistics -------------------------------*/

  thenSize = Ddi_DdSize (thenFunction);
  elseSize = Ddi_DdSize (elseFunction);

  currTime = util_cpu_time();

  if (verbosity >= Pdtutil_VerbLevelNone_c) {
    fprintf (stdout, "[%d--->%d,%d(time:%s)]\n",
      bddSize, thenSize, elseSize,
      util_print_time(currTime - startTime));
  }

  /*--------------------------------- Recur ---------------------------------*/

  partPartitionCudd (thenFunction, partitionMethod, threshold,
    partition, verbosity);
  partPartitionCudd (elseFunction, partitionMethod, threshold,
    partition, verbosity);

  return;
}



