/**CFile***********************************************************************

  FileName    [trBuild.c]

  PackageName [tr]

  Synopsis    [Functions to Build a partitioned, clustered or monolithic
    transition relation]

  Description []

  SeeAlso     []  

  Author      [Gianpiero Cabodi and Stefano Quer]

  Copyright   [This file was created at the Politecnico di Torino,
    Torino, Italy. 
    The  Politecnico di Torino makes no warranty about the suitability of 
    this software for any purpose.  
    It is presented on an AS IS basis.]
  
  Revision    []

******************************************************************************/

#include "trInt.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/
  
/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/


/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Definition of external functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Build clustered Transition relation]

  Description [Create the clustered TR from transition relation by 
    clustering several function together.<br>
    The size of clustering is controlled by a threshold value.
    -1 is used to specify that there is no threshold.]

  SideEffects []

  SeeAlso     [Trav_BuildPartTranRel]

******************************************************************************/

Ddi_Dd_t *
Tr_BuildClusteredTR ( 
  Tr_Mgr_t      *trMgr     /* Tr manager */,
  Ddi_Dd_t      *tr        /* Input partitioned/clustered relation */,
  Ddi_VarSet_t  *svars     /* Set of smoothing variables */,
  Ddi_VarSet_t  *qvars     /* Set of quantifying variables */,
  Ddi_VarSet_t  *rvars     /* Set of range variables */,
  int           threshold  /* Threshold for creating clusters */
)
{
  Ddi_Dd_t *ClustTR;  
  long trTime;
  int sizeMax;
 
  trTime = util_cpu_time ();

  ClustTR = Ddi_DdDup (tr);
  if (threshold==0) {
    return (ClustTR);  /* no clustering */
  }

  if (Tr_MgrReadVerbosity (trMgr) >= Pdtutil_VerbLevelAppMin_c) {
    fprintf(stdout, "TR Clustering\n");
    fprintf(stdout, "|q|: %d\n", Ddi_VarSet_n (qvars));
    fprintf(stdout, "|s|: %d\n", Ddi_VarSet_n (svars));
    fprintf(stdout, "|r|: %d\n", Ddi_VarSet_n (rvars));
    fprintf(stdout, "thr: %d\n", threshold);
  }

  Tr_DdSortPartitions (trMgr, ClustTR, svars, qvars, rvars,
    Tr_MgrReadSortMethod (trMgr));

  ClustTR = Ddi_DdEvalFree(
    Part_DdMultiwayLinearAndExist (ClustTR, qvars, threshold,
      Tr_MgrReadVerbosity (trMgr), &sizeMax),
    ClustTR);

  if ((Ddi_DdReadCode(ClustTR)!=DDI_MONO_BOOLF)
    &&(Ddi_DdReadPartNum(ClustTR)==1)) {
    ClustTR = Ddi_DdEvalFree(Ddi_DdPart2Mono(ClustTR),ClustTR);
  }

  trTime = util_cpu_time() - trTime;
  trMgr->trTime += trTime;

  return (ClustTR);
}


/**Function********************************************************************

  Synopsis    [Build monolithic Transition relation]

  Description [Create the monilithic transition relation using clustering
               with infinite threshold.]

  SideEffects []

  SeeAlso     [Trav_BuildClusteredTR]

******************************************************************************/

Ddi_Dd_t *
Tr_BuildMonoTR ( 
  Tr_Mgr_t      *trMgr     /* TR manager */,
  Ddi_Dd_t      *tr        /* Input partitioned/clustered relation */,
  Ddi_VarSet_t  *svars     /* Set of smoothing variables */,
  Ddi_VarSet_t  *qvars     /* Set of quantifying variables */,
  Ddi_VarSet_t  *rvars     /* Set of range variables */
)
{
  return (Tr_BuildClusteredTR (trMgr, tr, svars, qvars, rvars, -1));
}


/**Function*******************************************************************

  Synopsis    [Build partitioned Transition Relation]

  Description [The function receives the array of next state functions
    and the array of next state variables.<br>
    It returns a cojunctively partitioned transition relation, or NULL in
    case of failure.]

  SideEffects [None]

  SeeAlso     [Trav_BuildClusteredTR]

******************************************************************************/

Ddi_Dd_t *
Tr_BuildPartTR (
  Ddi_DdArray_t *delta  /* array of state function */,
  Ddi_VarArray_t *y     /* array of next state variables */
)
{
  Ddi_Dd_t *F, *t_i, *temp;
  Ddi_Var_t *var;
  Ddi_DdArray_t *part_TR;
  Ddi_Dd_t *result;
  Ddi_DdMgr *dd;
  int i,n;

  Pdtutil_Assert(Ddi_DdArray_n(delta)==Ddi_VarArray_n(y),
    "number of deltas and y are not matching");

  dd = Ddi_DdReadMgr (Ddi_DdArrayFetch (delta, 0));

  n = Ddi_DdArray_n (delta); 

  part_TR = Ddi_DdArrayAlloc (n);
  if (part_TR==NULL) {
    return (NULL);
  }

  for (i=0;i<n;i++) {
    F = Ddi_DdArrayFetch(delta,i);
    var = Ddi_VarArrayFetch(y,i);
    t_i = Ddi_DdXnor((temp=Ddi_DdLiteral(dd,var)),F);
    Ddi_DdFree(temp);
    Ddi_DdArrayInsert(part_TR,i,t_i);
    Ddi_DdFree(t_i);
  }

  result = Ddi_DdPartFromArray(part_TR,DDI_CPART_BOOLF);
  Ddi_DdArrayFree(part_TR);

  return (result);
}

/**Function********************************************************************

  Synopsis    [Remove Lambda Latches]

  Description [Remove the so-called lambda-latches following the strategy
    presented at ICCAD'96 by Cabodi, Camurati, Quer.]

  SideEffects []

  SeeAlso     [Trav_BuildClusteredTR]

******************************************************************************/

void
Tr_RemoveLambdaLatches (
  Tr_Mgr_t *trMgr     /* TR manager */
)
{
  Ddi_VarSet_t *LLs, *LLy, *suppTR;

  if (trMgr->tr == NULL) {
    Pdtutil_Warning (1, "Command aborted: No TR found. Please build one.");
    return;
  }

  suppTR = Ddi_DdSupport (trMgr->tr);

  /* Create a variable set with all PS vars */
  LLs = Ddi_VarArrayToVarSet(Ddi_DdReadMgr (trMgr->tr),
    Tr_MgrReadPS (trMgr));

  /* Remove support of TR: only lambda latches survive */
  LLs = Ddi_VarSetEvalFree(Ddi_VarSetDifference(LLs,suppTR),LLs);

  if (!(Ddi_VarSetIsVoid(LLs))) {
    if (Tr_MgrReadVerbosity (trMgr) >= Pdtutil_VerbLevelAppMin_c) {
      fprintf (stdout, "Removing %d Lambda Latches\n", Ddi_VarSet_n(LLs));
    }

    /* the following is not formally correct for a varset! */
    LLy = Ddi_VarSetSwapVariables(LLs,Tr_MgrReadPS (trMgr),
      Tr_MgrReadNS (trMgr));
    trMgr->tr = Ddi_DdEvalFree(Ddi_DdExist (trMgr->tr,LLy), trMgr->tr);

    Ddi_VarSetFree(LLy);

    /* from and reached could be simplified in trav with Exist(set,LLs) */

  }
  Ddi_VarSetFree(LLs);

  return;
}
