/**CFile***********************************************************************

  FileName    [trClosure.c]

  PackageName [tr]

  Synopsis    [Functions doing transitive closure]

  Description []

  SeeAlso     []  

  Author      [Gianpiero Cabodi and Stefano Quer]

  Copyright   [This file was created at the Politecnico di Torino,
    Torino, Italy. 
    The  Politecnico di Torino makes no warranty about the suitability of 
    this software for any purpose.  
    It is presented on an AS IS basis.]
  
  Revision    []

******************************************************************************/

#include "trInt.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/
  
/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static Ddi_Dd_t * TrCompose(Tr_Mgr_t *trMgr, Ddi_Dd_t *f, Ddi_Dd_t *g, Ddi_VarArray_t *s, Ddi_VarArray_t *y, Ddi_VarArray_t *z, Ddi_VarSet_t *suppZ);

/**AutomaticEnd***************************************************************/

/*---------------------------------------------------------------------------*/
/* Definition of external functions                                          */
/*---------------------------------------------------------------------------*/


/**Function********************************************************************

  Synopsis    [Build transitive closure of a Transition Relation
    by means of iterative squaring]

  Description [Transitive closure of a Monolithic or Disjunctively 
    partitioned Transition Relation is operated using the iterative squaring
    method, expressed by the following recurrence equations:
    <p>
    T(0)(s,y) = TR(s,y) <br>
    T(i+1)(s,y)= TR(s,y) + Exist(z) ( T(i)(s,z) * T(i)(z,y) )
    <p>
    The least fixed point is T* (<em>transitive closure</em>)
    and the number of iterations required to compute T* is 
    logarithmic in the sequential depth of TR (the diameter of
    the state transition graph represented by TR).<br>
    The transitive closure describes the pairs of states that are
    connected by at least one path in the state graph of FSM.
    This function is a shell to TrBuildTransClosure, where the
    job is really done. Nothing is done here if the array of
    "intermediate" z variables is not NULL, whereas
    a temporary BDD manager is created (and destroyed when
    the job is done) if z is NULL, to avoid
    creating new variables in the original manager.
    Primary input variables (NON quantifying and NON state
    variables) are not taken into account. This means that they
    are NOT duplicated at every step as in standart squaring.
    They should (since they could) be quantifyed out from a 
    monolithic or disjunctively partitioned TR. Otherwise the
    algorithm only closes paths with constant input values.
    ]

  SideEffects [A Ddi Manager (a BDD manager) is temporarily allocated 
    if required for all the operations involved, if  
    the auxiliary set of variables (Zs) must be created.]

  SeeAlso     [TrBuildTransClosure]

******************************************************************************/

Ddi_Dd_t *
Tr_BuildTransClosure (
  Tr_Mgr_t *trMgr        /* Tr manager */,
  Ddi_Dd_t  *TR          /* Transition Relation */,
  Ddi_VarArray_t *s      /* present state vars */,
  Ddi_VarArray_t *y      /* next state vars */,
  Ddi_VarArray_t *z      /* intermediate state vars */
)
{
  Ddi_Dd_t *TSQ;
  Ddi_Dd_t *trAux;
  Ddi_Var_t *var;
  int i, level;
  int ns;                /* number of state vars */
  Ddi_DdMgr *ddTr;       /* original dd manager */
  Ddi_DdMgr *ddAux;      /* aux dd manager */
  Ddi_VarArray_t *sAux;  /* present state vars */
  Ddi_VarArray_t *yAux;  /* next state vars */
  Ddi_VarArray_t *zAux;  /* compositional state vars */

  Pdtutil_Assert (
    ((Ddi_DdReadCode(TR)!=DDI_CPART_BOOLF)||(Ddi_DdReadPartNum(TR)==1)),
    "Trying to apply squaring to conjunctively partitioned TR");

  /* Convert a 1 partition conjunction to a monolithic BDD */
  if ((Ddi_DdReadCode(TR)==DDI_CPART_BOOLF)&&(Ddi_DdReadPartNum(TR)==1)) {
    trAux = Ddi_DdPart2Mono(TR);
  } else {
    trAux = Ddi_DdDup(TR);
  }

  ns = Ddi_VarArray_n(s);
  Pdtutil_Assert(ns==Ddi_VarArray_n(y),
    "sizes of s,y variable arrays don't match");

  if (z != NULL) {
    Pdtutil_Assert(ns==Ddi_VarArray_n(z),
      "sizes of the z variable array don't match s,y sizes");
    TSQ = TrBuildTransClosure(trMgr,trAux,s,y,z);
    Ddi_DdFree(trAux);
    return (TSQ);
  }

  /* an auxiliary dd manager is created */
  ddTr  = Ddi_DdReadMgr(TR); 
  ddAux = Ddi_MgrDup(ddTr);    

  /* input parameters are copied to the ddAux manager */
  trAux = Ddi_DdEvalFree (Ddi_DdCopy (ddAux, trAux), trAux);
  sAux  = Ddi_VarArrayCopy (ddAux,s);
  yAux  = Ddi_VarArrayCopy (ddAux,y);

  /* temporarily disable dynamic reordering */
  Ddi_MgrAutodynSuspend (ddAux);

  /* create new array of variables (z) */
  zAux = Ddi_VarArrayAlloc (ns);
  for (i=0;i<ns;i++) {
    level = Ddi_VarCurrPos (ddAux,Ddi_VarArrayFetch(sAux,i));
    var = Ddi_DdNewVarAtLevel (ddAux,level);
    Ddi_VarArrayInsert (zAux,i,var); 
    if (level > Ddi_VarCurrPos (ddAux,Ddi_VarArrayFetch(yAux,i)))
      var = Ddi_VarArrayFetch (yAux,i);
    Ddi_VarMakeGroup (ddAux,var,3);
  }

  Ddi_MgrAutodynResume(ddAux);

  if (Tr_MgrReadVerbosity (trMgr) > Pdtutil_VerbLevelAppMed_c) {
    fprintf(stdout, "Tr Closure - created z vars: ");
    Ddi_PrintVarArray (zAux);
  }

  TSQ = TrBuildTransClosure (trMgr,trAux,sAux,yAux,zAux);

  Ddi_MgrAlign (ddTr, ddAux);

  TSQ = Ddi_DdEvalFree(Ddi_DdCopy(ddTr,TSQ),TSQ);

  Ddi_DdFree (trAux);
  Ddi_VarArrayFree (sAux);
  Ddi_VarArrayFree (yAux);

  Ddi_DdiMgrQuit (ddAux);

  return (TSQ);
}

/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Build transitive closure of a Transition Relation
    by means of iterative squaring]

  Description [Internal procedure to build the transitive closure of 
    a transition relation. This can now handle a Monolithic TR
    as well as a disjunctively partitioned one. In the latter case
    closure is only partial: it is operated within single partitions.
    The iterative squaring method, expressed by
    the following recurrence equations:
    <p>
    T(0)(s,y) = TR(s,y) <br>
    T(i+1)(s,y)= TR(s,y) + Exist(z) ( T(i)(s,z) * T(i)(z,y) )
    <p>
    The least fixed point is T* (<em>transitive closure</em>)
    and the number of iterations required to compute T* is 
    logarithmic in the sequential depth of TR (the diameter of
    the state transition graph represented by TR).<br>
    The transitive closure describes the pairs of states that are
    connected by at least one path in the state graph of FSM.]

  SideEffects []

  SeeAlso     [TrCompose]

******************************************************************************/

Ddi_Dd_t *
TrBuildTransClosure (
  Tr_Mgr_t *trMgr        /* TR manager */,
  Ddi_Dd_t  *TR          /* Transition Relation */,
  Ddi_VarArray_t *s      /* Present state vars */,
  Ddi_VarArray_t *y      /* Next state vars */,
  Ddi_VarArray_t *z      /* Intermediate state vars */
)
{
  Ddi_VarSet_t *suppZ;   /* Set of new variables z */
  Ddi_Dd_t *TR_i, *TC, *TC0, *TSQ, *TSQ_i;
  Ddi_Dd_t *temp;
  int i,
      size,              /* size of T.C. */
      peak=0;            /* max size of T.C. */
  long step;             /* number of iterations */
  int maxIter;           /* max allowed squaring iterations */
  Ddi_DdMgr *dd;         /* dd manager */

  Pdtutil_Assert((Ddi_DdReadCode(TR)!=DDI_CPART_BOOLF),
    "Trying to apply squaring to conjunctively partitioned TR");

  dd  = Ddi_DdReadMgr(TR); 
  TC0 = NULL;

  /*
   *  Disjunctively partitioned TR
   */

  if (Ddi_DdReadCode(TR)==DDI_DPART_BOOLF) {

    if (Tr_MgrReadVerbosity(trMgr) >= Pdtutil_VerbLevelUsrMed_c) {
      fprintf (stdout, "Squaring of a disjunctively partitioned TR\n");
      fprintf (stdout, "Each partition is closed locally\n");
      fprintf (stdout, "Only a partial closure is globally obtained\n");
    }

    TSQ = NULL;
    for (i=0;i<Ddi_DdReadPartNum(TR);i++) {
      TR_i = Ddi_DdReadIthPart(TR,i);
      TSQ_i = TrBuildTransClosure(trMgr, TR_i, s, y, z);
      TSQ = Ddi_DdEvalFree(Ddi_DdAddDpart(TSQ,TSQ_i),TSQ);
      Ddi_DdFree(TSQ_i);
    }
    return TSQ;

  }

  suppZ = Ddi_VarArrayToVarSet(dd,z);

  /* TC is the relation to be composed */ 
  TC = Ddi_DdDup(TR);

  /* TC0 is the relation to be or-ed after composition */ 
  switch (Tr_MgrReadSquaringMethod(trMgr)) {
    case TR_SQUARING_FULL:
      TC0 = Ddi_DdDup(TC);
      break;
    case TR_SQUARING_POWER:
      TC0 = Ddi_DdDup(Ddi_MgrReadZero(dd));
      break;
  }

  /*-------------------- START Iterative squaring --------------------------*/

  maxIter     = Tr_MgrReadMaxIter(trMgr);
  TSQ = Ddi_DdDup(Ddi_MgrReadOne(dd));

  for (step=1; (maxIter<0)||(step<=maxIter); step++) {

    /* composition */
    temp = TrCompose(trMgr,TC,TC,s,y,z,suppZ);
    /* or with TC0 */
    temp = Ddi_DdEvalFree(Ddi_DdOr(temp,TC0),temp);
    if (Ddi_DdEqual(temp,TC))   /* fixed point reached */
      break;

    if (Tr_MgrReadVerbosity (trMgr) >= Pdtutil_VerbLevelAppMed_c) {
      size = Ddi_DdSize(TC);
      if (size>peak)  
         peak = size; /* remember the max value of size */
      printf("It.Squaring step %ld : [ T.C. = %d[%d] ]\n",step,
        size,Ddi_DdSize(TSQ));
      fflush(stdout);
    }

    Ddi_DdFree(TC);
    TC = temp;
    
    /* TSQ is the upgraded closure */
    switch (Tr_MgrReadSquaringMethod(trMgr)) {
      case TR_SQUARING_FULL:
        Ddi_DdFree(TSQ);
        TSQ = Ddi_DdDup(TC);
        break;
      case TR_SQUARING_POWER:
        TSQ = Ddi_DdEvalFree(Ddi_DdOr(TC,TSQ),TSQ);
        break;
    }

  }
  /*----------------------- END Iterative squaring -------------------------*/
 
  Ddi_VarSetFree(suppZ);

  Ddi_DdFree(TC);
  Ddi_DdFree(TC0);

  if (Tr_MgrReadVerbosity (trMgr) >= Pdtutil_VerbLevelAppMed_c) {
    Pdtutil_ChrPrint (stdout, '*', 50);
    fprintf (stdout, "\nIterative Squaring Results Summary:\n");
    fprintf (stdout, "# Size of Transitive Closure = %d\n",Ddi_DdSize(TSQ));
    fprintf (stdout, "# Number of iterations = %ld\n",step-1);
    fprintf (stdout, "# |TC| peak = %d\n\n",peak);
    Pdtutil_ChrPrint (stdout, '*', 50);
    fprintf (stdout, "\n");
  }

  return (TSQ);
}


/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Compute : EXIST(z) (f(s,z) * g(z,y) )]

  Description [Computes the relational composition of f(s,z) and g(z,y), 
    after proper variable substitutions on the initial
    f(s,y): initially both f and g are expressed as functions
    of (s,y).
    This is the basic step of iterative squaring using monolithic
    transition relations.
    ]

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

static Ddi_Dd_t *
TrCompose (
  Tr_Mgr_t *trMgr         /* Tr Manager */,
  Ddi_Dd_t *f             /* first function */,
  Ddi_Dd_t *g             /* second function */,
  Ddi_VarArray_t *s       /* array of s variables */,
  Ddi_VarArray_t *y       /* array of y variables */,
  Ddi_VarArray_t *z       /* array of z variables */,
  Ddi_VarSet_t *suppZ     /* z variables as a set */
)
{
  Ddi_Dd_t *f1,*g1,*fSq;

  Pdtutil_Assert(Ddi_DdReadCode(f)==DDI_MONO_BOOLF,
    "Squaring step received a non monolithic function");

  f1 = Ddi_DdSwapVariables(f,z,y); 
  g1 = Ddi_DdSwapVariables(g,z,s);

  if (Tr_MgrReadVerbosity(trMgr) >= Pdtutil_VerbLevelAppMax_c) {
    fprintf(stdout,"Composing: [%d]*[%d] -> ",
      Ddi_DdSize(f1),Ddi_DdSize(g1));fflush(stdout);
  } 

  fSq = Ddi_DdAndExist(f1,g1,suppZ);

  if (Tr_MgrReadVerbosity(trMgr) >= Pdtutil_VerbLevelAppMax_c) {
    fprintf(stdout,"[%d]\n",
      Ddi_DdSize(fSq));fflush(stdout);
  } 

  Ddi_DdFree(f1);
  Ddi_DdFree(g1);

  return (fSq);  
}

