/**CFile***********************************************************************

  FileName    [trMgr.c]

  PackageName [tr]

  Synopsis    [Functions to handle a TR manager]

  Description []

  SeeAlso     []

  Author      [Gianpiero Cabodi and Stefano Quer]

  Copyright   [This file was created at the Politecnico di Torino,
    Torino, Italy.
    The  Politecnico di Torino makes no warranty about the suitability of
    this software for any purpose.
    It is presented on an AS IS basis.]

  Revision    []

******************************************************************************/

#include "trInt.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/
  
/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/
  
/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static void SetSortWeights(Tr_Mgr_t *trMgr, char *weights);
static char * ShowSortWeights(Tr_Mgr_t *trMgr);

/**AutomaticEnd***************************************************************/

/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Creates a DdManager.]

  Description []

  SideEffects [none]

  SeeAlso     [Ddi_DdiMgrQuit]

******************************************************************************/

Tr_Mgr_t *
Tr_MgrInit (
  char *trName    /* Name of the FSM structure */,
  Ddi_DdMgr *dd   /* Decision Diagram Manager */
)
{
  Tr_Mgr_t *trMgr;

  trMgr = Pdtutil_Alloc(Tr_Mgr_t, 1);
  if (trMgr == NULL) {
    Pdtutil_Warning (1, "Out of memory.");
    return (NULL);
  }

  /*------------------- Set the Name of the FSM Structure ------------------*/

  if (trName != NULL) {
     Tr_MgrSetTrName (trMgr, trName);
  } else {
     Tr_MgrSetTrName (trMgr, "tr");
  }

#if 0 
  /* not yet supported Ddi_VarArrayReadtrMgr */
  Pdtutil_Assert(Ddi_DdReadtrMgr(tr)==Ddi_VarArrayReadtrMgr(i),
    "Different DD Managers in use");
  Pdtutil_Assert(Ddi_DdReadtrMgr(tr)==Ddi_VarArrayReadtrMgr(ns), 
    "Different DD Managers in use");
  Pdtutil_Assert(Ddi_DdReadtrMgr(tr)==Ddi_VarArrayReadtrMgr(ps), 
    "Different DD Managers in use");
#endif

  trMgr->DdiMgr = dd;
  trMgr->tr = NULL;
  trMgr->i  = NULL;
  trMgr->ns = NULL;
  trMgr->ps = NULL;
  trMgr->settings.verbosity = Pdtutil_VerbLevelNone_c;
  trMgr->settings.sortMethod = Tr_SortNone_c;
  trMgr->settings.ClustThreshold = TR_IMAGE_CLUSTER_SIZE;
  trMgr->settings.SortW[0] = TRSORT_DEFAULT_W1;
  trMgr->settings.SortW[1] = TRSORT_DEFAULT_W2;
  trMgr->settings.SortW[2] = TRSORT_DEFAULT_W3;
  trMgr->settings.SortW[3] = TRSORT_DEFAULT_W4;
  trMgr->settings.SortW[4] = TRSORT_DEFAULT_W5;
  trMgr->settings.maxIter = -1;
  trMgr->settings.SquaringMethod = TR_SQUARING_FULL;
  
  trMgr->trTime = 0;

  return (trMgr);
}

/**Function********************************************************************

  Synopsis     [Closes a Transition Relation Manager.]

  Description  [Closes a Transition Relation Manager freeing all the
    correlated fields.]

  SideEffects  [none]

  SeeAlso      [Ddi_DdiMgrInit]

******************************************************************************/

void
Tr_MgrQuit(
  Tr_Mgr_t *trMgr  /* Tr manager */
)
{
  Ddi_DdFree (trMgr->tr);
  Ddi_VarArrayFree (trMgr->i);
  Ddi_VarArrayFree (trMgr->ps);
  Ddi_VarArrayFree (trMgr->ns);

  Pdtutil_Free (trMgr);
}

/**Function********************************************************************

  Synopsis    [Prints Statistics on a Transition Relation Manager.]

  Description [Prints Statistics on a Transition Relation Manager on
    standard output.]

  SideEffects []

  SeeAlso     []

******************************************************************************/

int
Tr_MgrPrintStats (
  Tr_Mgr_t *trMgr  /* Tr manager */
  )
{ 
  fprintf (stdout, "Transition Relation Manager Name %s\n",
    trMgr->trName);
  fprintf (stdout, "Total CPU Time: %s\n",
    util_print_time (trMgr->trTime));
 
  return (1);
}

/**Function********************************************************************

  Synopsis    [Performs an Operation on a Transition Relation Manager.]

  Description [Performs an Operation on a Transition Relation Manager.
    The allowed operations are specified by the enumerated type
    Pdtutil_MgrOp_t. Returns the result of the operation, the enumerated
    type Pdtutil_MgrRet_t.]

  SideEffects []

  SeeAlso     [CmdMgrOperation, CmdRegOperation, Fsm_MgrOperation,
    Trav_MgrOperation]

******************************************************************************/

int
Tr_MgrOperation (
  Tr_Mgr_t *trMgr                 /* TR Manager */,
  char *string                    /* String */,
  Pdtutil_MgrOp_t operationFlag   /* Operation Flag */,
  void **voidPointer              /* Generic Pointer */,
  Pdtutil_MgrRet_t *returnFlagP   /* Type of the Pointer Returned */
  )
{
  char *stringFirst, *stringSecond;
  Ddi_DdMgr *defaultDdMgr, *tmpDdMgr;

  /*------------ Check for the Correctness of the Command Sequence ----------*/

  if (trMgr == NULL) {
    Pdtutil_Warning (1, "Command Out of Sequence.");
    return (1);
  }

  defaultDdMgr = Tr_MgrReadDdiMgrDefault (trMgr);
  
  /*----------------------- Take Main and Secondary Name --------------------*/
 
  Pdtutil_ReadSubstring (string, &stringFirst, &stringSecond);

  /*----------------------- Operation on the FSM Manager --------------------*/

  if (stringFirst == NULL) {
    switch (operationFlag) {
      case Pdtutil_MgrOpOptionShow_c:
        fprintf (stdout, "TransitionRelationManager verbosity %s (%d)\n",
          Pdtutil_VerbosityEnum2String (Tr_MgrReadVerbosity (trMgr)),
          Tr_MgrReadVerbosity (trMgr));
        fprintf (stdout,
          "TransitionRelationManager sortWeights %s\n",
          ShowSortWeights (trMgr)); 
        fprintf (stdout, "TransitionRelationManager clusterThreshold %d\n",
          Tr_MgrReadClustThreshold (trMgr));
        break;
      case Pdtutil_MgrOpStats_c:
        Tr_MgrPrintStats (trMgr);
        break;
      case Pdtutil_MgrOpMgrRead_c:
        *voidPointer = (void *) Tr_MgrReadDdiMgrDefault (trMgr);
        *returnFlagP= Pdtutil_MgrRetDdMgr_c;
        break;
      case Pdtutil_MgrOpMgrSet_c:
        tmpDdMgr = (Ddi_DdMgr *) *voidPointer;
        Tr_MgrSetDdiMgrDefault (trMgr, tmpDdMgr);
        break;
      default:
        Pdtutil_Warning (1, "Operation Non Allowed on TR Manager");
        break;
    }

    return (0);
  }

  /*----------------------------- Package is DDI ----------------------------*/

  if (strcmp (stringFirst, "ddm")==0) {
    Ddi_MgrOperation (&(trMgr->DdiMgr), stringSecond, operationFlag,
      voidPointer, returnFlagP);

    if (*returnFlagP == Pdtutil_MgrOpMgrDelete_c) {
      trMgr->DdiMgr = NULL;
    }

    Pdtutil_Free (stringFirst);
    Pdtutil_Free (stringSecond);
    return (0);
  }

  /*--------------------------------- Options -------------------------------*/

  if (strcmp(stringFirst, "verbosity")==0) {
    switch (operationFlag) {
      case Pdtutil_MgrOpOptionSet_c:
        Tr_MgrSetVerbosity (trMgr,
          Pdtutil_VerbosityString2Enum (*voidPointer));
        break;
      case Pdtutil_MgrOpOptionShow_c:
        fprintf (stdout, "Verbosity %s (%d)\n",
          Pdtutil_VerbosityEnum2String (Tr_MgrReadVerbosity (trMgr)),
          Tr_MgrReadVerbosity (trMgr));
        break;
      default:
        Pdtutil_Warning (1, "Operation Non Allowed on TR Manager");
        break;
    }

    Pdtutil_Free (stringFirst);
    Pdtutil_Free (stringSecond);
    return (0);
  }

  if (strcmp(stringFirst, "sortWeights")==0) {
    switch (operationFlag) {
      case Pdtutil_MgrOpOptionSet_c:
	SetSortWeights (trMgr, (char *) *voidPointer); 
        break;
      case Pdtutil_MgrOpOptionShow_c:
        fprintf (stdout, "SortWeights %s\n", ShowSortWeights (trMgr)); 
        break;
      default:
        Pdtutil_Warning (1, "Operation Non Allowed on TR Manager");
        break;
    }

    Pdtutil_Free (stringFirst);
    Pdtutil_Free (stringSecond);
    return (0);
  }

  if (strcmp(stringFirst, "clustTh")==0) {
    switch (operationFlag) {
      case Pdtutil_MgrOpOptionSet_c:
        Tr_MgrSetClustThreshold (trMgr, atoi (*voidPointer));
        break;
      case Pdtutil_MgrOpOptionShow_c:
        fprintf (stdout, "ClusterThreshold %d\n",
          Tr_MgrReadClustThreshold (trMgr));
        break;
      default:
        Pdtutil_Warning (1, "Operation Non Allowed on TR Manager");
        break;
    }

    Pdtutil_Free (stringFirst);
    Pdtutil_Free (stringSecond);
    return (0);
  }

  /*-------------------------- Transition Relation --------------------------*/

  if (strcmp(stringFirst, "tr")==0) {
    Ddi_BddOperation (defaultDdMgr, &(trMgr->tr), stringSecond,
      operationFlag, voidPointer, returnFlagP);

    Pdtutil_Free (stringFirst);
    Pdtutil_Free (stringSecond);
    return (0);
  }

  /*------------------------------- No Match --------------------------------*/

  Pdtutil_Warning (1, "Operation on TR manager not allowed");
  return (1);
}

/**Function********************************************************************

  Synopsis    [Set default DDi Mgr on the Transition Relation Manager.]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

void
Tr_MgrSetDdiMgrDefault (
  Tr_Mgr_t *trMgr      /* trersal manager */,
  Ddi_DdMgr *mgr         /* dd Manager */
)
{
  trMgr->DdiMgr = mgr;

  return;
}


/**Function*******************************************************************

  Synopsis    [Set sort settings in the Transition Relation Manager.]

  Description [Set sort settings in the Transition Relation Manager
    reading them from a string.]

  SideEffects [none]

  SeeAlso     []

*****************************************************************************/

static void
SetSortWeights(
  Tr_Mgr_t *trMgr            /* Traversal Manager */,
  char *weights              /* Weights */
  )
{
  float w[5];
  int i, n;

  n = sscanf (weights, "%f/%f/%f/%f/%f", &w[0],&w[1],&w[2],&w[3],&w[4]);
  
  if (n < 5) {
    Pdtutil_Warning (1, "Error reading weights; Format: w1/w2/w3/w4/w5.");
    return;
  }

  for (i=0; i<5; i++) {
    Tr_MgrSetSortW (trMgr, i+1, (double)w[i]);
  }

  return;
}
  
/**Function*******************************************************************

  Synopsis    [Show sort settings.]

  Description [Show sort settings reading them from the Transition
    Relation manager. Return a string containing them.]

  SideEffects [none]

  SeeAlso     []

*****************************************************************************/

static char *
ShowSortWeights(
  Tr_Mgr_t *trMgr            /* Traversal Manager */
  )
{
  static char buf[200];
  double w[5];
  int i;

  for (i=0; i<5; i++) {
    w[i] = Tr_MgrReadSortW(trMgr,i+1);
  }
  sprintf (buf, "%g/%g/%g/%g/%g", w[0],w[1],w[2],w[3],w[4]);

  return (buf);
}

/**Function********************************************************************

  Synopsis     [Read default DDi Mgr]

  Description  []

  SideEffects  [none]

  SeeAlso      []

******************************************************************************/

Ddi_DdMgr *
Tr_MgrReadDdiMgrDefault (
  Tr_Mgr_t *trMgr      /* trersal manager */
)
{
  return (trMgr->DdiMgr);
}

/**Function********************************************************************

  Synopsis     [Read PI array]

  Description  []

  SideEffects  [none]

  SeeAlso      []

******************************************************************************/

Ddi_VarArray_t *
Tr_MgrReadI(
  Tr_Mgr_t *trMgr            /* tr manager */
)
{
  return (trMgr->i);
}

/**Function********************************************************************

  Synopsis    [Set the PI array]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

void
Tr_MgrSetI(
  Tr_Mgr_t *trMgr       /* Transition Relation Manager */,
  Ddi_VarArray_t *i     /* Array of variables */
)
{
  trMgr->i = Ddi_VarArrayCopy(trMgr->DdiMgr,i);

  return;
}

/**Function********************************************************************

  Synopsis    [Read PS array in the Transition Relation Manager.]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

Ddi_VarArray_t *
Tr_MgrReadPS(
  Tr_Mgr_t *trMgr            /* tr manager */
)
{
  return (trMgr->ps);
}

/**Function********************************************************************

  Synopsis    [Set the PS array]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

void
Tr_MgrSetPS(
  Tr_Mgr_t *trMgr        /* Transition Relation Manager */,
  Ddi_VarArray_t *ps     /* Array of Variables */
)
{
  trMgr->ps = Ddi_VarArrayCopy(trMgr->DdiMgr,ps);

  return;
}

/**Function********************************************************************

  Synopsis    [Read NS array in the Transition Relation Manager.]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

Ddi_VarArray_t *
Tr_MgrReadNS(
  Tr_Mgr_t *trMgr    /* tr manager */
)
{
  return (trMgr->ns);
}

/**Function********************************************************************

  Synopsis    [Set the NS array in the Transition Relation Manager.]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

void
Tr_MgrSetNS(
  Tr_Mgr_t *trMgr      /* Transition Relation Manager */,
  Ddi_VarArray_t *ns   /* Array of variables */
)
{
  trMgr->ns = Ddi_VarArrayCopy(trMgr->DdiMgr,ns);

  return;
}

/**Function********************************************************************

  Synopsis    [Read TR in the Transition Relation Manager.]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

Ddi_Dd_t *
Tr_MgrReadTr(
  Tr_Mgr_t *trMgr   /* Transition Relation Manager */
)
{
  return (trMgr->tr);
}

/**Function********************************************************************

  Synopsis    [Set the Transition Relation in the Transition Relation
    Manager.]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

void
Tr_MgrSetTr(
  Tr_Mgr_t *trMgr    /* Transition Relation Manager */,
  Ddi_Dd_t *tr       /* Transition Relation */
)
{
  if (trMgr->tr != NULL) {
    Ddi_DdFree(trMgr->tr);
  }

  trMgr->tr = Ddi_DdCopy(trMgr->DdiMgr,tr);

  return;
}


/**Function********************************************************************

  Synopsis    [Read verbosity]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

Pdtutil_VerbLevel_e
Tr_MgrReadVerbosity(
  Tr_Mgr_t *trMgr            /* Transition Relation Manager */
)
{
  return (trMgr->settings.verbosity);
}

/**Function********************************************************************

  Synopsis    [Set the period for verbosity enabling.]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

void
Tr_MgrSetVerbosity(
  Tr_Mgr_t *trMgr                 /* Transition Relation Manager */,
  Pdtutil_VerbLevel_e verbosity   /* Verbosity */
)
{
  trMgr->settings.verbosity = verbosity;

  return;
}

/**Function********************************************************************

  Synopsis    [Read sort method]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

Tr_Sort_e
Tr_MgrReadSortMethod(
  Tr_Mgr_t *trMgr     /* Transition Relation Manager */
)
{
  return (trMgr->settings.sortMethod);
}

/**Function********************************************************************

  Synopsis    [Set the sort Method]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

void
Tr_MgrSetSortMethod(
  Tr_Mgr_t *trMgr            /* Transition Relation Manager */,
  Tr_Sort_e sortMethod       /* Method */
)
{
  trMgr->settings.sortMethod = sortMethod;

  return;
}

/**Function********************************************************************

  Synopsis    [Read cluster threshold]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

int
Tr_MgrReadClustThreshold(
  Tr_Mgr_t *trMgr      /* Transition Relation Manager */
)
{
  return (trMgr->settings.ClustThreshold);
}

/**Function********************************************************************

  Synopsis    [Set the cluster threshold]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

void
Tr_MgrSetClustThreshold(
  Tr_Mgr_t *trMgr          /* Transition Relation Manager */,
  int ClustThreshold       /* Threshold */
)
{
  trMgr->settings.ClustThreshold = ClustThreshold;

  return;
}

/**Function********************************************************************

  Synopsis    [Read sort weight]

  SideEffects [none]

  Description []

  SeeAlso     []

******************************************************************************/

double
Tr_MgrReadSortW(
  Tr_Mgr_t *trMgr            /* Transition Relation Manager */,
  int       i
)
{
  Pdtutil_Assert(((i<=5)&&(i>0)),"Weight index out of bounds: 1..5");
  return (trMgr->settings.SortW[i-1]);
}

/**Function********************************************************************

  Synopsis    [Set the sort weight]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

void
Tr_MgrSetSortW(
  Tr_Mgr_t *trMgr    /* Transition Relation Manager */,
  int i,
  double SortW       /* Weight */
)
{
  Pdtutil_Assert(((i<=5)&&(i>0)),"Weight index out of bounds: 1..5");
  trMgr->settings.SortW[i-1] = SortW;

  return;
}


/**Function********************************************************************

  Synopsis    []

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

void
Tr_MgrSetTrName (
  Tr_Mgr_t *trMgr     /* Transition Relation Manager */,
  char *trName        /* Traversal Manager Name */
)
{
  trMgr->trName = Pdtutil_StrDup (trName);

  return;
}

/**Function********************************************************************

  Synopsis    [Read verbosity]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

char *
Tr_MgrReadTrName (
  Tr_Mgr_t *trMgr    /* Transition Relation Manager */
)
{
  return (trMgr->trName);
}

/**Function********************************************************************

  Synopsis    [Read the maximum number of closure iterations.]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

int
Tr_MgrReadMaxIter(
  Tr_Mgr_t *trMgr    /* Transition Relation Manager */
)
{
  return (trMgr->settings.maxIter);
}


/**Function********************************************************************

  Synopsis    [Set the maximum number of closure iterations]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

void
Tr_MgrSetMaxIter(
  Tr_Mgr_t *trMgr    /* Transition Relation Manager */,
  int maxIter        /* max iterations */
)
{
  trMgr->settings.maxIter = maxIter;

  return;
}


/**Function********************************************************************

  Synopsis    [Read the squaring method]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

int
Tr_MgrReadSquaringMethod(
  Tr_Mgr_t *trMgr    /* Transition Relation Manager */
)
{
  return (trMgr->settings.SquaringMethod);
}


/**Function********************************************************************

  Synopsis    [Set the squaring method]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/

void
Tr_MgrSetSquaringMethod(
  Tr_Mgr_t *trMgr    /* Transition Relation Manager */,
  int method         /* Method */
)
{
  trMgr->settings.SquaringMethod = method;

  return;
}
