/**CHeaderFile*****************************************************************

  FileName    [trav.h]

  PackageName [trav]

  Synopsis    [Main module for a simple traversal of finite state machine]

  Description [This package contains the main function to exploring the state
    space of a FSM.<br>
    There are three methods for image computation:
    <ol>
    <li> <b>Monolithic:</b> This is the most naive approach possible.<br>
       This technique is based on building <em>monolithic transition 
       relation</em>. It is the conjunction of all latches transition
       relation. If we use <em>y</em> to denote the next state vector,
       <em>s</em> the present state vector, <em>x</em> the input 
       vector and <em>delta()</em> the next state function, we define the
       trasition relation of <em>i</em>-th latch to be the function 
       Ti (x,s,y) = yi <=> delta(i)(x,s).<br> Then, for a FSM of n latches,
       the monolhitic transition relation is:
       <p>
       T(x,s,y) = T1(x,s,y)*T2(x,s,y)* ... *Tn(x,s,y)
       <p>
       When the monolithic TR is built, the traversal algorithm is executed.
       This method is suitable for circuits with less than 20 latches

    <li> <b>IWLS95:</b> This technique is based on the early quantification
    heuristic.
       The initialization process consists of following steps:   
       <ul>
       <li> Create the clustered TR from transition relation by clustering 
            several function together. The size of clustering is controlled
            by a threshold value controlled by the user.
       <li> Order the clusters using the algorithm given in
            "Efficient BDD Algorithms for FSM Synthesis and
            Verification", by R. K. Ranjan et. al. in the proceedings of
            IWLS'95. 
       <li> For each cluster, quantify out the quantify variables which
            are local to that particular cluster.
       </ul>
    <li> <b>Iterative squaring:</b> This technique is based on building the
       <em>transitive closure (TC)</em> of a monolithic TR. Afterwards, TC
       replace TR in the traversal algorithm. 
    </ol>
    ]
 
  SeeAlso     []  

  Author      [Gianpiero Cabodi and Stefano Quer]

  Copyright   [This file was created at the Politecnico di Torino,
    Torino, Italy. 
    The  Politecnico di Torino makes no warranty about the suitability of 
    this software for any purpose.  
    It is presented on an AS IS basis.]
  
  Revision    []

******************************************************************************/

#ifndef _TRAV
#define _TRAV

#include "pdtutil.h"
#include "ddi.h"
#include "part.h"
#include "fsm.h"

/*---------------------------------------------------------------------------*/
/* Nested includes                                                           */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/* smoothed variables */
#define TRAV_SMOOTH_SX     0
#define TRAV_SMOOTH_S      1

/* buffer size */
#define TRAV_FILENAME_BUFSIZE 80


/*---------------------------------------------------------------------------*/
/* Structure declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/**Enum************************************************************************

  Synopsis    [Type for From Selection]

  Description []

******************************************************************************/

typedef enum
  {
  Trav_FromSelectNew_c,
  Trav_FromSelectReached_c,
  Trav_FromSelectTo_c,
  Trav_FromSelectCofactor_c,
  Trav_FromSelectRestrict_c,
  Trav_FromSelectBest_c,
  Trav_FromSelectSame_c
  }
Trav_FromSelect_e;


/**Enum************************************************************************

  Synopsis    [Type for Image Selection]

  Description []

******************************************************************************/

typedef enum
  {
   /* image method of symbolic traversal */
   Trav_ImgMethodSame_c,
   Trav_ImgMethodNone_c,
   Trav_ImgMethodMonolithic_c,                   /* old number 1 */
   Trav_ImgMethodIwls_c,                         /* old number 2 */
   Trav_ImgMethodApprox_c,                       /* old number 3 */
   /* variants of IWLS95 method */
   Trav_ImgMethodIwlsClassic_c,
   Trav_ImgMethodIwlsPdtrav_c,
   Trav_ImgMethodIwlsConstraint_c,
   Trav_ImgMethodIwlsRestrict_c
  }
Trav_ImgMethod_e;


/**Enum************************************************************************

  Synopsis    []

  Description []

******************************************************************************/

typedef struct Trav_Mgr_s Trav_Mgr_t;

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/


/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Function prototypes                                                       */
/*---------------------------------------------------------------------------*/

EXTERN void Trav_Main(Trav_Mgr_t *travMgr);
EXTERN Ddi_Dd_t * Trav_ImgMonoTr(Trav_Mgr_t *TravMgr, Ddi_Dd_t *TR, Ddi_VarArray_t *psv, Ddi_VarArray_t *nsv, Ddi_Dd_t *from, Ddi_VarSet_t *smoothV);
EXTERN Ddi_Dd_t * Trav_ImgConjPartTr(Trav_Mgr_t *travMgr, Ddi_Dd_t *TR, Ddi_VarArray_t *psv, Ddi_VarArray_t *nsv, Ddi_Dd_t *from, Ddi_VarSet_t *smoothV);
EXTERN Ddi_Dd_t * Trav_ImgDisjPartTr(Trav_Mgr_t *TravMgr, Ddi_Dd_t *TR, Ddi_VarArray_t *psv, Ddi_VarArray_t *nsv, Ddi_Dd_t *from, Ddi_VarSet_t *smoothV);
EXTERN Ddi_Dd_t * Trav_ImgDisjPartSet(Trav_Mgr_t *TravMgr, Ddi_Dd_t *TR, Ddi_VarArray_t *psv, Ddi_VarArray_t *nsv, Ddi_Dd_t *part_from, Ddi_VarSet_t *smoothV);
EXTERN Ddi_Dd_t * Trav_ImgApproxConjPartTr(Trav_Mgr_t *TravMgr, Ddi_Dd_t *TR, Ddi_VarArray_t *psv, Ddi_VarArray_t *nsv, Ddi_Dd_t *from, Ddi_VarSet_t *smoothV);
EXTERN Trav_Mgr_t * Trav_MgrInit(char *travName, Ddi_DdMgr *dd);
EXTERN void Trav_MgrQuit(Trav_Mgr_t *travMgr);
EXTERN int Trav_MgrPrintStats(Trav_Mgr_t *travMgr);
EXTERN int Trav_MgrOperation(Trav_Mgr_t *travMgr, char *string, Pdtutil_MgrOp_t operationFlag, void **voidPointer, Pdtutil_MgrRet_t *returnFlagP);
EXTERN Ddi_Dd_t * Trav_MgrReadTr(Trav_Mgr_t *travMgr);
EXTERN void Trav_MgrSetTr(Trav_Mgr_t *travMgr, Ddi_Dd_t *tr);
EXTERN int Trav_MgrReadLevel(Trav_Mgr_t *travMgr);
EXTERN int Trav_MgrReadProductPeak(Trav_Mgr_t *travMgr);
EXTERN void Trav_MgrSetProductPeak(Trav_Mgr_t *travMgr, int productPeak);
EXTERN Ddi_VarArray_t * Trav_MgrReadI(Trav_Mgr_t *travMgr);
EXTERN void Trav_MgrSetI(Trav_Mgr_t *travMgr, Ddi_VarArray_t *i);
EXTERN Ddi_VarArray_t * Trav_MgrReadPS(Trav_Mgr_t *travMgr);
EXTERN void Trav_MgrSetPS(Trav_Mgr_t *travMgr, Ddi_VarArray_t *ps);
EXTERN void Trav_MgrSetMgrAuxFlag(Trav_Mgr_t *travMgr, int flag);
EXTERN Ddi_VarArray_t * Trav_MgrReadNS(Trav_Mgr_t *travMgr);
EXTERN void Trav_MgrSetNS(Trav_Mgr_t *travMgr, Ddi_VarArray_t *ns);
EXTERN Ddi_Dd_t * Trav_MgrReadReached(Trav_Mgr_t *travMgr);
EXTERN void Trav_MgrSetReached(Trav_Mgr_t *travMgr, Ddi_Dd_t *reached);
EXTERN Ddi_Dd_t * Trav_MgrReadAssert(Trav_Mgr_t *travMgr);
EXTERN void Trav_MgrSetAssert(Trav_Mgr_t *travMgr, Ddi_Dd_t *assert);
EXTERN Ddi_Dd_t * Trav_MgrReadFrom(Trav_Mgr_t *travMgr);
EXTERN void Trav_MgrSetFrom(Trav_Mgr_t *travMgr, Ddi_Dd_t *from);
EXTERN char * Trav_MgrReadName(Trav_Mgr_t *travMgr);
EXTERN void Trav_MgrSetName(Trav_Mgr_t *travMgr, char *travName);
EXTERN int Trav_MgrReadSmoothVar(Trav_Mgr_t *travMgr);
EXTERN double Trav_MgrReadW1(Trav_Mgr_t *travMgr);
EXTERN double Trav_MgrReadW2(Trav_Mgr_t *travMgr);
EXTERN double Trav_MgrReadW3(Trav_Mgr_t *travMgr);
EXTERN double Trav_MgrReadW4(Trav_Mgr_t *travMgr);
EXTERN int Trav_MgrReadEnableDdR(Trav_Mgr_t *travMgr);
EXTERN int Trav_MgrReadEnableTrProfile(Trav_Mgr_t *travMgr);
EXTERN int Trav_MgrReadTrPruneThreshold(Trav_Mgr_t *travMgr);
EXTERN int Trav_MgrReadThreshold(Trav_Mgr_t *travMgr);
EXTERN int Trav_MgrReadPaDiTrTh(Trav_Mgr_t *travMgr);
EXTERN Part_Method_e Trav_MgrReadPartitionMethod(Trav_Mgr_t *travMgr);
EXTERN void Trav_MgrSetPartitionMethod(Trav_Mgr_t *travMgr, Part_Method_e partitionMethod);
EXTERN int Trav_MgrReadSquaring(Trav_Mgr_t *travMgr);
EXTERN Pdtutil_VerbLevel_e Trav_MgrReadVerbosity(Trav_Mgr_t *travMgr);
EXTERN void Trav_MgrSetSorting(Trav_Mgr_t *travMgr, int sorting);
EXTERN int Trav_MgrReadSorting(Trav_Mgr_t *travMgr);
EXTERN int Trav_MgrReadMaxIter(Trav_Mgr_t *travMgr);
EXTERN int Trav_MgrReadLogPeriod(Trav_Mgr_t *travMgr);
EXTERN int Trav_MgrReadSavePeriod(Trav_Mgr_t *travMgr);
EXTERN char * Trav_MgrReadSavePeriodName(Trav_Mgr_t *travMgr);
EXTERN Trav_ImgMethod_e Trav_MgrReadImgMethod(Trav_Mgr_t *travMgr);
EXTERN int Trav_MgrReadImgCofactor(Trav_Mgr_t *travMgr);
EXTERN Trav_FromSelect_e Trav_MgrReadFromSelect(Trav_Mgr_t *travMgr);
EXTERN int Trav_MgrReadPartThFrom(Trav_Mgr_t *travMgr);
EXTERN int Trav_MgrReadPartThTr(Trav_Mgr_t *travMgr);
EXTERN void Trav_MgrSetDdiMgrTr(Trav_Mgr_t *travMgr, Ddi_DdMgr *mgrTr);
EXTERN void Trav_MgrSetDdiMgrDefault(Trav_Mgr_t *travMgr, Ddi_DdMgr *mgrTr);
EXTERN void Trav_MgrSetDdiMgrR(Trav_Mgr_t *travMgr, Ddi_DdMgr *mgrR);
EXTERN Ddi_DdMgr * Trav_MgrReadDdiMgrDefault(Trav_Mgr_t *travMgr);
EXTERN Ddi_DdMgr * Trav_MgrReadDdiMgrTr(Trav_Mgr_t *travMgr);
EXTERN Ddi_DdMgr * Trav_MgrReadDdiMgrR(Trav_Mgr_t *travMgr);
EXTERN void Trav_MgrSetVerbosity(Trav_Mgr_t *travMgr, Pdtutil_VerbLevel_e verbosity);
EXTERN void Trav_MgrSetMaxIter(Trav_Mgr_t *travMgr, int maxIter);
EXTERN void Trav_MgrSetLogPeriod(Trav_Mgr_t *travMgr, int logPeriod);
EXTERN void Trav_MgrSetSavePeriod(Trav_Mgr_t *travMgr, int savePeriod);
EXTERN void Trav_MgrSetSavePeriodName(Trav_Mgr_t *travMgr, char *savePeriodName);
EXTERN void Trav_MgrSetImgMethod(Trav_Mgr_t *travMgr, Trav_ImgMethod_e imgMethod);
EXTERN void Trav_MgrSetImgCofactor(Trav_Mgr_t *travMgr, int imgCofactor);
EXTERN void Trav_MgrSetFromSelect(Trav_Mgr_t *travMgr, Trav_FromSelect_e fromSelect);
EXTERN void Trav_MgrSetPartThFrom(Trav_Mgr_t *travMgr, int threshold);
EXTERN void Trav_MgrSetPartThTr(Trav_Mgr_t *travMgr, int threshold);
EXTERN int Trav_MgrReadMgrAuxFlag(Trav_Mgr_t *travMgr);
EXTERN Ddi_Dd_t * Trav_Traversal(Trav_Mgr_t *travMgr, Ddi_Dd_t *TR, Ddi_Dd_t *from, Ddi_Dd_t *reached, Ddi_VarArray_t *psv, Ddi_VarArray_t *nsv, Ddi_VarSet_t *quantify);
EXTERN Trav_FromSelect_e Trav_FromSelectString2Enum(char *string);
EXTERN char * Trav_FromSelectEnum2String(Trav_FromSelect_e enumType);
EXTERN Trav_ImgMethod_e Trav_ImgMethodString2Enum(char *string);
EXTERN char * Trav_ImgMethodEnum2String(Trav_ImgMethod_e enumType);

/**AutomaticEnd***************************************************************/


#endif /* _TRAV */
