/**CFile***********************************************************************

  FileName    [travImg.c]

  PackageName [trav]

  Synopsis    [Functions for image/preimage computations]

  Description []

  SeeAlso     []  

  Author      [Gianpiero Cabodi and Stefano Quer]

  Copyright   [This file was created at the Politecnico di Torino,
    Torino, Italy. 
    The  Politecnico di Torino makes no warranty about the suitability of 
    this software for any purpose.  
    It is presented on an AS IS basis.]
  
  Revision    []

******************************************************************************/

#include "travInt.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/
  
/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/


/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of external functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Compute image/preimage of a monolithic transition relation]

  Description [Given the Transition Relation, the domain/codomain set and the 
              the set of quantifying variables, image/preimage is computed in a
              straighforward way through the AndExist operator]

  SeeAlso     [trav.c]

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

Ddi_Dd_t *
Trav_ImgMonoTr (
  Trav_Mgr_t    *TravMgr    /* Traversal manager */,
  Ddi_Dd_t *TR              /* Monolithic Transition Relation */,
  Ddi_VarArray_t *psv       /* Array of present state variables */,
  Ddi_VarArray_t *nsv       /* Array of next state variables */,
  Ddi_Dd_t *from            /* Constrain */,
  Ddi_VarSet_t *smoothV     /* Variables to be abstracted */ 
)
{
  Ddi_VarSet_t *suppTR,
               *smooth_from;   
  Ddi_Dd_t *to;
  Ddi_DdMgr *dd = Ddi_DdReadMgr(TR);        /* TR DD manager */
  Ddi_DdMgr *ddR = Ddi_DdReadMgr(from);     /* Reached DD manager */

  /* Isolate variables not in TR */
  /* @ this could be done once forall when Ddi_Dd_t will have a support
     field */
  suppTR = Ddi_DdSupport(TR); 

  smooth_from = Ddi_VarSetDifference(smoothV,suppTR);
  Ddi_VarSetFree(suppTR);

  /* MULVAR test */
  Ddi_MulvarInterleave(TR, from);
  /* end MULVAR test */

  from = Ddi_DdCopyWithAuxMgr (dd,from,TravMgr->ddiMgrAux);

  /* early smoothing on from */ 
  if (!(Ddi_VarSetIsVoid(smooth_from))) {
    from = Ddi_DdEvalFree(Ddi_DdExist(from,smooth_from),from);
  }

  to = Ddi_DdAndExist(from,TR,smoothV);

  Ddi_VarSetFree(smooth_from);
  Ddi_DdFree(from);

  /* Swap present state variables (PS) next state variables (NS) */
  to = Ddi_DdEvalFree (Ddi_DdSwapVariablesWithAuxMgr(to,nsv,psv,
    TravMgr->ddiMgrAux),to);

  /* Transfer to reached manager */
  to = Ddi_DdEvalFree(Ddi_DdCopyWithAuxMgr (ddR,to,TravMgr->ddiMgrAux),to);

  return (to);
}

/**Function********************************************************************

  Synopsis    [Compute image of a conjunctively partitioned transition
    relation.]

  Description []

  SideEffects [None]

  SeeAlso     [Part_DdArrayMultiwayAndExist]

******************************************************************************/

Ddi_Dd_t *
Trav_ImgConjPartTr (
  Trav_Mgr_t *travMgr        /* Traversal manager */,
  Ddi_Dd_t *TR               /* Partitioned TR */,
  Ddi_VarArray_t *psv        /* Array of present state variables */,
  Ddi_VarArray_t *nsv        /* Array of next state variables */,
  Ddi_Dd_t *from             /* Input constrain */,
  Ddi_VarSet_t *smoothV      /* Variables to be abstracted */
)
{
  Ddi_Dd_t *to;
  Ddi_Dd_t *Product;
  Ddi_DdMgr *dd = Ddi_DdReadMgr(TR);     /* TR DD manager */
  Ddi_DdMgr *ddR = Ddi_DdReadMgr(from);  /* Reached DD manager */
  int sizeMax;
  int verbosity = Trav_MgrReadVerbosity (travMgr);
  
  if (Ddi_DdReadCode(from)==DDI_DPART_BOOLF) {
    return (Trav_ImgDisjPartSet(travMgr, TR, psv, nsv, from, smoothV));
  }

  if (Ddi_DdReadCode(TR)==DDI_DPART_BOOLF) {
    return (Trav_ImgDisjPartTr(travMgr, TR, psv, nsv, from, smoothV));
  }

  if (Ddi_DdReadCode(TR)==DDI_MONO_BOOLF) {
    return (Trav_ImgMonoTr (travMgr, TR, psv, nsv, from, smoothV));
  }

  Pdtutil_Assert (Ddi_DdReadCode(TR)==DDI_CPART_BOOLF,
    "Unexpected TR decomposition in image computation");

  /* MULVAR test */
  Ddi_MulvarInterleave (TR, from);
  /* end MULVAR test */

  from = Ddi_DdCopyWithAuxMgr (dd, from, travMgr->ddiMgrAux);

  if (Trav_MgrReadImgCofactor(travMgr)==1) {
    /* constrain is done for all partitions */
    Ddi_MgrAutodynSuspend(dd);
    Product = Ddi_DdConstrain(TR,from);
    Ddi_MgrAutodynResume(dd);
  } else {
    /* insert 'from' as the first partition */
    Product = Ddi_DdDup(TR);
    Ddi_DdArrayInclude(Ddi_DdReadPart(Product),0,from);
  }

  /*@ still to provide optional sorting */

  to = Part_DdMultiwayLinearAndExist (Product, smoothV, -1, verbosity,
    &sizeMax);

  if (verbosity >= Pdtutil_VerbLevelAppMin_c) {
    fprintf (stdout, "[|ProductPeak|: %d]", sizeMax);
    fflush (stdout);
  }

  if (sizeMax > Trav_MgrReadProductPeak (travMgr)) {
    Trav_MgrSetProductPeak (travMgr, sizeMax);
  }

  Pdtutil_Assert (Ddi_DdReadCode(to)==DDI_MONO_BOOLF,
    "Unexpected partitioned to obtained from image");

  Ddi_DdFree (Product);
  Ddi_DdFree (from);

  /* Swap present state variables (PS) next state variables (NS) */
  to = Ddi_DdEvalFree (Ddi_DdSwapVariablesWithAuxMgr(to,nsv,psv,
    travMgr->ddiMgrAux),to);

  /* transfer to reached manager */
  to = Ddi_DdEvalFree (Ddi_DdCopyWithAuxMgr (ddR, to, travMgr->ddiMgrAux),
    to);

  return (to);
}

/**Function********************************************************************

  Synopsis    [Compute image of a disjunctively partitioned transition
    relation]

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

Ddi_Dd_t *
Trav_ImgDisjPartTr (
  Trav_Mgr_t *TravMgr       /* Traversal manager */,
  Ddi_Dd_t *TR              /* Partitioned TR */,
  Ddi_VarArray_t *psv       /* Array of present state variables */,
  Ddi_VarArray_t *nsv       /* Array of next state variables */,
  Ddi_Dd_t *from            /* Input state set */,
  Ddi_VarSet_t *smoothV     /* Variables to be abstracted */
)
{
  Ddi_Dd_t *to, *to_i, *TR_i;
  int i;
  Ddi_DdMgr *ddR = Ddi_DdReadMgr(from);             /* dd manager */
  int verbosity = Trav_MgrReadVerbosity(TravMgr);
  
  if (Ddi_DdReadCode(TR)==DDI_CPART_BOOLF)
    return Trav_ImgConjPartTr (TravMgr, TR, psv, nsv, from, smoothV);

  to = Ddi_DdDup (Ddi_MgrReadZero (ddR));

  if (verbosity >= Pdtutil_VerbLevelUsrMax_c) {
    fprintf (stdout, "DIsj IMG (%d partitions)\n",
       Ddi_DdReadPartNum(TR));
  }
    
  for (i=0;i<Ddi_DdReadPartNum(TR);i++) {
    TR_i = Ddi_DdReadIthPart(TR,i);
    to_i = Trav_ImgConjPartTr (TravMgr, TR_i, psv, nsv, from, smoothV);
    if (verbosity >= Pdtutil_VerbLevelAppMax_c) {
      fprintf (stdout, "[%d]", Ddi_DdSize (to_i));
      fflush (stdout);
    }
    to = Ddi_DdEvalFree (Ddi_DdOr (to, to_i), to);
    Ddi_DdFree (to_i);
  }

  return (to);
}


/**Function********************************************************************

  Synopsis    [Compute image of a disjunctively partitioned from set.]

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

Ddi_Dd_t *
Trav_ImgDisjPartSet (
  Trav_Mgr_t *TravMgr      /* Traversal manager */,
  Ddi_Dd_t *TR             /* Partitioned TR */,
  Ddi_VarArray_t *psv      /* Array of present state variables */,
  Ddi_VarArray_t *nsv      /* Array of next state variables */,
  Ddi_Dd_t *part_from      /* Input state set */,
  Ddi_VarSet_t *smoothV    /* Variables to be abstracted */
)
{
  Ddi_Dd_t *to, *to_i, *from_i;
  int i;
  Ddi_DdMgr *ddR = Ddi_DdReadMgr(part_from);             /* dd manager */
  int verbosity = Trav_MgrReadVerbosity(TravMgr);
  
  Pdtutil_Assert (Ddi_DdReadCode(part_from)==DDI_DPART_BOOLF,
    "The from set of a partitioned image is not partitioned");

  to = Ddi_DdDup (Ddi_MgrReadZero (ddR));

  if (verbosity >= Pdtutil_VerbLevelUsrMax_c) {
    fprintf (stdout, "DIsj FROM (%d partitions)\n",
       Ddi_DdReadPartNum (part_from));
  }
    
  for (i=0; i<Ddi_DdReadPartNum (part_from); i++) {
    from_i = Ddi_DdReadIthPart (part_from, i);
    to_i = Trav_ImgConjPartTr (TravMgr, TR, psv, nsv, from_i, smoothV);
    if (verbosity >= Pdtutil_VerbLevelUsrMax_c) {
      fprintf (stdout, "[%d]",Ddi_DdSize (to_i));
      fflush (stdout);
    }
#if 1 /* warning: part_from modified ! */
    Ddi_DdFree(from_i);
    Ddi_DdArrayInsert (Ddi_DdReadPart(part_from), i, Ddi_MgrReadZero(ddR));
#endif
    to = Ddi_DdEvalFree(Ddi_DdOr(to,to_i),to);
    Ddi_DdFree(to_i);
  }

  return (to);
}

/**Function********************************************************************

  Synopsis    [Compute approx image of a conjunctively partitioned transition 
    relation.]

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

Ddi_Dd_t *
Trav_ImgApproxConjPartTr (
  Trav_Mgr_t *TravMgr      /* Traversal manager */,
  Ddi_Dd_t *TR             /* Partitioned TR */,
  Ddi_VarArray_t *psv      /* Array of present state variables */,
  Ddi_VarArray_t *nsv      /* Array of next state variables */,
  Ddi_Dd_t *from           /* Input state set */,
  Ddi_VarSet_t *smoothV    /* Variables to be abstracted */
)
{
  Ddi_Dd_t *to, *to_i, *TR_i;
  int i;
  Ddi_DdMgr *ddR = Ddi_DdReadMgr(from);             /* dd manager */
  int verbosity = Trav_MgrReadVerbosity (TravMgr);
  
  if (Ddi_DdReadCode(TR)!=DDI_CPART_BOOLF) {
    Pdtutil_Warning (1,
      "No conjunctive TR for approx trav - exact image done.");
    return (Trav_ImgConjPartTr (TravMgr, TR, psv, nsv, from, smoothV));
  }

  to = Ddi_DdDup (Ddi_MgrReadOne (ddR));

  if (verbosity > Pdtutil_VerbLevelUsrMin_c) {
    fprintf (stdout, "Approx IMG (%d partitions)\n",
       Ddi_DdReadPartNum (TR));
  }

  for (i=0; i<Ddi_DdReadPartNum(TR); i++) {
    TR_i = Ddi_DdReadIthPart (TR, i);
    to_i = Trav_ImgConjPartTr (TravMgr, TR_i, psv, nsv, from, smoothV);
    if (verbosity > Pdtutil_VerbLevelUsrMin_c) {
      fprintf (stdout, "[%d]",Ddi_DdSize(to_i));
      fflush(stdout);
    }
    to = Ddi_DdEvalFree (Ddi_DdAnd (to, to_i), to);
    Ddi_DdFree (to_i);
  }

  return (to);
}

/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

