/**CFile***********************************************************************

  FileName    [travTrav.c]

  PackageName [trav]

  Synopsis    [Functions to traverse a FSM]

  Description [External procedures included in this file are:
    <ul>
    <li>Trav_Traversal()
    </ul> 
    ]

  SeeAlso     []  

  Author      [Gianpiero Cabodi and Stefano Quer]

  Copyright   [This file was created at the Politecnico di Torino,
    Torino, Italy. 
    The  Politecnico di Torino makes no warranty about the suitability of 
    this software for any purpose.  
    It is presented on an AS IS basis.]
  
  Revision    []

******************************************************************************/

#include "travInt.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/
  
/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static Ddi_Dd_t * ComputeFrom(Ddi_Dd_t *to, Ddi_Dd_t *reached, int option);
static void TraceFrozenLatches(Trav_Mgr_t *travMgr, Ddi_DdMgr *dd, Ddi_VarArray_t *s, Ddi_VarArray_t *y, Ddi_VarSet_t *quantify, Ddi_Dd_t *TR, Ddi_Dd_t *from, Ddi_Dd_t *to, Ddi_Dd_t *reached, int *frozen_latches, Pdtutil_VerbLevel_e verbosity);

/**AutomaticEnd***************************************************************/

/*---------------------------------------------------------------------------*/
/* Definition of external functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [FSM traversal given a transition relation]

  Description [The pseudo-code of traversal algorithm is:<br>
               <code>
               traverse ( delta , S0 )<br>
  		   {<br>
               reached = from = S0;<br> 
               do<br> 
               {<br>
               to = Img  ( delta , from );<br>
               new = to - reached;<br>
               reached = reached + new;<br>
               from = new;<br>
               }<br>
               while ( new!= 0 )<br>
               <p>
               return reached ;<br>
               }<br>
               </code>
               We use the following notations:
               <ul>
               <li>from = initial set of states
               <li>to  = set of reached state in one step from current states
               <li>new  = new reached states 
               <li>reached  = set of reached states from 0 up to current
                   iteration
               </ul>]

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

Ddi_Dd_t *
Trav_Traversal (
  Trav_Mgr_t *travMgr      /* Traversal Manager */,
  Ddi_Dd_t *TR             /* Transition relation */,
  Ddi_Dd_t *from           /* Starting from */,
  Ddi_Dd_t *reached        /* Already reached states */,
  Ddi_VarArray_t *psv      /* Array of present state variables */,
  Ddi_VarArray_t *nsv      /* Array of next state variables */,
  Ddi_VarSet_t *quantify   /* Set of quantifying variables */
)
{
  int step;              /* number of iteration */
  Ddi_Dd_t *to;          /* new states reached every iteration */
  long startTime, currTime;
  Pdtutil_VerbLevel_e verbosity, saveVerbosity;
  int maxIter, imgMethod, logPeriod, savePeriod, partThFrom, l;
  char *savePeriodName, *tmpName; 
  Ddi_VarSet_t *part_vars=NULL;
  Ddi_DdMgr *dd;                /* TR DD manager */
  Ddi_DdMgr *ddR;               /* reached DD manager */

  Pdtutil_Assert (travMgr!=NULL, "NULL traversal manager");

  dd = travMgr->ddiMgrTr;
  ddR = travMgr->ddiMgrR;
  verbosity = Trav_MgrReadVerbosity (travMgr);
  maxIter = Trav_MgrReadMaxIter (travMgr);
  logPeriod = Trav_MgrReadLogPeriod (travMgr);

  savePeriod = Trav_MgrReadSavePeriod (travMgr);
  if (Trav_MgrReadSavePeriodName (travMgr) != NULL) {
    savePeriodName = Pdtutil_StrDup (Trav_MgrReadSavePeriodName (travMgr));
  } else {
    savePeriodName = Pdtutil_StrDup ("");
  }

  partThFrom = Trav_MgrReadPartThFrom (travMgr);
  imgMethod = Trav_MgrReadImgMethod (travMgr);

  if (verbosity >= Pdtutil_VerbLevelUsrMax_c) {
    if (ddR != dd) {
      fprintf (stdout, "Two dd managers used for traversal\n");
    }
  }

  if (Trav_MgrReadMgrAuxFlag (travMgr) == 1) {
    travMgr->ddiMgrAux = Ddi_MgrDup(dd);
  }

  startTime = util_cpu_time();

#if 0
@@@
  part_vars = Trav_MgrReadPartVars (travMgr);
#endif

  /*---------------------------- START Traversal ----------------------------*/

  travMgr->reached = Ddi_DdCopy(ddR,reached);
  travMgr->from    = Ddi_DdCopy(ddR,from);

  saveVerbosity = verbosity;

  for (step=1; (maxIter<0)||(step<=maxIter); step++,travMgr->level++) {

    /*
     * Cope with logPeriod through verbosity level
     */

    if ((step%logPeriod) == 0) {
      verbosity = travMgr->settings.verbosity = saveVerbosity;
    } else { 
      verbosity = travMgr->settings.verbosity = Pdtutil_VerbLevelNone_c;
    }

    /*
     * Save BDDs and Variable Orders every "savePeriod" cycles
     */

    if (savePeriod!=(-1) && (step%savePeriod) == 0) {
      l = strlen (savePeriodName) + 20;
      tmpName = Pdtutil_Alloc (char, l);

      /* Save Ord */
      sprintf (tmpName, "%sord-l%d.ord", savePeriodName, step);
      Ddi_MgrWriteOrd (dd, tmpName, NULL, Pdtutil_VariableOrderComment_c);

      /* Save From */
      sprintf (tmpName, "%sfrom-l%d.bdd", savePeriodName, step);
      Ddi_DdStore (travMgr->from, tmpName, DDDMP_MODE_BINARY,
        tmpName, NULL);
 
      /* Save Reached */
      sprintf (tmpName, "%sreached-l%d.bdd", savePeriodName, step);
      Ddi_DdStore (travMgr->reached, tmpName, DDDMP_MODE_BINARY,
        tmpName, NULL);

      Pdtutil_Free (tmpName);
    }

    /* check assertion */
    if (travMgr->assert != NULL) {
      if (!Ddi_DdIncluded(travMgr->reached,travMgr->assert)) {
        if (saveVerbosity >= Pdtutil_VerbLevelUsrMax_c) {
          fprintf (stdout, "TravLevel %d: Assertion failed!\n",
            travMgr->level);
        }
        break;
      }
    }

    if (verbosity >= Pdtutil_VerbLevelUsrMed_c) {
      fprintf (stdout, "TravLevel %d: ", travMgr->level);
      fprintf (stdout, "[|Tr|: %d] ", Ddi_DdSize (travMgr->tr));
      fprintf (stdout, "[|From|: %d]", Ddi_DdSize(travMgr->from));
      if (verbosity >= Pdtutil_VerbLevelAppMed_c) { 
        fprintf (stdout, "\n");
      }
      fflush(stdout);
    }

    /*
     * Compute image of next state function: to = Img(TR, from)
     */

    /* Do Disjoined Partitioning */
    if (partThFrom>=0) {
      travMgr->from = Ddi_DdEvalFree(
        Part_PartitionDisjSet (travMgr->from, part_vars,
          Trav_MgrReadPartitionMethod (travMgr), partThFrom, verbosity
          ),
        travMgr->from);
    }

    switch (imgMethod ) {

      case Trav_ImgMethodMonolithic_c:
        if (Ddi_DdReadCode(TR)!=DDI_MONO_BOOLF) {
          if (verbosity >= Pdtutil_VerbLevelUsrMed_c) { 
            fprintf (stdout, "Partitioned TR found\n");
            fprintf (stdout, "Monolithic TR automatically generated\n");
          }
          TR = Ddi_DdEvalFree(Ddi_DdPart2Mono (TR), TR);
          travMgr->tr = TR;
        }
        to = Trav_ImgMonoTr (travMgr, TR, psv, nsv, travMgr->from, quantify);
        break;

      case Trav_ImgMethodIwls_c:
        to = Trav_ImgConjPartTr (travMgr, TR, psv, nsv, travMgr->from,
          quantify);
        break;

      case Trav_ImgMethodApprox_c:
        to = Trav_ImgApproxConjPartTr (travMgr, TR, psv, nsv,
          travMgr->from, quantify);
        break;

      default: 
        to = Trav_ImgConjPartTr (travMgr, TR, psv, nsv, travMgr->from,
          quantify);
        break;
    }

    /* 
     * Print Statistics for this Iteration
     */

    if (verbosity >= Pdtutil_VerbLevelUsrMed_c) {
      if (verbosity >= Pdtutil_VerbLevelAppMed_c) { 
        fprintf (stdout, "\n");
      }
      fprintf (stdout, "[|To|: %d]", Ddi_DdSize (to));
      currTime = util_cpu_time();
      fprintf (stdout, "(Time: %s)\n", util_print_time (currTime - startTime));
      fflush(stdout);
    }

    /*
     * Check Fix Point
     */

    if (Ddi_DdIncluded (to, travMgr->reached)) {
      Ddi_DdFree (to);
      travMgr->from = Ddi_DdDup (Ddi_MgrReadZero (ddR));
      break;
    }

    /* 
     *  Prepare Sets for the New Iteration
     */
     
    /* Compute new from according to the selection option
       and transfer to main manager */
    Ddi_DdFree (travMgr->from);
    travMgr->from = ComputeFrom (to, travMgr->reached,
      Trav_MgrReadFromSelect(travMgr));

    /* Update the Set of Reached States */	
    travMgr->reached = Ddi_DdEvalFree (Ddi_DdOr (travMgr->reached, to), 
      travMgr->reached);
    Ddi_DdFree (to);

    /* 
     * Print Statistics for this Iteration
     */

    if (verbosity >= Pdtutil_VerbLevelUsrMax_c) {
      double nstates;

      fprintf (stdout, "[|Reached|: %d]", Ddi_DdSize (travMgr->reached));

      nstates = Ddi_CountMinterm(travMgr->reached, Ddi_ReadSize(dd)) / 
        pow (2.0,(double) (Ddi_ReadSize(dd) - Ddi_VarArray_n(psv)));
      fprintf (stdout, "[#ReachedStates: %g]\n", nstates);
    }

    if (verbosity >= Pdtutil_VerbLevelAppMin_c) {
      fprintf (stdout,
        "Memory: %ld Kbytes; Unique Table: %u nodes; DDI-DD Num: %d;\n",
        Ddi_ReadMemoryInUse(dd)/1024,Ddi_ReadKeys(dd),
        Ddi_MgrReadDdNum(dd));
      fprintf (stdout, "Cache: %u slots\n", Ddi_ReadCacheSlots(dd));
      fflush(stdout);
    }

  }

  verbosity = travMgr->settings.verbosity = saveVerbosity;

  /*-------------------------- Traversal ... End ----------------------------*/

  Pdtutil_Free (savePeriodName);

  if (part_vars != NULL) {
    Ddi_VarSetFree(part_vars);
  }

  if (travMgr->ddiMgrAux != NULL) {
    Ddi_DdiMgrQuit (travMgr->ddiMgrAux);
    travMgr->ddiMgrAux = NULL;
  }

  return (travMgr->reached);
}

/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Compute from given to and reached]

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

static Ddi_Dd_t *
ComputeFrom(
  Ddi_Dd_t *to               /* result of image computation */,
  Ddi_Dd_t *reached          /* old reached */,
  int option                 /* selection option */
)
{
  Ddi_Dd_t *from, *best, *tmp;
  int size, bestsize;

  best = NULL;
  bestsize = 0;

  switch (option) {
    case Trav_FromSelectBest_c:
      /* compute all following cases */
    case Trav_FromSelectNew_c:
      /* compute from as newly reached states */
      from = Ddi_DdAnd(to,(tmp=Ddi_DdNot(reached)));
      Ddi_DdFree(tmp);
      if (option != Trav_FromSelectBest_c) {
        return (from);
      }
      best = from;
      bestsize = Ddi_DdSize (best);
    case Trav_FromSelectReached_c: 
      from = Ddi_DdOr (to, reached);
      if (option != Trav_FromSelectBest_c)
        return (from);
      size = Ddi_DdSize(from);
      if (size < bestsize) {
        Ddi_DdFree (best);
        best = from;
        bestsize = size;
      }
      else Ddi_DdFree(from);
    case Trav_FromSelectTo_c:
      from = Ddi_DdDup(to);
      if (option != Trav_FromSelectBest_c)
        return (from);
      size = Ddi_DdSize(from);
      if (size < bestsize) {
        Ddi_DdFree(best);
        best = from;
        bestsize = size;
      }
      else Ddi_DdFree(from);
    case Trav_FromSelectCofactor_c: 
      from = Ddi_DdConstrain(to,(tmp=Ddi_DdNot(reached)));
      Ddi_DdFree(tmp);
      if (option != Trav_FromSelectBest_c)
        return (from);
      size = Ddi_DdSize(from);
      if (size < bestsize) {
        Ddi_DdFree(best);
        best = from;
        bestsize = size;
      }
      else Ddi_DdFree(from);
    case Trav_FromSelectRestrict_c: 
      from = Ddi_DdRestrict(to,(tmp=Ddi_DdNot(reached)));
      Ddi_DdFree(tmp);
      if (option != Trav_FromSelectBest_c)
        return (from);
      size = Ddi_DdSize(from);
      if (size < bestsize) {
        Ddi_DdFree(best);
        best = from;
        bestsize = size;
      }
      else Ddi_DdFree(from);
  }

  return (best);
}

/**Function********************************************************************

  Synopsis    [analyze latches keeping s=y]

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

static void
TraceFrozenLatches(
  Trav_Mgr_t *travMgr        /* traversal manager */,
  Ddi_DdMgr *dd              /* dd manager */,
  Ddi_VarArray_t *s          /* present state vars */,
  Ddi_VarArray_t *y          /* next state vars */,
  Ddi_VarSet_t *quantify     /* set of quantifying variables */,
  Ddi_Dd_t *TR               /* Transition Relation */,
  Ddi_Dd_t *from             /* result of image computation */,
  Ddi_Dd_t *to               /* result of image computation */,
  Ddi_Dd_t *reached          /* old reached */,
  int *frozen_latches        /* array of frozen latches */,
  Pdtutil_VerbLevel_e verbosity
)
{
  int i, ns;
  Ddi_Var_t *vs, *vy;
  Ddi_Dd_t *constr_TR, *tmp, *eq, *newY, *reachedY;

  /*  assert(Ddi_DdReadCode(TR)==DDI_MONO_BOOLF); */

  ns = Ddi_VarArray_n(s);

  reachedY = Ddi_DdSwapVariables(reached,s,y);
  newY = Ddi_DdAnd(to,(tmp=Ddi_DdNot(reachedY)));

  Ddi_DdFree(tmp);
  Ddi_DdFree(reachedY);

  for (i=0;i<ns;i++) {
    if (frozen_latches[i]==1) {
      vs = Ddi_VarArrayFetch(s,i);
      vy = Ddi_VarArrayFetch(y,i);
      eq = Ddi_DdXnor(Ddi_DdLiteral(dd,vs),Ddi_DdLiteral(dd,vy));

      constr_TR = Ddi_DdAnd(TR,eq);

      to = Trav_ImgConjPartTr(travMgr,constr_TR,y,y,from,quantify);

      if (!(Ddi_DdIncluded(newY,to)))
        frozen_latches[i]=0;

      Ddi_DdFree(eq);
      Ddi_DdFree(constr_TR);
      Ddi_DdFree(to);
    }
  }

  return;
}


